Package: maotai
Type: Package
Title: Tools for Matrix Algebra, Optimization and Inference
Version: 0.3.0
Authors@R: c(person("Kisung", "You", role = c("aut", "cre"), email = "kisung.you@outlook.com",comment=c(ORCID="0000-0002-8584-459X")))
Description: Matrix is an universal and sometimes primary object/unit in applied mathematics and statistics. We provide a number of algorithms for selected problems in optimization and statistical inference. For general exposition to the topic with focus on statistical context, see the book by Banerjee and Roy (2014, ISBN:9781420095388).
Encoding: UTF-8
License: MIT + file LICENSE
Suggests: covr, igraph, testthat (>= 3.0.0)
Imports: Matrix, Rcpp, Rdpack, RSpectra, Rtsne, RANN, cluster, labdsv,
        stats, utils, fastcluster, dbscan, pracma, gsignal
LinkingTo: Rcpp, RcppArmadillo, RcppDist
RdMacros: Rdpack
RoxygenNote: 7.3.3
URL: https://github.com/kisungyou/maotai
BugReports: https://github.com/kisungyou/maotai/issues
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-01-12 17:35:25 UTC; kyou
Author: Kisung You [aut, cre] (ORCID: <https://orcid.org/0000-0002-8584-459X>)
Maintainer: Kisung You <kisung.you@outlook.com>
Repository: CRAN
Date/Publication: 2026-01-13 06:10:23 UTC

# Generated by roxygen2: do not edit by hand

export(LEiDA)
export(WLbarycenter)
export(WLmedian)
export(WLpdist)
export(bmds)
export(boot.mblock)
export(boot.stationary)
export(bwproj)
export(cayleymenger)
export(checkdist)
export(checkmetric)
export(cmds)
export(cov2corr)
export(cov2pcorr)
export(dpmeans)
export(ecdfdist)
export(ecdfdist2)
export(ecdfdistS)
export(epmeans)
export(kmeanspp)
export(lyapunov)
export(matderiv)
export(metricdepth)
export(mmd2test)
export(movMF_convert)
export(movMF_info)
export(movMF_reduce_greedy)
export(movMF_reduce_partitional)
export(nef)
export(nem)
export(pdeterminant)
export(rotationS2)
export(shortestpath)
export(sylvester)
export(trio)
export(tsne)
export(weiszfeld)
import(Rdpack)
importFrom(Matrix,rankMatrix)
importFrom(RANN,nn2)
importFrom(RSpectra,eigs)
importFrom(Rcpp,evalCpp)
importFrom(Rtsne,Rtsne)
importFrom(cluster,agnes)
importFrom(cluster,pam)
importFrom(cluster,silhouette)
importFrom(dbscan,dbscan)
importFrom(fastcluster,hclust)
importFrom(gsignal,butter)
importFrom(gsignal,filtfilt)
importFrom(gsignal,hilbert)
importFrom(pracma,cross)
importFrom(pracma,detrend)
importFrom(stats,as.dist)
importFrom(stats,cov)
importFrom(stats,dist)
importFrom(stats,ecdf)
importFrom(stats,knots)
importFrom(stats,lm)
importFrom(stats,quantile)
importFrom(stats,rgamma)
importFrom(stats,rgeom)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,var)
importFrom(utils,packageVersion)
useDynLib(maotai)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Tools for Matrix Algebra, Optimization and Inference Problems

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/maotai)](https://CRAN.R-project.org/package=maotai)
[![R-CMD-check](https://github.com/kisungyou/maotai/workflows/R-CMD-check/badge.svg)](https://github.com/kisungyou/maotai/actions)
[![Codecov test
coverage](https://codecov.io/gh/kisungyou/maotai/branch/master/graph/badge.svg)](https://app.codecov.io/gh/kisungyou/maotai?branch=master)
[![metacran
downloads](https://cranlogs.r-pkg.org/badges/grand-total/maotai)](https://cran.r-project.org/package=maotai)
<!-- badges: end -->

**maotai** is an R package whose name is an acronym for **M**atrix
**A**lgebra, **O**p**T**imization, **A**nd **I**nference problems -
though I can’t deny motivation from one of [my father’s
favorite](https://en.wikipedia.org/wiki/Maotai) for the namesake. More
detailed introduction will be added later.

## Installation

Install the latest release from
[CRAN](https://CRAN.R-project.org/package=maotai) with:

``` r
install.packages("maotai")
```

or the up-to-date development version from github:

``` r
# install.packages("devtools")
devtools::install_github("kisungyou/maotai")
```

#' Pairwise Wasserstein-like Distance between two vMF distributions
#' 
#' Given a collection of von Misees-Fisher (vMF) distributions, compute the pairwise 
#' distance using the Wasserstein-like distance from an approximate Wasserstein geometry.
#' 
#' @param means An \eqn{(n \times p)} matrix where each row represents the mean 
#' direction of one of the \eqn{n} vMF distributions.
#' @param concentrations A length-\eqn{n} vector of nonnegative concentration parameters.
#' 
#' @return An \eqn{(n \times n)} matrix of pairwise distances.
#' 
#' @examples
#' \donttest{
#' # Set seed for reproducibility
#' set.seed(123)
#' 
#' # Generate two classes of mean directions around north and south poles
#' means1 = array(0,c(50,2)); means1[,2] = rnorm(50, mean=1, sd=0.25)
#' means2 = array(0,c(50,2)); means2[,2] = rnorm(50, mean=-1, sd=0.25)
#' means1 = means1/sqrt(rowSums(means1^2))
#' means2 = means2/sqrt(rowSums(means2^2))
#' 
#' # Concatenate the mean directions
#' data_means = rbind(means1, means2)
#' 
#' # Generate concentration parameters
#' data_concentrations = rnorm(100, mean=20, sd=1)
#' 
#' # Compute the pairwise distance matrix
#' pdmat = WLpdist(data_means, data_concentrations)
#' 
#' # Visualise the pairwise distance matrix
#' opar <- par(no.readonly=TRUE)
#' image(pdmat, main="Pairwise Wasserstein-like Distance")
#' par(opar)
#' }
#' 
#' @export
WLpdist <- function(means, concentrations){
  ###############################################
  # Preprocessing
  # means
  if (!is.matrix(means)){
    data_means = cpp_WL_normalise(as.matrix(means))
  } else {
    data_means = cpp_WL_normalise(means)
  }
  
  # concentrations
  data_kappa = as.vector(concentrations)
  if (length(data_kappa)!=base::nrow(data_means)){
    stop("* WLpdist : cardinalities of the means and concentrations do not match.")
  }
  
  # parameters
  N = base::nrow(means)
  d = base::ncol(means)
  
  
  ###############################################
  # COMPUTATION
  output = array(0,c(N,N))
  for (i in 1:(N-1)){
    for (j in (i+1):N){
      term1 = sphere_dist(as.vector(data_means[i,]), as.vector(data_means[j,]))
      term2sq = (d-1)*((1/sqrt(data_kappa[i]) - 1/sqrt(data_kappa[j]))^2)
      output[i,j] <- output[j,i] <- sqrt((term1*term1) + term2sq)
    }
  }
  
  ###############################################
  # RETURN
  return(output)
}

# auxiliary functions to deal with ECDF objects
# (1) elist_check   : list of 'ecdf' objects
# (2) elist_fform   : make a function form in a discrete grid
# (3) elist_epmeans : either a vector or something


# (1) elist_check ---------------------------------------------------------
#' @keywords internal
#' @noRd
elist_check <- function(elist){
  cond1 = (is.list(elist))
  cond2 = all(unlist(lapply(elist, inherits, "ecdf"))==TRUE)
  if (cond1&&cond2){
    return(TRUE)
  } else {
    return(FALSE)
  }
}

# (2) elist_fform ---------------------------------------------------------
#' @keywords internal
#' @noRd
elist_fform <- function(elist){
  nlist = length(elist)
  # compute knot points
  allknots = array(0,c(nlist,2))
  for (i in 1:nlist){
    tgt = stats::knots(elist[[i]])
    allknots[i,] = c(min(tgt), max(tgt))
  }
  mint = min(allknots[,1]) - 0.01
  maxt = max(allknots[,2]) + 0.01
  tseq  = seq(mint, maxt, length.out=1000)
  # return the list of y values
  outY = list()
  for (i in 1:nlist){
    tgt       = elist[[i]]
    outY[[i]] = tgt(tseq)
  }
  # return the result
  output = list()
  output$tseq = tseq
  output$fval = outY # list of function values
  return(output)
}

# (3) elist_epmeans -------------------------------------------------------
#' @keywords internal
#' @noRd
elist_epmeans <- function(elist){
  N      = length(elist)
  output = list()
  for (n in 1:N){
    tgt = elist[[n]]
    if (is.vector(tgt)&&(!any(is.infinite(tgt)))&&(!any(is.na(tgt)))){ # Case 1. just a vector
      output[[n]] = stats::ecdf(tgt)
    } else if (inherits(tgt, "ecdf")){
      output[[n]] = tgt
    } else {
      smsg = paste("* epmeans : ",n,"-th element from 'elist' is neither an 'ecdf' object nor a vector.")
      stop(smsg)
    }
  }
  return(output)
}

#' Bures-Wasserstein Projection
#' 
#' Projects a set of high-dimensional Gaussian distributions specified by their means and covariance matrices 
#' onto a lower-dimensional subspace using the distance-preserving projection method. 
#' 
#' @param means an \eqn{(n,p)} matrix of Gaussian means, where \eqn{n} is the number of Gaussians and \eqn{p} is the original dimension.
#' @param covs a \eqn{(p,p,n)} array of Gaussian covariance matrices.
#' @param target_dim an integer specifying the target lower dimension \eqn{d} (default is 2).
#' @param max_iter an integer specifying the maximum number of iterations for the optimization (default is 100).
#' @param verbose a logical flag indicating whether to print progress messages (default is TRUE).
#' 
#' @return a named list containing \describe{
#' \item{U}{the \eqn{(p,d)} projection matrix mapping original space to the lower-dimensional space.}
#' \item{proj_means}{the \eqn{(n,d)} matrix of projected Gaussian means.}
#' \item{proj_covs}{the \eqn{(d,d,n)} array of projected Gaussian covariance matrices.}
#' \item{iter_obj}{a vector of objective function values at each iteration.}
#' \item{iter_gnorm}{a vector of gradient norms at each iteration.}
#' }
#' 
#' @export
bwproj <- function(means, covs, target_dim=2, max_iter = 100, verbose=TRUE){
  # check
  if (!check_datamat(means)){
    stop("* bwproj: 'means' must be a matrix without NA/Inf.")
  }
  par_target_dim = max(1, round(target_dim))
  par_max_iter   = max(2, round(max_iter))
  par_verbose = as.logical(verbose)
  
  # run
  result = bwp_project_gaussians_rcpp(
    means, 
    covs, 
    d=par_target_dim, 
    max_iter=par_max_iter,
    verbose=par_verbose
  )
  
  # prepare
  d = par_target_dim
  p = dim(covs)[1]
  n = dim(covs)[3]
  
  output = list()
  output$U = result$U
  output$proj_means = means%*%output$U
  output$proj_covs = array(0,c(d,d,n))
  for (i in 1:n){
    output$proj_covs[,,i] = t(output$U)%*%covs[,,i]%*%output$U
  }
  output$iter_obj = result$obj
  output$iter_gnorm = result$grad_norm
  return(output)
}


# # personal example
# set.seed(123)
# 
# # Problem sizes
# K <- 3          # number of classes
# n_per <- 30     # Gaussians per class
# n <- K * n_per
# p <- 20
# d <- 2
# 
# # Class labels
# y <- rep(1:K, each = n_per)
# 
# # Means: n x p
# M <- matrix(0, n, p)
# 
# # Covariances: array dim=c(p,p,n)
# Sigma <- array(0, dim = c(p, p, n))
# 
# # Build three "prototype" covariance structures (SPD) + mean centers
# # Each class uses a different dominant subspace / conditioning pattern.
# make_spd_from_eigs <- function(eigs) {
#   Q <- qr.Q(qr(matrix(rnorm(p*p), p, p)))
#   S <- Q %*% diag(eigs) %*% t(Q)
#   (S + t(S))/2
# }
# 
# # Distinct eigenvalue profiles (controls shape/anisotropy)
# eigs1 <- c(rep(6, 3), rep(1, p-3))     # class 1: strong 3-d anisotropy
# eigs2 <- c(rep(1, 8), rep(4, 2), rep(1, p-10))  # class 2: different anisotropy
# eigs3 <- c(rep(2, p))                  # class 3: near-isotropic (but shifted mean)
# 
# S0_1 <- make_spd_from_eigs(eigs1)
# S0_2 <- make_spd_from_eigs(eigs2)
# S0_3 <- make_spd_from_eigs(eigs3)
# 
# # Mean centers (separated mostly in first few coordinates)
# sep <- 4.0   # try 3, 4, 5 ...
# 
# c1 <- sep * c(rep( 2.5, 3), rep(0, p-3))
# c2 <- sep * c(rep(-2.5, 3), rep(0, p-3))
# c3 <- sep * c(rep( 0.0, 3), rep(2.5, 3), rep(0, p-6))
# 
# # Within-class variability controls
# mean_noise <- 0.6
# cov_noise  <- 0.25   # magnitude of random SPD perturbation
# jitter     <- 0.15   # diagonal jitter for stability
# 
# for (i in 1:n) {
#   cls <- y[i]
# 
#   # Means: class center + noise
#   if (cls == 1) M[i,] <- c1 + rnorm(p, sd = mean_noise)
#   if (cls == 2) M[i,] <- c2 + rnorm(p, sd = mean_noise)
#   if (cls == 3) M[i,] <- c3 + rnorm(p, sd = mean_noise)
# 
#   # Covariances: class prototype + SPD perturbation
#   A <- matrix(rnorm(p*p), p, p)
#   P <- crossprod(A) / p  # SPD perturbation
#   if (cls == 1) S <- S0_1 + cov_noise * P
#   if (cls == 2) S <- S0_2 + cov_noise * P
#   if (cls == 3) S <- S0_3 + cov_noise * P
# 
#   Sigma[,,i] <- (S + t(S))/2 + jitter * diag(p)
# }
# 
# # run
# # Run projection
# res <- bwproj(
#   means = M,
#   covs = Sigma,
#   target_dim = d,
#   max_iter = 200,
# )
# 
# Uhat <- res$U
# 
# # Draw a 2D Gaussian confidence ellipse (base graphics)
# # mu: length-2 mean
# # S : 2x2 SPD covariance
# # level: confidence level (0.95 by default)
# draw_gaussian_ellipse <- function(mu, S, level = 0.95, npoints = 200,
#                                   border = "black", lwd = 1, lty = 1) {
#   stopifnot(length(mu) == 2, all(dim(S) == c(2,2)))
#   S <- (S + t(S))/2
# 
#   # Chi-square radius for 2D
#   r2 <- qchisq(level, df = 2)
#   r  <- sqrt(r2)
# 
#   # eigendecomp for ellipse
#   ee <- eigen(S, symmetric = TRUE)
#   vals <- pmax(ee$values, 0)
#   vecs <- ee$vectors
# 
#   theta <- seq(0, 2*pi, length.out = npoints)
#   unit  <- rbind(cos(theta), sin(theta))  # 2 x npoints
# 
#   # Transform unit circle -> ellipse
#   A <- vecs %*% diag(sqrt(vals), 2, 2)
#   pts <- sweep(A %*% unit, 1, mu, "+")    # 2 x npoints
# 
#   lines(pts[1,], pts[2,], col = border, lwd = lwd, lty = lty)
#   invisible(pts)
# }
# 
# plot_projected_gaussians <- function(M, Sigma, Uhat, y, level = 0.95,
#                                      show_centers = TRUE,
#                                      cols = c("tomato", "royalblue", "darkgreen"),
#                                      lwd = 1.5, lty = 1) {
#   stopifnot(is.matrix(M), length(dim(Sigma)) == 3)
#   n <- nrow(M)
#   p <- ncol(M)
#   stopifnot(dim(Sigma)[1] == p, dim(Sigma)[2] == p, dim(Sigma)[3] == n)
#   stopifnot(nrow(Uhat) == p, ncol(Uhat) == 2)
#   stopifnot(length(y) == n)
# 
#   # Embedded means and covariances
#   M_low <- M %*% Uhat  # n x 2
#   Sig_low <- array(0, dim = c(2,2,n))
#   for (i in 1:n) {
#     Sig_low[,,i] <- t(Uhat) %*% Sigma[,,i] %*% Uhat
#     Sig_low[,,i] <- (Sig_low[,,i] + t(Sig_low[,,i]))/2
#   }
# 
#   # Determine plot limits from ellipses (cheap bounding)
#   # Use the largest axis length for each ellipse: sqrt(qchisq)*sqrt(max eigenvalue)
#   r <- sqrt(qchisq(level, df = 2))
#   rad <- numeric(n)
#   for (i in 1:n) {
#     ev <- eigen(Sig_low[,,i], symmetric = TRUE, only.values = TRUE)$values
#     rad[i] <- r * sqrt(max(pmax(ev, 0)))
#   }
#   xlim <- range(M_low[,1] + c(-1,1) %o% rad)
#   ylim <- range(M_low[,2] + c(-1,1) %o% rad)
# 
#   # Base plot canvas
#   plot(NA, xlim = xlim, ylim = ylim, asp = 1,
#        xlab = "Projected coordinate 1", ylab = "Projected coordinate 2",
#        main = sprintf("Projected Gaussians (%d%% ellipses)", round(100*level)))
# 
#   # Draw ellipses
#   for (i in 1:n) {
#     cls <- y[i]
#     border_col <- cols[cls]
#     draw_gaussian_ellipse(mu = M_low[i,], S = Sig_low[,,i],
#                           level = level, border = border_col, lwd = lwd, lty = lty)
#     if (show_centers) {
#       points(M_low[i,1], M_low[i,2], pch = 19, cex = 0.6, col = border_col)
#     }
#   }
# 
#   legend("topright", legend = paste("Class", sort(unique(y))),
#          col = cols[sort(unique(y))], lwd = lwd, bty = "n")
# 
#   invisible(list(M_low = M_low, Sigma_low = Sig_low))
# }
# 
# # After running:
# # res <- bwproj(...)
# # Uhat <- res$U
# 
# out <- plot_projected_gaussians(
#   M = M,
#   Sigma = Sigma,
#   Uhat = Uhat,
#   y = y,
#   level = 0.95,
#   show_centers = TRUE
# )

#' Classical Multidimensional Scaling
#' 
#' Classical multidimensional scaling aims at finding low-dimensional structure 
#' by preserving pairwise distances of data. 
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' @param ndim an integer-valued target dimension.
#' 
#' @return a named list containing
#' \describe{
#' \item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
#' \item{stress}{discrepancy between embedded and origianl data as a measure of error.}
#' }
#' 
#' @examples 
#' ## use simple example of iris dataset
#' data(iris) 
#' idata = as.matrix(iris[,1:4])
#' icol  = as.factor(iris[,5])   # class information
#' 
#' ## run Classical MDS
#' iris.cmds = cmds(idata, ndim=2)
#' 
#' ## visualize
#' opar <- par(no.readonly=TRUE)
#' plot(iris.cmds$embed, col=icol, 
#'      main=paste0("STRESS=",round(iris.cmds$stress,4)))
#' par(opar)
#' 
#' @references 
#' \insertRef{torgerson_multidimensional_1952}{maotai} 
#' 
#' @export
cmds <- function(data, ndim=2){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* cmds : an input 'data' should be a matrix without any missing/infinite values.")
  }
  xdiss = stats::as.dist(cpp_pdist(data))
  
  ############################################################
  # Run and Return
  mydim  = round(ndim)
  output = hidden_cmds(xdiss, ndim=mydim)
  return(output)
}

#' von Mises-Fisher mixture model reduction - Greedy method
#' 
#' When given parameters of the von Mises-Fisher mixture model, this function 
#' aims at mixture model reduction using a greedy method.
#' 
#' @param means a \eqn{(K \times p)} matrix of means of the von Mises-Fisher components.
#' @param concentrations a \eqn{K} vector of concentrations of the von Mises-Fisher components.
#' @param weights a \eqn{K} vector of weights of the von Mises-Fisher components.
#' @param target.num a desired number of components after reduction. Default is 2.
#' 
#' @return a named list of the reduced mixture model containing \describe{
#' \item{means}{a \eqn{(\code{target.num} \times p)} matrix of means of the von Mises-Fisher components.}
#' \item{concentrations}{a \eqn{\code{target.num}} vector of concentrations of the von Mises-Fisher components.}
#' \item{weights}{a \eqn{\code{target.num}} vector of weights of the von Mises-Fisher components.}
#' }
#' @export
movMF_reduce_greedy <- function(means, concentrations, weights, target.num=2){
  ###############################################
  # Preprocessing
  # means
  if (!is.matrix(means)){
    data_means = cpp_WL_normalise(as.matrix(means))
  } else {
    data_means = cpp_WL_normalise(means)
  }
  
  # concentrations
  data_concentrations = as.vector(concentrations)
  if (length(data_concentrations)!=base::nrow(data_means)){
    stop("* movMF_reduce_greedy : cardinalities of the means and concentrations do not match.")
  }
  
  # weights
  if ((length(weights)==0)&&(is.null(weights))){
    data_weights = rep(1/length(data_concentrations), length(data_concentrations))
  } else {
    data_weights = as.vector(weights)
    data_weights = data_weights/base::sum(data_weights)
  }
  if (any(data_weights < .Machine$double.eps)||(length(data_weights)!=length(data_concentrations))){
    stop("* movMF_reduce_greedy : invalid 'weights'. Please see the documentation.")
  }
  
  # desired number
  K = length(data_weights)
  par_target_num = round(target.num)
  if ((par_target_num <= 1)||(par_target_num >= K)){
    stop("* movMF_reduce_greedy : 'target.num' must be greater than 1 and less than the number of components.")
  }
  
  
  ###############################################
  # MAIN ROUTINE
  old_vmf <- list(means=data_means, concentrations=data_concentrations, weights=data_weights)
  
  while (length(old_vmf$weights) > par_target_num){
    # compute the pairwise distances
    dists = maotai::WLpdist(old_vmf$means, old_vmf$concentrations)
      
    # find the closest pair
    min_dist = Inf
    min_i = 0
    min_j = 0
    for (i in 1:(length(old_vmf$weights)-1)){
      for (j in (i+1):length(old_vmf$weights)){
        if (dists[i,j] < min_dist){
          min_dist = dists[i,j]
          min_i = i
          min_j = j
        }
      }
    }
    
    # save the non-closest pair
    new_means = old_vmf$means[-c(min_i, min_j),]
    new_concentrations = old_vmf$concentrations[-c(min_i, min_j)]
    new_weights = old_vmf$weights[-c(min_i, min_j)]
    
    # merge the closest pairs
    tgt_means = old_vmf$means[c(min_i, min_j),]
    tgt_kappa = old_vmf$concentrations[c(min_i, min_j)]
    tgt_alpha = old_vmf$weights[c(min_i, min_j)]
    tgt_alpha_norm = tgt_alpha/base::sum(tgt_alpha)
    tgt_barycenter <- maotai::WLbarycenter(tgt_means, tgt_kappa, tgt_alpha_norm)
    
    # update the old one
    old_vmf = list(means=rbind(new_means, tgt_barycenter$mean), 
                   concentrations=c(new_concentrations, tgt_barycenter$concentration),
                   weights=c(new_weights, base::sum(tgt_alpha)))
  }
  
  # return
  return(old_vmf)
}


# # simple example
# # data matrix normalized
# X = as.matrix(iris[,2:4])
# X = as.matrix(scale(X, center=TRUE, scale=FALSE))
# X = X%*%eigen(cov(X))$vectors[,1:2] # apply PCA
# X = X/sqrt(rowSums(X^2))
# 
# # fit the model with movMF package
# big_movMF <- movMF::movMF(X, 10)
# clust_big_movMF <- predict(big_movMF, X)
# convert_movMF <- movMF_convert(big_movMF)
# 
# big_means <- convert_movMF$means
# big_weights <- convert_movMF$weights
# big_concentrations <- convert_movMF$concentrations
# 
# # reduce to 3 components using different methods
# red3 <- movMF_reduce_greedy(big_means, big_concentrations, big_weights, target.num=3)
# red5 <- movMF_reduce_greedy(big_means, big_concentrations, big_weights, target.num=5)
# 
# clust3 <- movMF_info(X, red3$means, red3$concentrations, red3$weights)$clustering
# clust5 <- movMF_info(X, red5$means, red5$concentrations, red5$weights)$clustering
# 
# # visualize
# par(mfrow=c(1,3), pty="s")
# plot(X, col=clust_big_movMF, pch=19, main="Original")
# plot(X, col=clust3, pch=19, main="Greedy K=3")
# plot(X, col=clust5, pch=19, main="Greedy K=5")

## RETICULATE : global reference
# .onLoad <- function(libname, pkgname) {
#   if (!interactive()) {
#     # Tell rgl to use an off-screen "null" device in headless checks
#     options(rgl.useNULL = TRUE)
#     Sys.setenv(RGL_USE_NULL = "TRUE")
#   }
# }
# .pkgenv <- new.env(parent = emptyenv())

.onAttach <- function(...){
  ## Retrieve Year Information
  date <- date()
  x <- regexpr("[0-9]{4}", date)
  this.year <- substr(date, x[1], x[1] + attr(x, "match.length") - 1)

  # Retrieve Current Version
  this.version = packageVersion("maotai")

  ## Print on Screen
  packageStartupMessage("**-----------------------------------------------------------------**")
  packageStartupMessage("** maotai")
  packageStartupMessage("**  - Tools for Matrix Algebra, Optimization and Inference Problems")
  packageStartupMessage("**")
  packageStartupMessage("** Version    : ",this.version,"      (",this.year,")",sep="")
  packageStartupMessage("** Maintainer : Kisung You (kisung.you@outlook.com)")
  packageStartupMessage("** Website    : https://www.kisungyou.com/maotai")
  packageStartupMessage("**")
  packageStartupMessage("** Please share any bugs or suggestions to the maintainer.")
  packageStartupMessage("**-----------------------------------------------------------------**")
}

.onUnload <- function(libpath) {
  library.dynam.unload("maotai", libpath)
}

#' Extract meaningful information from the von Mises-Fisher mixture model
#' 
#' Given a mixture of von Mises-Fisher distributions, this function computes 
#' several related quantities of the data on the unit hypersphere with respect 
#' to the specified model.
#' 
#' @param data an \eqn{(n\times d)} data matrix.
#' @param means an \eqn{(k\times d)} matrix of means. 
#' @param concentrations a vector of length \eqn{k} of concentration parameters.
#' @param weights a vector of length \eqn{k} of mixing weights.
#' 
#' @return a named list containing \describe{
#' \item{densities}{a vector of length \eqn{n} of the densities of the data points.}
#' \item{clustering}{a vector of length \eqn{n} of the hard clustering results.}
#' \item{loglkd}{the log-likelihood of the data.}
#' \item{AIC}{the Akaike information criterion.}
#' \item{BIC}{the Bayesian information criterion.}
#' \item{HQIC}{the Hannan-Quinn information criterion.}
#' }
#' 
#' @export
movMF_info <- function(data, means, concentrations, weights){
  ###############################################
  # Preprocessing
  # data
  if (!is.matrix(data)){
    my_data = cpp_WL_normalise(as.matrix(data))
  } else {
    my_data = cpp_WL_normalise(data)
  }
  
  # means
  if (!is.matrix(means)){
    my_means = cpp_WL_normalise(as.matrix(means))
  } else {
    my_means = cpp_WL_normalise(means)
  }
  
  # concentrations
  my_concentrations = as.vector(concentrations)
  if (length(my_concentrations)!=base::nrow(my_means)){
    stop("* movMF_info : cardinalities of the means and concentrations do not match.")
  }
  
  # weights
  if ((length(weights)==0)&&(is.null(weights))){
    my_weights = rep(1/length(my_concentrations), length(my_concentrations))
  } else {
    my_weights = as.vector(weights)
    my_weights = my_weights/base::sum(my_weights)
  }
  if (any(my_weights < .Machine$double.eps)||(length(my_weights)!=length(my_concentrations))){
    stop("* movMF_info : invalid 'weights'. Please see the documentation.")
  }
  
  ###############################################
  # BASIC COMPUTATION
  par_n = base::nrow(my_data)
  par_p = base::ncol(my_data)
  par_k = base::length(my_weights)
  membership <- array(0,c(par_n, par_k))
  for (k in 1:par_k){
    membership[,k] <- aux_vmf_density(my_data, my_means[k,], my_concentrations[k])*my_weights[k]
  }
  
  ###############################################
  # ADVANCED
  # density (n,) vector
  out_densities = base::rowSums(membership) 
  
  # hard clustering results
  out_cluster = rep(0, par_n)
  for (i in 1:par_n){
    tmp_clust <- as.vector(membership[i,])
    tmp_clust <- tmp_clust/base::sum(tmp_clust)
    out_cluster[i] <- which.max(tmp_clust)
  }
  
  # log-likelihood
  out_loglkd <- base::sum(base::log(out_densities))
  
  # information criteria
  df <- (par_k - 1) + par_k*((par_p-1)+1)
  out_aic <- -2*out_loglkd + 2*df # df = k for wikipedia documentation
  out_bic <- -2*out_loglkd + base::log(par_n)*df
  out_hqic <- -2*out_loglkd + base::log(base::log(par_n))*2*df
  
  ###############################################
  # RETURN
  output = list()
  output$densities = out_densities
  output$clustering = out_cluster
  output$loglkd <- out_loglkd
  output$AIC <- out_aic 
  output$BIC <- out_bic
  output$HQIC <- out_hqic
  
  return(output)
}



# auxiliary functions -----------------------------------------------------
#' @keywords internal
#' @noRd
aux_vmf_density <- function(x, mu, kappa){
  # x : (n x d) data matrix
  # mu : (d,)   mean vector
  # kappa : >0  concentration parameter
  
  # initialization
  d <- base::ncol(x)
  
  # normalizing constant
  log_c_p_kappa <- (d/2 - 1) * log(kappa) - (d/2) * log(2 * pi) - log(base::besselI(kappa, nu = d/2 - 1, expon.scaled = TRUE)) - kappa
  c_p_kappa <- exp(log_c_p_kappa)
  
  
  # iterate
  densities = rep(0, base::nrow(x))
  for (i in 1:base::nrow(x)){
    densities[i] <- base::exp(kappa*base::sum(as.vector(x[i,])*mu))*c_p_kappa
  }
  return(densities)
}

# # simple example
# # data matrix normalized
# X = as.matrix(iris[,2:4])
# X = as.matrix(scale(X, center=TRUE, scale=FALSE))
# X = X/sqrt(rowSums(X^2))
# 
# # fit the model with movMF package
# out_movMF <- movMF::movMF(X, 3)
# clust_movMF <- predict(out_movMF, X)
# 
# # use my function
# fit_weights <- out_movMF$alpha
# fit_means   <- out_movMF$theta/sqrt(rowSums(out_movMF$theta^2))
# fit_concentrations <- sqrt(rowSums(out_movMF$theta^2))
# clust_mine <- movMF_info(X, fit_means, fit_concentrations, fit_weights)$clustering

#' Compute a Rotation on the 2-dimensional Sphere
#' 
#' A vector of unit norm is an element on the hypersphere. When two unit-norm 
#' vectors \eqn{x} and \eqn{y} in 3-dimensional space are given, this function 
#' computes a rotation matrix \eqn{Q} on the 2-dimensional sphere such that
#' \deqn{y=Qx}.
#' 
#' @param x a length-\eqn{3} vector. If \eqn{\|x\|\neq 1}, normalization is applied.
#' @param y a length-\eqn{3} vector. If \eqn{\|y\|\neq 1}, normalization is applied.
#' 
#' @return a \eqn{(3\times 3)} rotation matrix.
#' 
#' @examples
#' \donttest{
#' ## generate two data points
#' #  one randomly and another on the north pole
#' x = stats::rnorm(3)
#' x = x/sqrt(sum(x^2))
#' y = c(0,0,1)
#' 
#' ## compute the rotation
#' Q = rotationS2(x,y)
#' 
#' ## compare 
#' Qx = as.vector(Q%*%x)
#' 
#' ## print
#' printmat = rbind(Qx, y)
#' rownames(printmat) = c("rotated:", "target:")
#' print(printmat)
#' }
#' 
#' @export
rotationS2 <- function(x, y){
  ############################################################
  # Preprocessing 
  vec_x = as.vector(x); vec_x = vec_x/sqrt(sum(vec_x^2))
  vec_y = as.vector(y); vec_y = vec_y/sqrt(sum(vec_y^2))

  if (length(vec_x)!=3){    stop("rotationS2 : an input 'x' is not of length 3.")  }
  if (length(vec_y)!=3){    stop("rotationS2 : an input 'y' is not of length 3.")  }
  
  ############################################################
  # Run and Return
  output = rotateS2_main(vec_x, vec_y)
  return(output)
}



# auxiliary ---------------------------------------------------------------
#' @keywords internal
#' @noRd
rotateS2_main <- function(vec_u, vec_v){
  # orthonormal vector 'n'
  vec_n = pracma::cross(vec_u, vec_v)
  vec_n = vec_n/sqrt(sum(vec_n^2))
  
  # another orthonormal vector
  vec_t = pracma::cross(vec_n, vec_u)
  vec_t = vec_t/sqrt(sum(vec_t^2))
  
  # compute the angle
  alpha = base::atan2(sum(vec_v*vec_t), sum(vec_v*vec_u))
  
  # Rn(alpha)
  Rnalpha = cbind(c(base::cos(alpha), base::sin(alpha), 0),
                  c(-base::sin(alpha), base::cos(alpha), 0),
                  c(0,0,1))
  
  # T
  mat_T = cbind(vec_u,vec_t,vec_n)
  
  # rotator
  output = mat_T%*%Rnalpha%*%base::solve(mat_T)
  return(output)
}

#' Numerical Approximation to Gradient of a Function with Matrix Argument
#' 
#' For a given function \eqn{f:\mathbf{R}^{n\times p} \rightarrow \mathbf{R}}, 
#' we use finite difference scheme that approximates a gradient at a given point \eqn{x}. 
#' In Riemannian optimization, this can be used as a proxy for 
#' ambient gradient. Use with care since it may accumulate numerical error.
#' 
#' @param fn a function that takes a matrix of size \eqn{(n\times p)} and returns a scalar value.
#' @param x an \eqn{(n\times p)} matrix where the gradient is to be computed.
#' @param h step size for centered difference scheme. 
#' 
#' @return an approximate numerical gradient matrix of size \eqn{(n\times p)}.
#' 
#' @examples 
#' ## function f(X) = <a,Xb> for two vectors 'a' and 'b'
#' #  derivative w.r.t X is ab'
#' #  take an example of (5x5) symmetric positive definite matrix
#' 
#' #  problem settings
#' a   <- rnorm(5)
#' b   <- rnorm(5)
#' ftn <- function(X){
#'   return(sum(as.vector(X%*%b)*a))
#' }       # function to be taken derivative
#' myX <- matrix(rnorm(25),nrow=5)  # point where derivative is evaluated
#' myX <- myX%*%t(myX)
#' 
#' # main computation
#' sol.true <- base::outer(a,b)
#' sol.num1 <- matderiv(ftn, myX, h=1e-1) # step size : 1e-1
#' sol.num2 <- matderiv(ftn, myX, h=1e-5) #             1e-3
#' sol.num3 <- matderiv(ftn, myX, h=1e-9) #             1e-5
#' 
#' ## visualize/print the results
#' expar = par(no.readonly=TRUE)
#' par(mfrow=c(2,2),pty="s")
#' image(sol.true, main="true solution")
#' image(sol.num1, main="h=1e-1")
#' image(sol.num2, main="h=1e-5")
#' image(sol.num3, main="h=1e-9")
#' par(expar)
#' 
#' \donttest{
#' ntrue = norm(sol.true,"f")
#' cat('* Relative Errors in Frobenius Norm ')
#' cat(paste("*  h=1e-1   : ",norm(sol.true-sol.num1,"f")/ntrue,sep=""))
#' cat(paste("*  h=1e-5   : ",norm(sol.true-sol.num2,"f")/ntrue,sep=""))
#' cat(paste("*  h=1e-9   : ",norm(sol.true-sol.num3,"f")/ntrue,sep=""))
#' }
#' 
#' @references 
#' \insertRef{kincaid_numerical_2009}{maotai}
#' 
#' @export
matderiv <- function(fn, x, h=0.001){
  if (h <= 0){
    stop("* matderiv : 'h' should be a nonnegative real number.")
  }
  hval = max(sqrt(.Machine$double.eps), h)
  return(gradF(fn,x,hval))
}
# h = 0.001
# X = matrix(rnorm(9),nrow=3)
# X = X%*%t(X)
# dX = array(0,c(3,3))
# fX = function(x){return(sum(diag(x%*%x)))}
# for (i in 1:3){
#   for (j in 1:3){
#     Xp = X
#     Xm = X
#     Xp[i,j] = Xp[i,j] + h
#     Xm[i,j] = Xm[i,j] - h
#     dX[i,j] = (fX(Xp)-fX(Xm))/(2*h)
#   }
# }
# dX

# Hidden Functions : Operations
# these functions can be loaded using 'utils::getFromNamespace'
# by the command 'getFromNamespace("function_name","maotai");
# 
# 01. hidden_pinv           : pseudo-inverse
# 02. hidden_vech           : half vectorization including the diagonal.
#     hidden_ivech            diagonal inclusion is also taken care. 
# 03. hidden_lab2ind        : create an index list from a label vector
#     hidden_ind2lab          given an index list, create a label vector
# 04. hidden_subsetid       : generate split of the subset id
# 05. hidden_geigen         : do 'geigen' operation; Decreasing order
# 06. hidden_knn
# 07. hidden_knee_clamped   : knee-point detection with clamped least squares - return idx
# 08. hidden_knn_binary     : return a sparse binary matrix for Euclidean data excluding selfl dgCMatrix


# 01. hidden_pinv ---------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_pinv <- function(A){
  return(aux_pinv(A))
}

# 02. hidden_vech & hidden_ivech ------------------------------------------
#' @keywords internal
#' @noRd
hidden_vech <- function(A, diag=TRUE){
  if ((!is.matrix(A))||(nrow(A)!=ncol(A))){
    stop("* hidden_vech : input should be a square matrix.")
  }
  mydiag = as.logical(diag)
  return(A[base::lower.tri(A, diag=mydiag)])
}
#' @keywords internal
#' @noRd
hidden_ivech <- function(a, diag=TRUE){
  k = length(a)
  if (diag){
    n = round((-1 + sqrt(1+(8*k)))/2)
    output = array(0,c(n,n))
    output[lower.tri(output, diag = TRUE)] = a
    output = output + t(output)
    diag(output) = diag(output)/2
  } else {
    n = round((1+sqrt(1+8*k))/2)
    output = array(0,c(n,n))
    output[lower.tri(output, diag = FALSE)] = a
    output = output + t(output)
  }
  return(output)
}


# 03. hidden_lab2ind & hidden_ind2lab -------------------------------------
#' @keywords internal
#' @noRd
hidden_lab2ind <- function(label){
  ulabel = base::unique(label)
  nlabel = length(ulabel)
  
  index  = list()
  for (k in 1:nlabel){
    index[[k]] = which(label==ulabel[k])
  }
  return(index)
}
#' @keywords internal
#' @noRd
hidden_ind2lab <- function(index){
  K = length(index)
  N = sum(unlist(lapply(index, length)))
  
  output = rep(0,N)
  for (k in 1:K){
    output[index[[k]]] = k
  }
  return(output)
}

# 04. hidden_subsetid -----------------------------------------------------
#' @keywords internal
#' @noRd
hidden_subsetid <- function(n, k){
  return(base::split(base::sample(n), base::sort(n%%k)))
}

# 05. hidden_geigen -------------------------------------------------------
#' It mimics the behavior of 'geigen' function with normalization added
#' @keywords internal
#' @noRd
hidden_geigen <- function(A, B, normalize=TRUE){
  n    = nrow(A)
  runs = cpp_geigen(A,B)
  
  tval = as.vector(base::Re(runs$values))[n:1]
  tvec = base::Re(runs$vectors)[,n:1]
  if (normalize){
    for (i in 1:n){
      tgt = as.vector(tvec[,i])
      tvec[,i] = tgt/sqrt(sum(tgt^2))
    }  
  }

  output = list()
  output$values  = tval
  output$vectors = tvec
  return(output)
}

# 06. hidden_knn ----------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_knn <- function(dat, nnbd=2, ...){
  nnbd = round(nnbd)
  return(RANN::nn2(dat, k=nnbd, ...))
}


# 07. hidden_knee_clamped -------------------------------------------------
#' @keywords internal
#' @noRd
hidden_knee_clamped_basic <- function(x, y){
  m = length(x)
  c = x[1]
  d = y[1]
  a = x[m]
  b = y[m]
  
  y2 = (((b-d)/(a-c))*(x-c))+d
  return(sum((y-y2)^2))
}
#' @keywords internal
#' @noRd
hidden_knee_clamped <- function(x, y){
  x = as.vector(x)
  y = as.vector(y)
  n = length(x)
  if (n < 3){
    stop("* knee point detection : length must be larger than 2.")
  }
  scores = rep(Inf, n)
  for (i in 2:(n-1)){
    x.left = x[1:i]
    y.left = y[1:i]
    
    x.right = x[i:n]
    y.right = y[i:n]
    
    term1 = hidden_knee_clamped_basic(x.left, y.left)
    term2 = hidden_knee_clamped_basic(x.right, y.right)
    scores[i] = term1+term2
  }
  return(which.min(scores)) # return the index of the minimal SSE's
}

# 08. hidden_knn_binary ---------------------------------------------------
#     excluding self and return a binary sparse adjacency matrix
#' @keywords internal
#' @noRd
hidden_knn_binary <- function(data, nnbd=1, construction=c("or","and")){
  n = base::nrow(data)
  nnbd = max(round(nnbd), 1)
  construction = match.arg(construction)
  if (all(construction=="and")){
    intersect = TRUE
  } else {
    intersect = FALSE
  }
  
  run_knn = RANN::nn2(data, k=nnbd+1)$nn.idx[,2:(nnbd+1)]-1 # -1 for C++ convention
  run_res = src_construct_by_knn(run_knn, intersect)
  return(run_res)
}

#' #' Distance Measures between Multiple Samples using Gaussian Mixture Models
#' #' 
#' #' Taking multiple observations (a sample) as a unit of analysis requires 
#' #' a measure of discrepancy between samples. \code{distgmm} fits finite 
#' #' Gaussian mixture models to each sample and use the fitted model as 
#' #' a representation of a single sample. A single model is selected via 
#' #' Bayesian Information Criterion (BIC). 
#' #' 
#' #' @param datalist a length \eqn{N} list of samples. All elements of the list should be of same type, either \code{vector} or \code{matrix} of same dimension (number of columns).
#' #' @param method name of the distance/dissimilarity measure.
#' #' @param maxk maximum number of clusters to be fitted using GMM.
#' #' @param as.dist a logical; \code{TRUE} to return \code{dist} object, \code{FALSE} to return an \eqn{(N\times N)} symmetric matrix of pairwise distances.
#' #' 
#' #' @return either \code{dist} object of an \eqn{(N\times N)} symmetric matrix of pairwise distances by \code{as.dist} argument.
#' #' 
#' #' @examples 
#' #' ## let's try two-dimensional data of 30 samples
#' #' ## single or mixture of two and three gaussian distributions.
#' #' dlist = list()
#' #' for (i in 1:10){
#' #'   dlist[[i]] = matrix(rnorm(120),ncol=2) 
#' #' }
#' #' for (i in 11:20){
#' #'   A = matrix(rnorm(60,mean=-4),ncol=2)
#' #'   B = matrix(rnorm(60,mean= 4),ncol=2)
#' #'   dlist[[i]] = rbind(A,B)
#' #' }
#' #' for (i in 21:30){
#' #'   A = matrix(rnorm(40,mean=-4),ncol=2)
#' #'   B = matrix(rnorm(40),ncol=2)
#' #'   C = matrix(rnorm(40,mean= 4),ncol=2)
#' #'   dlist[[i]] = rbind(A,B,C)
#' #' }
#' #' 
#' #' ## compute pairwise distances, expecting (3 x 3) block structure.
#' #' mm = distgmm(dlist, maxk=5)
#' #' 
#' #' ## visualize
#' #' opar <- par(no.readonly=TRUE)
#' #' par(pty="s")
#' #' image(mm[,nrow(mm):1], main="3-block pattern as expected")
#' #' par(opar)
#' #' 
#' #' @keywords internal
#' #' @noRd
#' distgmm <- function(datalist, method=c("L2"), maxk=5, as.dist=FALSE){
#'   #######################################################
#'   # Preprocessing : checkers
#'   if (!check_datalist(datalist)){
#'     stop("* distgmm : an input should be a list containing samples of same dimension.")
#'   }
#'   maxk   = round(maxk)
#'   method = match.arg(method)
#'   nlist  = length(datalist)
#'   
#'   #######################################################
#'   # Compute : GMM
#'   list.gmm = list()
#'   if (is.vector(datalist[[1]])){
#'     vec.flag = TRUE
#'     for (n in 1:nlist){
#'      list.gmm[[n]] = mclust::Mclust(datalist[[n]], G=1:maxk, modelNames="V", verbose=FALSE)$parameters
#'     }  
#'   } else {
#'     vec.flag = FALSE
#'     for (n in 1:nlist){
#'       list.gmm[[n]] = mclust::Mclust(datalist[[n]], G=1:maxk, verbose=FALSE, modelNames="VVV")$parameters
#'     }
#'   }
#'   
#'   #######################################################
#'   # Compute : Pairwise Distance
#'   output = array(0,c(nlist,nlist))
#'   for (i in 1:(nlist-1)){
#'     objA = list.gmm[[i]]
#'     for (j in (i+1):nlist){
#'       objB = list.gmm[[j]]
#'       if (vec.flag){
#'         theval = switch(method, 
#'                         "L2" = distgmm_l2_1d(objA, objB))
#'         output[i,j] <- output[j,i] <- theval
#'       } else {
#'         theval = switch(method,
#'                         "L2" = distgmm_l2_nd(objA, objB))
#'         output[i,j] <- output[j,i] <- theval
#'       }
#'     }
#'   }
#'   
#'   #######################################################
#'   if (as.dist){
#'     return(stats::as.dist(output))
#'   } else {
#'     return(output)  
#'   }
#' }
#' 
#' 
#' # use Mclust 'parameters' object ------------------------------------------
#' #' @keywords internal
#' #' @noRd
#' distgmm_l2_1d <- function(objA, objB){
#'   weightA = as.vector(objA$pro)
#'   muA     = matrix(objA$mean, ncol=1)
#'   covA    = array(0,c(1,1,length(weightA)))
#'   for (i in 1:length(weightA)){
#'     covA[,,i] = objA$variance$sigmasq[i]
#'   }
#'   weightB = as.vector(objB$pro)
#'   muB     = matrix(objB$mean, ncol=1)
#'   covB    = array(0,c(1,1,length(weightB)))
#'   for (i in 1:length(weightB)){
#'     covB[,,i] = objB$variance$sigmasq[i]
#'   }
#'   
#'   ## run CPP (same for both 1d and nd cases)
#'   cpp.res = cpp_pairwise_L2(muA, muB, covA, covB)
#'   A = cpp.res$A
#'   B = cpp.res$B
#'   C = cpp.res$C
#'   
#'   ## matrix multiplication
#'   term1 = base::sum(as.vector(A%*%weightA)*weightA)
#'   term2 = base::sum(as.vector(B%*%weightB)*weightB)
#'   term3 = -2*base::sum(as.vector(C%*%weightB)*weightA)
#'   
#'   ## return distance/ L2 needs to be taken square root.
#'   return(base::sqrt(term1+term2+term3))
#' }
#' #' @keywords internal
#' #' @noRd
#' distgmm_l2_nd <- function(objA, objB){
#'   weightA = as.vector(objA$pro)
#'   muA     = t(objA$mean)
#'   covA    = objA$variance$sigma
#'   
#'   weightB = as.vector(objB$pro)
#'   muB     = t(objB$mean)
#'   covB    = objB$variance$sigma
#'   
#'   if (length(dim(covA)) < 3){
#'     tmpA = covA
#'     covA = array(0,c(ncol(muA),ncol(muA),1))
#'     covA[,,1] = as.matrix(tmpA)
#'   }
#'   if (length(dim(covB)) < 3){
#'     tmpB = covB
#'     covB = array(0,c(ncol(muB),ncol(muB),1))
#'     covB[,,1] = as.matrix(tmpB)
#'   }
#'   
#'   ## run CPP (same for both 1d and nd cases)
#'   cpp.res = cpp_pairwise_L2(muA, muB, covA, covB)
#'   A = cpp.res$A
#'   B = cpp.res$B
#'   C = cpp.res$C
#'   
#'   ## matrix multiplication
#'   term1 = base::sum(as.vector(A%*%weightA)*weightA)
#'   term2 = base::sum(as.vector(B%*%weightB)*weightB)
#'   term3 = -2*base::sum(as.vector(C%*%weightB)*weightA)
#'   
#'   ## return distance/ L2 needs to be taken square root.
#'   return(base::sqrt(term1+term2+term3))
#' }
#' 


# # personal experiment -----------------------------------------------------
# x = list()
# for (i in 1:20){
#   x[[i]] = matrix(rnorm(300*2),ncol=2)
# }
# for (i in 21:40){
#   x[[i]] = rbind(matrix(rnorm(150*2,mean=-4),ncol=2), matrix(rnorm(150*2,mean=4),ncol=2))
# }
# for (i in 41:60){
#   x[[i]] = rbind(matrix(rnorm(100*2,mean=-4),ncol=2), matrix(rnorm(100*2),ncol=2), matrix(rnorm(150*2,mean=4),ncol=2))
# }
# mm = distgmm(x, maxk=10)
# image(mm[,nrow(mm):1])

# bestgmm <- function(dat){
#   # belows are all automatically implemented in mclustBIC
#   # # run mclustBIC
#   # opt_gmm <- (mclust::mclustBIC(dat, G=1:9, verbose = FALSE))
#   # colgmm  <- colnames(opt_gmm)
#   # rowgmm  <- 1:9
#   # 
#   # # extract mclustBIC information
#   # mm      <- matrix(opt_gmm, nrow=nrow(opt_gmm))
#   # mm[is.na(mm)] = -Inf
#   # idmax   <- as.integer(which(mm == max(mm), arr.ind = TRUE)) # show the
#   # 
#   # nclust  <- rowgmm[idmax[1]]
#   # vartype <- colgmm[idmax[2]]
#   
#   # run Mclust
#   runobj <- mclust::Mclust(dat, G=1:10, verbose=FALSE)
#   
#   # run GMM with prespecified results
#   output  = list()
#   output$weight = runobj$parameters$pro
#   output$mu     = t(runobj$parameters$mean)
#   output$cov    = runobj$parameters$variance
#   return(output)
# }

#' DP-means Algorithm for Clustering Euclidean Data
#' 
#' DP-means is a nonparametric clustering method motivated by DP mixture model in that 
#' the number of clusters is determined by a parameter \eqn{\lambda}. The larger 
#' the \eqn{\lambda} value is, the smaller the number of clusters is attained. 
#' In addition to the original paper, we added an option to randomly permute 
#' an order of updating for each observation's membership as a common 
#' heuristic in the literature of cluster analysis. 
#' 
#' @param data an \eqn{(n\times p)} data matrix for each row being an observation.
#' @param lambda a threshold to define a new cluster.
#' @param maxiter maximum number of iterations.
#' @param abstol stopping criterion 
#' @param permute.order a logical; \code{TRUE} if random order for permutation is used, \code{FALSE} otherwise.
#' 
#' @return a named list containing
#' \describe{
#' \item{cluster}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
#' \item{centers}{a list containing information for out-of-sample prediction.}
#' }
#' 
#' @examples
#' ## define data matrix of two clusters
#' x1  = matrix(rnorm(50*3,mean= 2), ncol=3)
#' x2  = matrix(rnorm(50*3,mean=-2), ncol=3)
#' X   = rbind(x1,x2)
#' lab = c(rep(1,50),rep(2,50))
#' 
#' ## run dpmeans with several lambda values
#' solA <- dpmeans(X, lambda= 5)$cluster
#' solB <- dpmeans(X, lambda=10)$cluster
#' solC <- dpmeans(X, lambda=20)$cluster
#' 
#' ## visualize the results
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,4), pty="s")
#' plot(X,col=lab,  pch=19, cex=.8, main="True", xlab="x", ylab="y")
#' plot(X,col=solA, pch=19, cex=.8, main="dpmeans lbd=5", xlab="x", ylab="y")
#' plot(X,col=solB, pch=19, cex=.8, main="dpmeans lbd=10", xlab="x", ylab="y")
#' plot(X,col=solC, pch=19, cex=.8, main="dpmeans lbd=20", xlab="x", ylab="y")
#' par(opar)
#' 
#' \donttest{
#' ## let's find variations by permuting orders of update
#' ## used setting : lambda=20, we will 8 runs
#' sol8 <- list()
#' for (i in 1:8){
#'   sol8[[i]] = dpmeans(X, lambda=20, permute.order=TRUE)$cluster
#' }
#' 
#' ## let's visualize
#' vpar <- par(no.readonly=TRUE)
#' par(mfrow=c(2,4), pty="s")
#' for (i in 1:8){
#'   pm = paste("permute no.",i,sep="")
#'   plot(X,col=sol8[[i]], pch=19, cex=.8, main=pm, xlab="x", ylab="y")
#' }
#' par(vpar)
#' }
#' 
#' @references 
#' \insertRef{kulis_revisiting_2012}{maotai}
#' 
#' @export
dpmeans <- function(data, lambda=1, maxiter=1234, abstol=1e-6, permute.order=FALSE){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* dpmeans : an input 'data' should be a matrix without any missing/infinite values.")
  }
  # Parameter and Initialization
  n = nrow(data)
  p = ncol(data)
  k = 1                                   # set k=1
  labels = rep(1,n)                       # labels={1,2,...,n}
  mu     = matrix(colMeans(data), nrow=1) # global mean
  lambda = as.double(lambda)
  
  ############################################################
  # Main Iteration
  ss.old = compute.ss(data, labels, mu)+ k*lambda
  ss.new = 0
  for (iter in 1:maxiter){
    # 0. updating order of observations
    if (permute.order){
      idseq = sample(1:n)
    } else {
      idseq = 1:n
    }
    # 1. update the class membership per each class
    for (i in idseq){
      # 1-1. compute distances to the centers
      # dic = rep(0, k); for (j in 1:k){dic[j] = sum((as.vector(data[i,])-as.vector(mu[j,]))^2)}
      dic = as.vector(dat2centers(data[i,], mu)); # cpp conversion
      # 1-2. assign new or stay
      if (min(dic) > lambda){
        k = k+1
        labels[i] = k
        mu = rbind(mu, data[i,])
      } else {
        idmins = which(dic==min(dic))
        if (length(idmins)>1){
          labels[i] = sample(idmins, 1)
        } else {
          labels[i] = idmins 
        }
      }
    }
    
    # 2. rearrange the label (remove empty ones)
    labels = as.factor(labels)
    ulabel = sort(unique(labels))
    labnew = rep(0,n)
    for (i in 1:length(ulabel)){
      labnew[(labels==ulabel[i])] = i
    }
    labels = labnew
    k      = round(max(labels))
    
    # 3. compute per-class means
    uassign = sort(unique(labels))
    mu = array(0,c(k,p))
    for (i in 1:k){
      idmean = which(labels==uassign[i])
      if (length(idmean)==1){
        mu[i,] = as.vector(data[idmean,])
      } else {
        mu[i,] = as.vector(colMeans(data[idmean,]))
      }
    }
    
    # 4. compute DPMEANS objective function
    ss.new   = compute.ss(data, labels, mu) + k*lambda
    ss.delta = ss.old-ss.new
    ss.old   = ss.new
    
    # 5. stop if updating is not significant
    if (ss.delta < abstol){
      break
    }
  }
  
  ############################################################
  # Return the results
  output = list()
  output$cluster = as.factor(labels)
  output$centers = mu
  return(output)
}

# auxiliary functions -----------------------------------------------------
#' @keywords internal
#' @noRd
compute.ss <- function(data, label, centers){
  p = ncol(data)
  if (is.vector(centers)){
    centers = matrix(centers, nrow=1)
  }
  ulabel = sort(unique(label))
  output = 0
  for (i in 1:length(ulabel)){
    subdata = data[(label==ulabel[i]),]
    if (!is.vector(subdata)){
      nn = nrow(subdata)
      for (j in 1:nn){
        output = output + sum((as.vector(subdata[j,])-as.vector(centers[i,]))^2)
      }
    }
  }
  return(output)
}

#' Negative Eigenvalue Magnitude
#' 
#' Negative Eigenvalue Magnitude (NEM) is a measure of distortion for the data 
#' whether they are lying in Euclidean manner or not. When the value is exactly 0, it means 
#' the data is Euclidean. On the other hand, when NEM is far away from 0, it means not Euclidean. 
#' The concept of NEM is closely related to the definiteness of a Gram matrix. 
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' 
#' @return a nonnegative NEM value.
#' 
#' @examples 
#' ## use simple example of iris dataset 
#' data(iris) 
#' mydat = as.matrix(iris[,1:4])
#' 
#' ## calculate NEM
#' nem(mydat)
#' 
#' @references 
#' \insertRef{pekalska_noneuclidean_2006}{maotai}
#' 
#' @export
nem <- function(data){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* nem : an input 'data' should be a matrix without any missing/infinite values.")
  }
  xdiss = stats::as.dist(cpp_pdist(data))
  
  ############################################################
  # Compute and Return
  output = hidden_nem(xdiss)
  return(output)
}

#' CAML: Frechet mean 
#' 
#' 
#' 
#' 
#' 
#' @keywords internal
#' @noRd
caml_mean <- function(list_samples, maxiter=100, abstol=1e-6){
  # check the input
  if (!check_datalist(list_samples)){
    stop("* caml_mean: 'list_samples' must have the same-sized matrices.")
  }
  par_maxiter = max(5, round(maxiter))
  par_abstol = max(1e-12, as.double(abstol))
  
  # weights
  N = length(list_samples)
  w = rep(1/N, N)
  
  # compute
  cpp_run <- caml_frechet_mean(
    Y_list_R = list_samples, # uncentered version is fine
    w = w,
    maxit = par_maxiter, 
    tol      = par_abstol,
    step0    = 1.0,
    c1       = 1e-4,
    beta     = 0.5,
    verbose  = TRUE
  )
 
  return(cpp_run)
}

# set.seed(42)
# 
# 
# # "True" latent positions (unknown in real scenarios)
# 
# mlbench_true = mlbench::mlbench.smiley()
# Y_true  <- as.matrix(scale(mlbench_true$x, center=TRUE, scale=FALSE))
# Y_class <- mlbench_true$classes
# 
# n <- nrow(Y_true)
# k <- ncol(Y_true)
# S <- 200  # number of MCMC samples
# 
# 
# # Generate MCMC-like samples: rotated + noisy
# list_samples <- vector("list", S)
# 
# for (s in 1:S) {
#   Qs <- qr.Q(qr(matrix(rnorm(k*k), k, k)))  # random orthogonal
#   noise <- 0.1 * matrix(rnorm(n*k), n, k)
#   list_samples[[s]] <- Y_true %*% Qs + noise
# }
# 
# # Compute the CAML Fréchet mean
# res <- caml_mean(list_samples, maxiter = 200, abstol = 1e-8)
# 
# Y_hat  <- res$Y
# X_hat  <- res$YYT
# value  <- res$cost
# 
# # plot
# par(mfrow=c(1,3), pty="s")
# plot(Y_true, col=Y_class, main="True shape")
# plot(Y_hat, col=Y_class, main="Frechet mean")
# image(X_hat, main="Mean Gram matrix")
# 
# 
# # caml stands for
# # Coherent Analysis of MCMC samples from LPM models
# # present this as a framework

#' t-SNE Embedding
#' 
#' This function is a simple wrapper of \code{\link[Rtsne]{Rtsne}} function for 
#' t-Stochastic Neighbor Embedding for finding low-dimensional structure of 
#' the data embedded in the high-dimensional space. 
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' @param ndim an integer-valued target dimension.
#' @param ... extra parameters to be used in \code{\link[Rtsne]{Rtsne}} function.
#' 
#' @return a named list containing
#' \describe{
#' \item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
#' \item{stress}{discrepancy between embedded and origianl data as a measure of error.}
#' }
#' 
#' @examples
#' \donttest{
#' ## use simple example of iris dataset 
#' data(iris) 
#' mydat = as.matrix(iris[,1:4])
#' mylab = as.factor(iris[,5])
#' 
#' ## run t-SNE and MDS for comparison
#' iris.cmds = cmds(mydat, ndim=2)
#' iris.tsne = tsne(mydat, ndim=2)
#' 
#' ## extract coordinates and class information
#' cx = iris.cmds$embed # embedded coordinates of CMDS
#' tx = iris.tsne$embed #                         t-SNE
#' 
#' ## visualize
#' #  main title
#' mc = paste("CMDS with STRESS=",round(iris.cmds$stress,4),sep="")
#' mt = paste("tSNE with STRESS=",round(iris.tsne$stress,4),sep="")
#' 
#' #  draw a figure
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,2))
#' plot(cx, col=mylab, pch=19, main=mc)
#' plot(tx, col=mylab, pch=19, main=mt)
#' par(opar)
#' }
#' 
#' @export
tsne <- function(data, ndim=2, ...){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* tsne : an input 'data' should be a matrix without any missing/infinite values.")
  }
  dx = (stats::as.dist(cpp_pdist(data)))
  kk = round(ndim)
  if ((length(kk)>1)||(kk<1)||(kk>=ncol(data))){
    stop("* tsne : 'ndim' should be an integer in [1,col(data)). ")
  }
  
  ############################################################
  # Run and Return
  output = hidden_tsne(dx, ndim=kk, ...)
  return(output)
}

test_that("tSNE", {
  myn = 10
  myp = 3
  
  prep_mat = rbind(matrix(rnorm(myn*myp, mean=-2), ncol=myp),
                   matrix(rnorm(myn*myp, mean= 2), ncol=myp))
  run_tsne = maotai::tsne(prep_mat, ndim=2, perplexity=5)$embed

  expect_equal(base::nrow(run_tsne), 2*myn)
  expect_equal(base::ncol(run_tsne), 2)
})

#' Solve Sylvester Equation
#' 
#' The Sylvester equation is of form
#' \deqn{AX + XB = C}
#' where \eqn{X} is the unknown and others are given. Though it's possible to have non-square \eqn{A} and \eqn{B} matrices, 
#' we currently support square matrices only. This is a wrapper of \code{armadillo}'s \code{sylvester} function.
#' 
#' @param A a \eqn{(p\times p)} matrix as above.
#' @param B a \eqn{(p\times p)} matrix as above.
#' @param C a \eqn{(p\times p)} matrix as above.
#' 
#' @return a solution matrix \eqn{X} of size \eqn{(p\times p)}.
#' 
#' @examples 
#' ## simulated example
#' #  generate square matrices
#' A = matrix(rnorm(25),nrow=5)
#' X = matrix(rnorm(25),nrow=5)
#' B = matrix(rnorm(25),nrow=5)
#' C = A%*%X + X%*%B
#' 
#' #  solve using 'sylvester' function
#' solX = sylvester(A,B,C)
#' pm1 = "* Experiment with Sylvester Solver"
#' pm2 = paste("* Absolute Error  : ",norm(solX-X,"f"),sep="")
#' pm3 = paste("* Relative Error  : ",norm(solX-X,"f")/norm(X,"f"),sep="")
#' cat(paste(pm1,"\n",pm2,"\n",pm3,sep=""))
#' 
#' 
#' @references 
#' \insertRef{sanderson_armadillo_2016}{maotai}
#' 
#' \insertRef{eddelbuettel_rcpparmadillo_2014}{maotai}
#' 
#' @export
sylvester <- function(A,B,C){
  ###################################################################
  # check square matrix
  if (!check_sqmat(A)){
    stop("* sylvester : an input 'A' should be a square matrix.")
  }
  if (!check_sqmat(B)){
    stop("* sylvester : an input 'B' should be a square matrix.")
  }
  if (!check_sqmat(C)){
    stop("* sylvester : an input 'C' should be a square matrix.")
  }
  
  ###################################################################
  # arrange and solve
  return(cpp_sylvester(A,B,-C))
}

#' Bayesian Multidimensional Scaling
#' 
#' A Bayesian formulation of classical Multidimensional Scaling is presented.
#' Even though this method is based on MCMC sampling, we only return maximum a posterior (MAP) estimate
#' that maximizes the posterior distribution. Due to its nature without any special tuning,
#' increasing \code{mc.iter} requires much computation.
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' @param ndim an integer-valued target dimension.
#' @param par.a hyperparameter for conjugate prior on variance term, i.e., \eqn{\sigma^2 \sim IG(a,b)}. Note that \eqn{b} is chosen appropriately as in paper.
#' @param par.alpha hyperparameter for conjugate prior on diagonal term, i.e., \eqn{\lambda_j \sim IG(\alpha, \beta_j)}. Note that \eqn{\beta_j} is chosen appropriately as in paper.
#' @param par.step stepsize for random-walk, which is standard deviation of Gaussian proposal.
#' @param mc.iter the number of MCMC iterations.
#' @param verbose a logical; \code{TRUE} to show iterations, \code{FALSE} otherwise.
#' 
#' @return a named list containing
#' \describe{
#' \item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
#' \item{stress}{discrepancy between embedded and origianl data as a measure of error.}
#' }
#' 
#' @examples
#' \donttest{
#' ## use simple example of iris dataset
#' data(iris) 
#' idata = as.matrix(iris[,1:4])
#' 
#' ## run Bayesian MDS
#' #  let's run 10 iterations only.
#' iris.cmds = cmds(idata, ndim=2)
#' iris.bmds = bmds(idata, ndim=2, mc.iter=5, par.step=(2.38^2)) 
#' 
#' ## extract coordinates and class information
#' cx = iris.cmds$embed # embedded coordinates of CMDS
#' bx = iris.bmds$embed #                         BMDS
#' icol = iris[,5]      # class information
#' 
#' ## visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(2,1))
#' mc = paste0("CMDS with STRESS=",round(iris.cmds$stress,4))
#' mb = paste0("BMDS with STRESS=",round(iris.bmds$stress,4))
#' plot(cx, col=icol,pch=19,main=mc)
#' plot(bx, col=icol,pch=19,main=mb)
#' par(opar)
#' }
#' 
#' @references 
#' \insertRef{oh_bayesian_2001a}{maotai}
#' 
#' @export
bmds <- function(data, ndim=2, par.a=5, par.alpha=0.5, par.step=1, mc.iter=8128, verbose=TRUE){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* bmds : an input 'data' should be a matrix without any missing/infinite values.")
  }
  xdiss = stats::as.dist(cpp_pdist(data))
  
  ############################################################
  # Run the Hidden Function
  mydim   = round(ndim)
  mya     = as.double(par.a)
  myalpha = as.double(par.alpha)
  mystep  = as.double(par.step)
  myiter  = round(mc.iter)
  myshow  = as.logical(verbose)
  
  output = hidden_bmds(xdiss, ndim=mydim, 
                       par.a=mya, par.alpha=myalpha, 
                       par.step=mystep, mc.iter=myiter, verbose=myshow)
  
  ############################################################
  # Return the output
  return(output)
}

library(maotai)

test_that("Bayesian MDS", {
  myn = 10
  myp = 3
  
  prep_mat = rbind(matrix(rnorm(myn*myp, mean=-2), ncol=myp),
                   matrix(rnorm(myn*myp, mean= 2), ncol=myp))
  
  iris2d = maotai::bmds(prep_mat, ndim=2, mc.iter=2, par.step=4, verbose=FALSE)
  n_rows = base::nrow(iris2d$embed)
  n_cols = base::ncol(iris2d$embed)
  
  expect_equal(n_rows, base::nrow(prep_mat))
  expect_equal(n_cols, 2)
})

#' Cayley-Menger Determinant
#' 
#' Cayley-Menger determinant is a formula of a \eqn{n}-dimensional simplex 
#' with respect to the squares of all pairwise distances of its vertices.
#' 
#' @param data an \eqn{(n\times p)} matrix of row-stacked observations.
#' @return a list containing\describe{
#' \item{det}{determinant value.}
#' \item{vol}{volume attained from the determinant.}
#' }
#' 
#' @examples 
#' ## USE 'IRIS' DATASET
#' data(iris)
#' X = as.matrix(iris[,1:4])
#' 
#' ## COMPUTE CAYLEY-MENGER DETERMINANT
#' #  since k=4 < n=149, it should be zero.
#' cayleymenger(X)
#' 
#' @export
cayleymenger <- function(data){
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* cayleymenger : an input 'data' should be a matrix without any missing/infinite values.")
  }

  # compute pairwise distance
  Dtmp = stats::as.dist(cpp_pdist(data))
  
  # compute and return
  return(cayleymenger_internal(Dtmp))
}

#' @keywords internal
#' @noRd
cayleymenger_internal <- function(distobj){
  Dold = (as.matrix(distobj)^2)
  n    = base::nrow(Dold)
  Dnew = rbind(cbind(Dold, rep(1,n)), c(rep(1,n),0))
  
  val.det = base::det(Dnew)
  n       = n+1
  val.vol = base::sqrt(base::exp(base::log(((-1)^(n+1))*val.det) - n*log(2) - (2*base::lfactorial(n))))
  
  output = list(det=val.det, vol=val.vol)
  return(output)
}

# CHECKERS ----------------------------------------------------------------
# 01. check_sqmat    : if a square matrix
# 02. check_symm     : if a square, symmetric matrix
# 03. check_datalist : if a list of same-dimensional data
# 04. check_datamat  : if a matrix without weird values
# 05. check_veclist  : if a list of numeric vectors


# 05. check_veclist -------------------------------------------------------
#' @keywords internal
#' @noRd
check_veclist <- function(input_list){
  # check all are vectors
  cond1 = unlist(lapply(input_list, is.vector))
  if (any(isFALSE(cond1))){
    return(FALSE)
  }
  
  # check all are numeric
  cond2 = unlist(lapply(input_list, is.numeric))
  if (any(isFALSE(cond2))){
    return(FALSE)
  }
  
  return(TRUE)
}



# 01. check_sqmat ---------------------------------------------------------
#' @keywords internal
#' @noRd
check_sqmat <- function(x){
  cond1 = is.matrix(x)
  cond2 = (nrow(x)==ncol(x))
  cond3 = (!(any(is.infinite(x))||any(is.null(x))))
  if (cond1&&cond2&&cond3){
    return(TRUE)
  } else {
    return(FALSE)
  }
}


# 02. check_symm ----------------------------------------------------------
#' @keywords internal
#' @noRd
check_symm <- function(x){
  cond1 = check_sqmat(x)
  cond2 = isSymmetric(x)
  if (cond1&&cond2){
    return(TRUE)
  } else {
    return(FALSE)
  }
}

# 03. check_datalist ------------------------------------------------------
#' @keywords internal
#' @noRd
check_datalist <- function(dlist){
  cond1 = (is.list(dlist))
  if (is.vector(dlist[[1]])){
    cond2 = all(unlist(lapply(dlist, is.vector))==TRUE)
    cond3 = (unlist(lapply(dlist, check_datavec))==TRUE)
    if (cond1&&cond2&&cond3){
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else {
    cond2 = all(unlist(lapply(dlist, is.matrix))==TRUE)
    cond3 = (length(unique(unlist(lapply(dlist, ncol))))==1)
    cond4 = all(unlist(lapply(dlist, check_datamat))==TRUE)
    if (cond1&&cond2&&cond3&&cond4){
      return(TRUE)
    } else {
      return(FALSE)
    }
  }
}


# 04. check_datamat -------------------------------------------------------
#' @keywords internal
#' @noRd
check_datamat <- function(data){
  cond1 = (is.matrix(data))
  cond2 = all(!is.na(data))
  cond3 = all(!is.infinite(data))
  if (cond1&&cond2&&cond3){
    return(TRUE)
  } else {
    return(FALSE)
  }
}
#' @keywords internal
#' @noRd
check_datavec <- function(data){
  cond1 = (is.vector(data))
  cond2 = all(!is.na(data))
  cond3 = all(!is.infinite(data))
  if (cond1&&cond2&&cond3){
    return(TRUE)
  } else {
    return(FALSE)
  }
}

#' von Mises-Fisher mixture model reduction - partitional method
#' 
#' When given parameters of the von Mises-Fisher mixture model, this function 
#' aims at mixture model reduction using a partitional method.
#' 
#' @param means a \eqn{(K \times p)} matrix of means of the von Mises-Fisher components.
#' @param concentrations a \eqn{K} vector of concentrations of the von Mises-Fisher components.
#' @param weights a \eqn{K} vector of weights of the von Mises-Fisher components.
#' @param target.num a desired number of components after reduction. Default is 2.
#' @param method a clustering method to be used. Default is "hclust".
#' 
#' @return a named list of the reduced mixture model containing \describe{
#' \item{means}{a \eqn{(\code{target.num} \times p)} matrix of means of the von Mises-Fisher components.}
#' \item{concentrations}{a \eqn{\code{target.num}} vector of concentrations of the von Mises-Fisher components.}
#' \item{weights}{a \eqn{\code{target.num}} vector of weights of the von Mises-Fisher components.}
#' }
#' 
#' @export
movMF_reduce_partitional <- function(means, concentrations, weights, target.num=2, method=c("hclust","kmedoids")){
  ###############################################
  # Preprocessing
  # means
  if (!is.matrix(means)){
    data_means = cpp_WL_normalise(as.matrix(means))
  } else {
    data_means = cpp_WL_normalise(means)
  }
  d = base::ncol(data_means)
  
  # concentrations
  data_concentrations = as.vector(concentrations)
  if (length(data_concentrations)!=base::nrow(data_means)){
    stop("* reduce_movMF_greedy : cardinalities of the means and concentrations do not match.")
  }
  
  # weights
  if ((length(weights)==0)&&(is.null(weights))){
    data_weights = rep(1/length(data_concentrations), length(data_concentrations))
  } else {
    data_weights = as.vector(weights)
    data_weights = data_weights/base::sum(data_weights)
  }
  if (any(data_weights < .Machine$double.eps)||(length(data_weights)!=length(data_concentrations))){
    stop("* movMF_reduce_partitional : invalid 'weights'. Please see the documentation.")
  }
  
  # desired number
  K = length(data_weights)
  par_target_num = round(target.num)
  if ((par_target_num <= 1)||(par_target_num >= K)){
    stop("* movMF_reduce_partitional : 'target.num' must be greater than 1 and less than the number of components.")
  }
  
  ###############################################
  # MAIN ROUTINE
  # check the clustering method
  par_clustering = match.arg(method)
  
  # pairwise distance computation
  pdist_mat = array(0,c(K,K))
  for (i in 1:(K-1)){
    for (j in (i+1):K){
      term1 = sphere_dist(as.vector(data_means[i,]), as.vector(data_means[j,]))
      term2sq = (d-1)*((1/sqrt(data_concentrations[i]) - 1/sqrt(data_concentrations[j]))^2)
      pdist_mat[i,j] <- pdist_mat[j,i] <- sqrt((term1*term1) + term2sq)
    }
  }
  
  
  # clustering
  if (par_clustering=="hclust"){
    # old approach: directly apply clustering
    pdist_obj <- stats::as.dist(pdist_mat)
    clust_obj <- fastcluster::hclust(pdist_obj, method="single")
    obtained_clust <- stats::cutree(clust_obj, k=par_target_num)
    
    # # apply weighting scheme
    # weighted_dist_mat <- pdist_mat*sqrt(outer(data_weights, data_weights, "*"))
    # weighted_dist <- stats::as.dist(weighted_dist_mat)
    # # cut the tree
    # clust_obj <- cluster::agnes(weighted_dist, method="single")
    # obtained_clust <- stats::cutree(clust_obj, k=par_target_num)
  } else {
    # old approach: directly apply clustering
    pdist_obj <- stats::as.dist(pdist_mat)
    clust_obj <- cluster::pam(pdist_obj, k=par_target_num)
    obtained_clust <- clust_obj$clustering
    
    # # apply weighting scheme
    # weighted_dist_mat <- pdist_mat*outer(data_weights, data_weights)
    # clust_obj <- cluster::pam(weighted_dist_mat, 
    #                           k=par_target_num,
    #                           diss = TRUE)
    # obtained_clust <- clust_obj$clustering
  }
  
  # get the indices
  list_indices = vector("list", length=par_target_num)
  for (i in 1:par_target_num){
    list_indices[[i]] = which(obtained_clust==i)
  }
  
  # for each indices, compute the barycenter
  new_means = array(0,c(par_target_num, base::ncol(data_means)))
  new_concentrations = rep(0, par_target_num)
  new_weights = rep(0, par_target_num)
  
  for (it in 1:par_target_num){
    # current indices
    now_index = list_indices[[it]]
    
    # branching : single vs multiple observations in a cluster
    if (length(now_index)==1){
      new_means[it,] = data_means[now_index,]
      new_concentrations[it] = data_concentrations[now_index]
      new_weights[it] = data_weights[now_index]
    } else {
      tmp_means = data_means[now_index,]
      tmp_concentration = data_concentrations[now_index]
      tmp_weights = data_weights[now_index]
      new_weights[it] = base::sum(tmp_weights) # aggregated weights
      tmp_weights = tmp_weights/base::sum(tmp_weights)
      
      # compute the barycenter
      tmp_bary <- WLbarycenter(tmp_means, tmp_concentration, tmp_weights)
      new_means[it,] <- tmp_bary$mean
      new_concentrations[it] <- tmp_bary$concentration
    }
  }
  
  # return the output
  output = list(means=new_means,
                concentrations=new_concentrations,
                weights=new_weights)
  return(output)
}

# # simple example
# # data matrix normalized
# X = as.matrix(iris[,2:4])
# X = as.matrix(scale(X, center=TRUE, scale=FALSE))
# X = X%*%eigen(cov(X))$vectors[,1:2] # apply PCA
# X = X/sqrt(rowSums(X^2))
# 
# # fit the model with movMF package
# big_movMF <- movMF::movMF(X, 10)
# clust_big_movMF <- predict(big_movMF, X)
# convert_movMF <- movMF_convert(big_movMF)
# 
# big_means <- convert_movMF$means
# big_weights <- convert_movMF$weights
# big_concentrations <- convert_movMF$concentrations
# 
# # reduce to 3 components using different methods
# red_hclust <- movMF_reduce_partitional(big_means, big_concentrations, big_weights, target.num=3, method="hclust")
# red_medoid <- movMF_reduce_partitional(big_means, big_concentrations, big_weights, target.num=3, method="kmedoids")
# 
# clust_hclust <- movMF_info(X, red_hclust$means, red_hclust$concentrations, red_hclust$weights)$clustering
# clust_medoid <- movMF_info(X, red_medoid$means, red_medoid$concentrations, red_medoid$weights)$clustering
# 
# # visualize
# par(mfrow=c(1,3), pty="s")
# plot(X, col=clust_big_movMF, pch=19, main="Original")
# plot(X, col=clust_hclust, pch=19, main="Reduced (hclust)")
# plot(X, col=clust_medoid, pch=19, main="Reduced (kmedoids)")

#' EP-means Algorithm for Clustering Empirical Distributions
#' 
#' EP-means is a variant of k-means algorithm adapted to cluster 
#' multiple empirical cumulative distribution functions under metric structure 
#' induced by Earth Mover's Distance.
#' 
#' @param elist a length \eqn{N} list of either vector or \code{ecdf} objects.
#' @param k the number of clusters.
#' 
#' @return a named list containing \describe{
#' \item{cluster}{an integer vector indicating the cluster to which each \code{ecdf} is allocated.}
#' \item{centers}{a length \eqn{k} list of centroid \code{ecdf} objects.}
#' }
#' 
#' @examples
#' \donttest{
#' ## two sets of 1d samples, 10 each and add some noise
#' #    set 1 : mixture of two gaussians
#' #    set 2 : single gamma distribution
#' 
#' # generate data
#' elist = list()
#' for (i in 1:10){
#'    elist[[i]] = stats::ecdf(c(rnorm(100, mean=-2), rnorm(50, mean=2)))
#' }
#' for (j in 11:20){
#'    elist[[j]] = stats::ecdf(rgamma(100,1) + rnorm(100, sd=sqrt(0.5)))
#' }
#' 
#' # run EP-means with k clusters 
#' # change the value below to see different settings
#' myk   = 2
#' epout = epmeans(elist, k=myk)
#' 
#' # visualize
#' opar = par(no.readonly=TRUE)
#' par(mfrow=c(1,myk))
#' for (k in 1:myk){
#'   idk = which(epout$cluster==k)
#'   for (i in 1:length(idk)){
#'     if (i<2){
#'       pm = paste("class ",k," (size=",length(idk),")",sep="")
#'       plot(elist[[idk[i]]], verticals=TRUE, lwd=0.25, do.points=FALSE, main=pm)
#'     } else {
#'       plot(elist[[idk[i]]], add=TRUE, verticals=TRUE, lwd=0.25, do.points=FALSE)
#'     }
#'     plot(epout$centers[[k]], add=TRUE, verticals=TRUE, lwd=2, col="red", do.points=FALSE)
#'   }
#' }
#' par(opar)
#' }
#' 
#' @references 
#' \insertRef{henderson_epmeans_2015}{maotai}
#' 
#' @export
epmeans <- function(elist, k=2){
  ###############################################
  # Preprocessing
  clist = elist_epmeans(elist)  # will use quantized ones only / flist = elist_fform(qlist)
  myk   = round(k)
  myn   = length(clist)
  
  # Quantization
  mylength = 1000
  qseq = seq(from=1e-6, to=1-(1e-6), length.out=mylength)
  qmat = array(0,c(myn,mylength))
  for (n in 1:myn){
    qmat[n,] = as.vector(stats::quantile(clist[[n]], qseq))
  }
  
  ###############################################
  # Rcpp k-means
  tmpcpp = cpp_kmeans(qmat, myk)$means
  
  ###############################################
  # Pairwise Distance Computation
  #   wrap
  mylist1 = list()
  mylist2 = list()
  for (n in 1:myn){
    mylist1[[n]] = stats::ecdf(as.vector(qmat[n,]))
  }
  for (k in 1:myk){
    mylist2[[k]] = stats::ecdf(as.vector(tmpcpp[k,]))
  }
  #   compute pairwise distance using Earth Mover's Distance
  pdistmat = dist2_wasserstein(mylist1, mylist2, 1)
  #   index
  label = base::apply(pdistmat, 1, which.min)
  
  
  ###############################################
  # Return : we want to add 'Silhouette'
  output = list()
  output$cluster = as.integer(label)
  output$centers = mylist2
  return(output)
}
  

# ## personal examples
# cdf0  = stats::ecdf(rnorm(100, sd=3))    # original ECDF
# qseq  = seq(from=0,to=1,length.out=1000) # quantile sequence
# quant = stats::quantile(cdf0, qseq)
# cdf1  = stats::ecdf(quant)
# 
# par(mfrow=c(1,2))
# plot(cdf0, main="Original")
# plot(cdf1, main="Recovered")

#' Distance Measures between Multiple Empirical Cumulative Distribution Functions
#' 
#' We measure distance between two empirical cumulative distribution functions (ECDF). For 
#' simplicity, we only take an input of \code{\link[stats]{ecdf}} objects from \pkg{stats} package. 
#' 
#' @param elist a length \eqn{N} list of \code{ecdf} objects.
#' @param method name of the distance/dissimilarity measure. Case insensitive (default: \code{ks}).
#' @param p exponent for \code{Lp} or \code{Wasserstein} distance. 
#' @param as.dist a logical; \code{TRUE} to return \code{dist} object, \code{FALSE} to return an \eqn{(N\times N)} symmetric matrix of pairwise distances (default: \code{FALSE}).
#' @param useR a logical; \code{TRUE} to use R implementation, \code{FALSE} to use C++ implementation (default: \code{TRUE}).
#' 
#' @return either \code{dist} object of an \eqn{(N\times N)} symmetric matrix of pairwise distances by \code{as.dist} argument.
#' 
#' @seealso \code{\link[stats]{ecdf}}
#' 
#' @examples
#' \donttest{
#' ## toy example : 10 of random and uniform distributions
#' mylist = list()
#' for (i in 1:10){
#'   mylist[[i]] = stats::ecdf(stats::rnorm(50, sd=2))
#' }
#' for (i in 11:20){
#'   mylist[[i]] = stats::ecdf(stats::runif(50, min=-5))
#' }
#' 
#' ## compute Kolmogorov-Smirnov distance
#' dm = ecdfdist(mylist, method="KS")
#' 
#' ## visualize
#' mks  =" KS distances of 2 Types"
#' opar = par(no.readonly=TRUE)
#' par(pty="s")
#' image(dm[,nrow(dm):1], axes=FALSE, main=mks)
#' par(opar)
#' }
#' 
#' @export
ecdfdist <- function(elist, method=c("KS","Lp","Wasserstein"), p=2, as.dist=FALSE, useR=TRUE){
  ###############################################
  # Preprocessing
  if (!elist_check(elist)){
    stop("* ecdfdist : input 'elist' should be a list of 'ecdf' objects.")
  }
  methodss = c("ks","wasserstein","lp")
  mymethod = tolower(method)
  mymethod = match.arg(mymethod, methodss)
  myp      = round(p)
  if (myp <= 0){
    stop("* ecdfdist : exponent 'p' should be a nonnegative number.")
  }
  
  ###############################################
  # Computation
  if (as.logical(useR)){
    output = switch(mymethod, 
                    "ks"          = dist_ks(elist),
                    "wasserstein" = dist_wasserstein(elist, myp),
                    "lp"          = dist_lp(elist, myp))
  } else {
    output = switch(mymethod, 
                    "ks"          = fast_dist_ks(elist),
                    "wasserstein" = fast_dist_wasserstein(elist, myp),
                    "lp"          = fast_dist_lp(elist, myp))
  }
  
  ###############################################
  # Report
  if (as.dist){
    return(stats::as.dist(output))
  } else {
    return(output)
  }
}



# single functions --------------------------------------------------------
# (1) dist_ks          : kolmogorov-smirnov
# (2) dist_wasserstein : 1d wasserstein distance
# (3) dist_lp          : Lp distance


#' @keywords internal
#' @noRd
dist_ks <- function(elist){
  trflist  = elist_fform(elist)
  flist = trflist$fval
  nlist = length(flist)
  output = array(0,c(nlist,nlist))
  for (i in 1:(nlist-1)){
    fi = flist[[i]]
    for (j in (i+1):nlist){
      fj = flist[[j]]
      theval = max(abs(fi-fj))
      output[i,j] <- output[j,i] <- theval[1]
    }
  }
  return(output)
}
#' @keywords internal
#' @noRd
dist_lp <- function(elist, p){
  nlist = length(elist)
  trflist  = elist_fform(elist)
  flist = trflist$fval
  nlist = length(flist)
  output = array(0,c(nlist,nlist))
  if (is.infinite(p)){
    for (i in 1:(nlist-1)){
      fi = flist[[i]]
      for (j in (i+1):nlist){
        fj = flist[[j]]
        output[i,j] <- output[j,i] <- base::max(base::abs(fi-fj))[1]
      }
    } 
  } else {
    for (i in 1:(nlist-1)){
      fi = flist[[i]]
      for (j in (i+1):nlist){
        fj = flist[[j]]
        theval = ((integrate_1d(trflist$tseq, (abs(fi-fj)^p)))^(1/p))
        output[i,j] <- output[j,i] <- theval
      }
    }
  }
  return(output)
}
#' @keywords internal
#' @noRd
dist_wasserstein <- function(elist, p){
  nlist = length(elist)
  qseq  = base::seq(from=1e-6, to=1-(1e-6), length.out=8128)
  quants = list() # compute quantile functions first
  for (i in 1:nlist){
    quants[[i]] = as.double(stats::quantile(elist[[i]], qseq))
  }
  
  
  output = array(0,c(nlist,nlist))
  for (i in 1:(nlist-1)){
    vali = quants[[i]]
    for (j in (i+1):nlist){
      valj = quants[[j]]
      valij = abs(vali-valj)
      
      if (is.infinite(p)){
        output[i,j] <- output[j,i] <- base::max(valij)
      } else {
        theval <- ((integrate_1d(qseq, valij^p))^(1/p))
        output[i,j] <- output[j,i] <- theval
      }
    }
  } 
  
  return(output)
}

## wasserstein : http://www-users.math.umn.edu/~bobko001/preprints/2016_BL_Order.statistics_Revised.version.pdf

#' @keywords internal
#' @noRd
fast_dist_ks <- function(elist){
  trflist <- elist_fform(elist)        # returns list(tseq, fval=list of numeric)
  FF <- do.call(cbind, trflist$fval)   # L x N
  cpp_dist_ks(FF)
}

#' @keywords internal
#' @noRd
fast_dist_lp <- function(elist, p){
  trflist <- elist_fform(elist)
  tseq <- as.numeric(trflist$tseq)
  FF <- do.call(cbind, trflist$fval)
  cpp_dist_lp(tseq, FF, p)
}

#' @keywords internal
#' @noRd
fast_dist_wasserstein <- function(elist, p){
  nlist <- length(elist)
  qseq  <- base::seq(from = 1e-6, to = 1 - 1e-6, length.out = 8128L)
  # columns are quantiles for each ecdf on qseq
  Q <- vapply(elist,
              function(f) as.numeric(stats::quantile(f, qseq, names = FALSE)),
              numeric(length(qseq)))
  Q <- as.matrix(Q)                    # Lq x N
  cpp_dist_wasserstein(qseq, Q, p)
}









# ## toy example : random and uniform distributions
# n_test = 200
# n_half = round(n_test/2)
# mylist = vector("list", length=n_test)
# for (i in 1:n_half){
#   mylist[[i]] = stats::ecdf(stats::rnorm(500, sd=2))
# }
# for (i in (n_half+1):n_test){
#   mylist[[i]] = stats::ecdf(stats::runif(500, min=-5))
# }
# 
# # compute Kolmogorov-Smirnov distance
# microbenchmark::microbenchmark(
#   R_ks = ecdfdist(mylist, method="KS"),
#   R_lp = ecdfdist(mylist, method="Lp", p=2),
#   R_wt = ecdfdist(mylist, method="Wasserstein", p=2),
#   C_ks = ecdfdist(mylist, method="KS", useR=FALSE),
#   C_lp = ecdfdist(mylist, method="Lp", p=2, useR=FALSE),
#   C_wt = ecdfdist(mylist, method="Wasserstein", p=2, useR=FALSE),
#   times=3
# )

#' Calculate the Pseudo-Determinant of a Matrix
#' 
#' When a given square matrix \eqn{A} is rank deficient, determinant is zero. 
#' Still, we can compute the pseudo-determinant by multiplying all non-zero 
#' eigenvalues. Since thresholding to determine near-zero eigenvalues is subjective, 
#' we implemented the function as of original limit problem. When matrix is 
#' non-singular, it coincides with traditional determinant.
#' 
#' @param A a square matrix whose pseudo-determinant be computed.
#' 
#' @return a scalar value for computed pseudo-determinant.
#' 
#' @examples 
#' ## show the convergence of pseudo-determinant
#' #  settings
#' n = 10
#' A = cov(matrix(rnorm(5*n),ncol=n))   # (n x n) matrix
#' k = as.double(Matrix::rankMatrix(A)) # rank of A
#' 
#' # iterative computation
#' ntry = 11
#' del.vec = exp(-(1:ntry))
#' det.vec = rep(0,ntry)
#' for (i in 1:ntry){
#'   del = del.vec[i]
#'   det.vec[i] = det(A+del*diag(n))/(del^(n-k))
#' }
#' 
#' # visualize the results
#' opar <- par(no.readonly=TRUE)
#' plot(1:ntry, det.vec, main=paste("true rank is ",k," out of ",n,sep=""),"b", xlab="iterations")
#' abline(h=pdeterminant(A),col="red",lwd=1.2)
#' par(opar)
#' 
#' @references 
#' \insertRef{holbrook_differentiating_2018}{maotai}
#' 
#' @export
pdeterminant <- function(A){ ## wrapped-up function
  if (!is.matrix(A)){
    stop("* pdeterminant : input 'A' should be a matrix.")
  }
  if (nrow(A)!=ncol(A)){
    stop("* pdeterminant : input 'A' should be a square matrix.")
  }
  n = nrow(A)
  k = as.double(Matrix::rankMatrix(A))
  if (k==n){
    return(base::determinant(A))
  } else {
    multccc = 0.9
    del.old = 1
    det.old = det(A+del.old*diag(n))/(del.old^(n-k))
    for (i in 1:496){
      # print(paste("iteration ",i," initiated...",sep=""))
      del.new = del.old*multccc
      det.new = det(A+del.new*diag(n))/(del.new^(n-k))
      if ((abs(det.new-det.old)/abs(det.old))<1e-5){
        return(det.new)
      } else {
        del.old = del.new
        det.old = det.new
      }
    }
    return(det.old)
  }
}



# # # personal tests ----------------------------------------------------------
# n = 10
# A = cov(matrix(rnorm(5*n),ncol=n))
# k = as.double(Matrix::rankMatrix(A))
# 
# ntry = 20
# del.vec = exp(-(1:ntry))
# det.vec = rep(0,ntry)
# for (i in 1:ntry){
#   del = del.vec[i]
#   det.vec[i] = det(A+del*diag(n))/(del^(n-k))
# }
# plot(1:ntry, det.vec, main=paste("true rank is ",k,"/",n,sep=""),"b")
# abline(h=pdeterminant(A),col="red",lwd=1.2)

#' Pairwise Measures for Two Sets of Empirical CDFs
#' 
#' We measure distance between two sets of empirical cumulative distribution functions (ECDF). For 
#' simplicity, we only take an input of \code{\link[stats]{ecdf}} objects from \pkg{stats} package. 
#' 
#' @param elist1 a length \eqn{M} list of \code{ecdf} objects.
#' @param elist2 a length \eqn{N} list of \code{ecdf} objects.
#' @param method name of the distance/dissimilarity measure. Case insensitive.
#' @param p exponent for \code{Lp} or \code{Wasserstein} distance. 
#' 
#' @return an \eqn{(M\times N)} matrix of pairwise distances.
#' 
#' @seealso \code{\link[stats]{ecdf}} \code{\link{ecdfdist}}
#' 
#' @examples 
#' \donttest{
#' ## toy example
#' #  first list : 10 of random and uniform distributions
#' mylist1 = list()
#' for (i in 1:10){ mylist1[[i]] = stats::ecdf(stats::rnorm(50, sd=2))}
#' for (i in 11:20){mylist1[[i]] = stats::ecdf(stats::runif(50, min=-5))}
#' 
#' #  second list : 15 uniform and random distributions
#' mylist2 = list()
#' for (i in 1:15){ mylist2[[i]] = stats::ecdf(stats::runif(50, min=-5))}
#' for (i in 16:30){mylist2[[i]] = stats::ecdf(stats::rnorm(50, sd=2))}
#' 
#' ## compute Kolmogorov-Smirnov distance
#' dm2ks = ecdfdist2(mylist1, mylist2, method="KS")
#' dm2lp = ecdfdist2(mylist1, mylist2, method="lp")
#' dm2wa = ecdfdist2(mylist1, mylist2, method="wasserstein")
#' nrs   = nrow(dm2ks)
#' 
#' ## visualize
#' opar = par(no.readonly=TRUE)
#' par(mfrow=c(1,3), pty="s")
#' image(dm2ks[,nrs:1], axes=FALSE, main="Kolmogorov-Smirnov")
#' image(dm2lp[,nrs:1], axes=FALSE, main="L2")
#' image(dm2wa[,nrs:1], axes=FALSE, main="Wasserstein")
#' par(opar)
#' }
#' 
#' @export
ecdfdist2 <- function(elist1, elist2, method=c("KS","Lp","Wasserstein"), p=2){
  ###############################################
  # Preprocessing
  if (!elist_check(elist1)){stop("* ecdfdist2 : input 'elist1' should be a list of 'ecdf' objects.")}
  if (!elist_check(elist2)){stop("* ecdfdist2 : input 'elist2' should be a list of 'ecdf' objects.")}
  methodss = c("ks","wasserstein","lp")
  mymethod = tolower(method)
  mymethod = match.arg(mymethod, methodss)
  myp      = as.integer(p)
  if (myp <= 0){
    stop("* ecdfdist2 : exponent 'p' should be a nonnegative number.")
  }
  
  ###############################################
  # Computation
  output = switch(mymethod, 
                  "ks"          = dist2_ks(elist1, elist2),
                  "wasserstein" = dist2_wasserstein(elist1, elist2, myp),
                  "lp"          = dist2_lp(elist1, elist2, myp))
  
  ###############################################
  # Return
  return(output)
}



# single functions --------------------------------------------------------
# (1) dist2_ks          : kolmogorov-smirnov
# (2) dist2_wasserstein : 1d wasserstein distance
# (3) dist2_lp          : Lp distance

#' @keywords internal
#' @noRd
dist2_ks <- function(elist1, elist2){
  M = length(elist1)
  N = length(elist2)
  
  trflst = elist_fform(c(elist1, elist2))
  flist1 = trflst$fval[1:M]
  flist2 = trflst$fval[(M+1):(M+N)]
  
  output = array(0,c(M,N))
  for (i in 1:M){
    fi = flist1[[i]]
    for (j in 1:N){
      fj = flist2[[j]]
      theval = max(abs(fi-fj))
      output[i,j] <- theval[1]
    }
  }
  return(output)
}
#' @keywords internal
#' @noRd
dist2_lp <- function(elist1, elist2, p){
  M = length(elist1)
  N = length(elist2)
  
  trflst = elist_fform(c(elist1, elist2))
  flist1 = trflst$fval[1:M]
  flist2 = trflst$fval[(M+1):(M+N)]
  
  output = array(0,c(M,N))
  for (i in 1:M){
    fi = flist1[[i]]
    for (j in 1:N){
      fj = flist2[[j]]
      if (is.infinite(p)){
        output[i,j] = base::max(base::abs(fi-fj))[1]
      } else {
        output[i,j] <- ((integrate_1d(trflst$tseq, (abs(fi-fj)^p)))^(1/p))
      }
    }
  }
  return(output)
}
#' @keywords internal
#' @noRd
dist2_wasserstein <- function(elist1, elist2, p){
  M = length(elist1)
  N = length(elist2)
  
  trflst = elist_fform(c(elist1, elist2))
  flist1 = trflst$fval[1:M]
  flist2 = trflst$fval[(M+1):(M+N)]
  
  qseq    = base::seq(from=1e-6, to=1-(1e-6), length.out=8128)
  quants1 = list()  # compute quantile functions first
  quants2 = list()
  for (i in 1:M){
    quants1[[i]] = as.double(stats::quantile(elist1[[i]], qseq))
  }
  for (j in 1:N){
    quants2[[j]] = as.double(stats::quantile(elist2[[j]], qseq))
  }
  
  output = array(0,c(M,N))
  for (i in 1:M){
    vali = quants1[[i]]
    for (j in 1:N){
      valj  = quants2[[j]]
      valij = abs(vali-valj)
      
      if (is.infinite(p)){
        output[i,j] = base::max(valij)
      } else {
        output[i,j] <- ((integrate_1d(qseq, valij^p))^(1/p))
      }
    }
  }
  return(output)
}

# Hidden Functions : Computation
# 
# (01) hidden_gaussbary_2002R
#      - Ruschendorf, Uckelmann (2002) : On the n-coupling problem algorithm
#
# (02) hidden_gaussbary_2016A
#      - Alvarez-Esteban et al (2016) : A fixed-point approach to barycenters in Wasserstein space
#      - See Theorem 4.2 for the updating rules.
#
# (03) hidden_projsimplex_sorting : method by Wang and Carreira-Perpinan
#      Held, P. Wolfe, and H. Crowder, “Validation of subgradient optimization,” Mathematical Programming, vol. 6, pp. 62–88, 1974
#      (above is known by https://www.optimization-online.org/DB_FILE/2014/08/4498.pdf)
#      url : https://eng.ucmerced.edu/people/wwang5/papers/SimplexProj.pdf


# (01) hidden_gaussbary_2002R ---------------------------------------------
#      array3d : (p x p x N) array of SPD matrices.
#      weight  : N vector of weights. C++ handles L1 normalization.
#
#      RETURN a list containing
#             $mean : (p x p) barycenter matrix
#             $iter : the number of iterations
#' @keywords internal
#' @noRd
hidden_gaussbary_2002R <- function(array3d, weight, maxiter=50, abstol=(1e-8)){
  # minimal check
  weight = as.vector(weight)
  if (length(weight)!=dim(array3d)[3]){
    stop("* hidden_gaussbary_2002R : non-matching weight.")
  }
  
  # compute
  return(src_gaussbary_2002R(array3d, weight, maxiter, abstol))
}

# (02) hidden_gaussbary_2016A ---------------------------------------------
#      array3d : (p x p x N) array of SPD matrices.
#      weight  : N vector of weights. C++ handles L1 normalization.
#
#      RETURN a list containing
#             $mean : (p x p) barycenter matrix
#             $iter : the number of iterations
#' @keywords internal
#' @noRd
hidden_gaussbary_2016A <- function(array3d, weight, maxiter=50, abstol=(1e-8)){
  # minimal check
  weight = as.vector(weight)
  if (length(weight)!=dim(array3d)[3]){
    stop("* hidden_gaussbary_2016A : non-matching weight.")
  }
  
  # compute
  return(src_gaussbary_2016A(array3d, weight, maxiter, abstol))
}



# # test scenario 1 : random near-identity covariances
# # test scenario 2 : generate randomly some data + add noise
# p = 10
# n = 50
# noise = 0.01
# 
# somedat = matrix(runif(50*p), ncol=p)
# mycovs = array(0, c(p,p,n))
# for (i in 1:n){
#   mycovs[,,i] = cov(somedat + matrix(rnorm(50*p), ncol=p)*noise)
#   # print(paste0("The ",i,"-th matrix has rank=",round(Matrix::rankMatrix(mycovs[,,i]))))
# }
# myweight = rep(1/n, n)
# 
# 
# res_2002 = hidden_gaussbary_2002R(mycovs, myweight)
# res_2016 = hidden_gaussbary_2016A(mycovs, myweight)
# par(mfrow=c(1,3), pty="s")
# image(cov(somedat), main="true covariance")
# image(res_2002$mean, main=paste0("02R:iter=",round(res_2002$iter),"/error=",round(norm(res_2002$mean-cov(somedat),"F"),3)))
# image(res_2016$mean, main=paste0("16A:iter=",round(res_2016$iter),"/error=",round(norm(res_2016$mean-cov(somedat),"F"),3)))
# 
# microbenchmark(res_2002 = hidden_gaussbary_2002R(mycovs, myweight),
#                res_2016 = hidden_gaussbary_2016A(mycovs, myweight))



# (03) hidden_projsimplex_sorting ---------------------------------------------
#' @keywords internal
#' @noRd
hidden_projsimplex_sorting <- function(input){
  # preprocessing
  y = as.vector(input)
  D = length(y)
  u = base::sort(y, decreasing = TRUE)
  
  # main part
  uvec = rep(0,D)
  for (j in 1:D){
    uvec[j] = u[j] + (1/j)*(1-base::sum(u[1:j]))
  }
  rho = max(which(uvec > 0))
  lbd = (1/rho)*(1 - base::sum(u[1:rho]))
  
  # finalize
  x = pmax(y+lbd, 0)
  return(x)
}
# xx = matrix(rnorm(1000*2), ncol=2)
# yy = array(0,c(1000,2))
# for (i in 1:1000){
#   yy[i,] = hidden_projsimplex_13W(xx[i,])
# }
# plot(xx, main="scatter")
# points(yy, pch=19, col="red")

#' Generate Index for Stationary Bootstrapping
#' 
#' Assuming data being dependent with cardinality \code{N}, \code{boot.stationary} returns 
#' a vector of index that is used for stationary bootstrapping. To describe, starting points 
#' are drawn from uniform distribution over \code{1:N} and the size of each block is 
#' determined from geometric distribution with parameter \eqn{p}.
#' 
#' 
#' @param N the number of observations.
#' @param p parameter for geometric distribution with the size of each block.
#' 
#' @return a vector of length \code{N} for moving block bootstrap sampling.
#' 
#' @examples
#' \donttest{
#' ## example : bootstrap confidence interval of mean and variances
#' vec.x = seq(from=0,to=10,length.out=100)
#' vec.y = sin(1.21*vec.x) + 2*cos(3.14*vec.x) + rnorm(100,sd=1.5)
#' data.mu  = mean(vec.y)
#' data.var = var(vec.y)
#' 
#' ## apply stationary bootstrapping
#' nreps   = 50
#' vec.mu  = rep(0,nreps)
#' vec.var = rep(0,nreps)
#' for (i in 1:nreps){
#'    sample.id = boot.stationary(100)
#'    sample.y  = vec.y[sample.id]
#'    vec.mu[i]  = mean(sample.y)
#'    vec.var[i] = var(sample.y)
#'    print(paste("iteration ",i,"/",nreps," complete.", sep=""))
#' }
#' 
#' ## visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,3), pty="s")
#' plot(vec.x, vec.y, type="l", main="1d signal")  # 1d signal
#' hist(vec.mu, main="mean CI", xlab="mu")         # mean
#' abline(v=data.mu, col="red", lwd=4)
#' hist(vec.var, main="variance CI", xlab="sigma") # variance
#' abline(v=data.var, col="blue", lwd=4)
#' par(opar)
#' }
#' 
#' @references 
#' \insertRef{politis_stationary_1994}{maotai}
#' 
#' @export
boot.stationary <- function(N, p=0.25){
  ###################################################################
  # Preprocessing
  myn   = round(N)
  myp   = as.double(p)
  vec1n = 1:myn
  
  ###################################################################
  # Computation
  output = c()
  while (length(output)<myn){
    # 1. select the starting index
    idst = base::sample(vec1n,1)
    # 2. select the number
    nb = 0
    while (nb < 1){
      nb = stats::rgeom(1, myp)
    }
    # 3. select indices
    cnow = idst:(idst+nb-1)
    cnow = (cnow%%myn)
    cnow[(cnow==0)] = myn
    # 4. concatenate
    output = c(output, cnow)
  }
  
  ###################################################################
  # Return
  return(output[1:myn])
}

#' Solve Lyapunov Equation
#' 
#' The Lyapunov equation is of form
#' \deqn{AX + XA^\top = Q}
#' where \eqn{A} and \eqn{Q} are square matrices of same size. Above form is also known as \emph{continuous} form. 
#' This is a wrapper of \code{armadillo}'s \code{sylvester} function.
#' 
#' @param A a \eqn{(p\times p)} matrix as above.
#' @param Q a \eqn{(p\times p)} matrix as above.
#' 
#' @return a solution matrix \eqn{X} of size \eqn{(p\times p)}.
#' 
#' @examples 
#' ## simulated example
#' #  generate square matrices
#' A = matrix(rnorm(25),nrow=5)
#' X = matrix(rnorm(25),nrow=5)
#' Q = A%*%X + X%*%t(A)
#' 
#' #  solve using 'lyapunov' function
#' solX = lyapunov(A,Q)
#' \dontrun{
#' pm1 = "* Experiment with Lyapunov Solver"
#' pm2 = paste("* Absolute Error  : ",norm(solX-X,"f"),sep="")
#' pm3 = paste("* Relative Error  : ",norm(solX-X,"f")/norm(X,"f"),sep="")
#' cat(paste(pm1,"\n",pm2,"\n",pm3,sep=""))
#' }
#' 
#' @references 
#' \insertRef{sanderson_armadillo_2016}{maotai}
#' 
#' \insertRef{eddelbuettel_rcpparmadillo_2014}{maotai}
#' 
#' @export
lyapunov <- function(A, Q){
  ###################################################################
  # check square matrix
  if (!check_sqmat(A)){
    stop("* lyapunov : an input 'A' should be a square matrix.")
  }
  if (!check_sqmat(Q)){
    stop("* lyapunov : an input 'Q' should be a square matrix.")
  }
  
  ###################################################################
  # arrange for RcppArmadillo format
  B = t(A)
  C = -Q
  
  ###################################################################
  # pass and return
  return(solve_lyapunov(A,B,C))
}

#' Leading Eigenvector Dynamics Analysis
#' 
#' Compute the leading eigenvector dynamics analysis (LEiDA) of a 
#' multivariate time series as appearing in computational neuroscience.
#' 
#' @param X A \eqn{(T,N)} matrix of multivariate time series data, where \eqn{T} is the number of time points and \eqn{N} is the number of ROIs.
#' @param TR Repetition time (in secounds)
#' @param bp Bandpass filter, a vector of length 2 with the lower and upper bounds of the bandpass filter in Hz. Default is \code{c(0.01, 0.10)}.
#' @param b_ord Butterworth order, a positive integer. Default is 2.
#' 
#' @return a list containing \describe{
#' \item{V}{A \eqn{(T,N)} matrix of the leading eigenvector time series.}
#' \item{FCD_cos}{A \eqn{(T,T)} matrix of the functional connectivity dynamics (FCD) using cosine similarity.}
#' \item{FCD_cor}{A \eqn{(T,T)} matrix of the functional connectivity dynamics (FCD) using Pearson correlation.}
#' }
#' 
#' @export
LEiDA <- function(X, TR, bp=c(0.01, 0.10), b_ord = 2){
  # 1. PREPROCESSING
  fs <- 1/as.double(TR)
  par_T <- base::nrow(X)
  par_N <- base::ncol(X)
  
  Wn <- bp/(fs/2) # normalised cut-offs
  bpFilt <- gsignal::butter(round(b_ord), Wn, type="pass")
  
  # 2. DEMEANING, DETRENDING, AND FILTERING
  Xclean <- array(0,c(par_T, par_N)) # filtered data
  for (j in seq_len(par_N)){         # run for each ROI
    ts <- as.vector(X[,j])
    ts <- ts - base::mean(ts)                  # demeaning
    ts <- pracma::detrend(ts, tt = "linear")   # detrending
    Xclean[,j] = gsignal::filtfilt(bpFilt, ts) # filtering
  }
  
  # 3. PHASE 
  Phase <- array(0,c(par_T, par_N))
  for (j in seq_len(par_N)){
    Phase[,j] = base::Arg(gsignal::hilbert(Xclean[,j])) # phase
  }
  
  # 4. COMPUTE AND RETURN
  computed = src_leida(Phase)
  return(computed)
}

#' Check for Distance Matrix
#' 
#' This function checks whether the distance matrix \eqn{D:=d_{ij} = d(x_i, x_j)} satisfies 
#' three axioms to make itself a semimetric, which are (1) \eqn{d_{ii} = 0}, (2) \eqn{d_{ij} > 0} for \eqn{i\neq j}, and 
#' (3) \eqn{d_{ij} = d_{ji}}.
#' 
#' @param d \code{"dist"} object or \eqn{(N\times N)} matrix of pairwise distances.
#' 
#' @return a logical; \code{TRUE} if it satisfies metric property, \code{FALSE} otherwise.
#' 
#' @examples 
#' ## Let's use L2 distance matrix of iris dataset
#' data(iris)
#' dx = as.matrix(stats::dist(iris[,1:4]))
#' 
#' # perturb d(i,j) 
#' dy = dx  
#' dy[1,2] <- dy[2,1] <- 10
#' 
#' # run the algorithm
#' checkdist(dx)
#' checkdist(dy)
#' 
#' @seealso \code{\link{checkmetric}}
#' @export
checkdist <- function(d){
  if (inherits(d, "dist")){
    d = as.matrix(d)
  } else {
    if (!is.matrix(d)){
      stop("* checkdist : input 'd' should be a matrix.")
    }
  }
  # 1. square matrix
  if (nrow(d)!=ncol(d)){
    message(" checkdist : input 'd' is not a square matrix.")
    return(FALSE)
  }
  # 2. zero diagonals
  if (any(diag(d)!=0)){
    message(" checkdist : input 'd' has non-zero diagonals.")
    return(FALSE)
  }
  # 3. all positive elements
  if (any(d < 0)){
    message(" checkdist : input 'd' has negative values.")
    return(FALSE)
  }
  # 4. symmetric
  if (!base::isSymmetric(d)){
    message(" checkdist : input 'd' is not symmetric.")
    return(FALSE)
  }
  
  return(TRUE)
}

# auxiliary functions for computation -------------------------------------
# (1) aux_pinv       : pseudo-inverse
# (2) aux_pseudomean : compute distance from 1st observation to pseudo mean by rest points
 
# (1) aux_pinv ------------------------------------------------------------
#' @keywords internal
aux_pinv <- function(A){
  svdA      = base::svd(A)
  tolerance = (.Machine$double.eps)*max(c(nrow(A),ncol(A)))*as.double(max(svdA$d))
  
  idxcut    = which(svdA$d <= tolerance)
  invDvec   = (1/svdA$d)
  invDvec[idxcut] = 0
  
  output = (svdA$v%*%diag(invDvec)%*%t(svdA$u))
  return(output)
}


# (2) aux_pseudomean ------------------------------------------------------
#' @keywords internal
aux_pseudomean <- function(dmat){
  # we need embedding .. umm .. automatic dimension selection
  if (nrow(dmat)==1){
    stop("* aux_pseudomean : error..")
  } else if (nrow(dmat)==2){
    return(dmat[1,2])
  } else {
    embedded = aux_pseudomean_auto(dmat)
    n = nrow(embedded)
    p = ncol(embedded)
    
    # centering based on other points
    emcenter = as.vector(base::colMeans(embedded[2:n,]))
    embednew = embedded - matrix(rep(emcenter,n), ncol=p, byrow=TRUE)
    
    # compute scalar
    d1mat = dmat[2:n,2:n]                          # d(x,y)
    d2mat = as.matrix(stats::dist(embednew[2:n,])) # ||x-y||
    d12mat = (d1mat*d2mat)
    d22mat = (d2mat^2)
    dlower = base::lower.tri(d12mat)
    cstar =sum(d12mat[dlower])/sum(d22mat[dlower])
    
    # update embednew and compute 
    erow1 = cstar*as.vector(embednew[1,])
    return(sqrt(sum(erow1^2))) 
  }
}
#' @keywords internal
#' @noRd
aux_pseudomean_auto <- function(dmat){ # only positive eigenvalues' part
  n = nrow(dmat)
  J = diag(rep(1,n))-(1/n)*outer(rep(1,n),rep(1,n))
  B = -(J%*%(dmat^2)%*%J)/2.0
  eigB = base::eigen(B, symmetric = TRUE) # decreasing order
  
  m = max(length(which(eigB$values > 0)),2)
  X = (eigB$vectors[,1:m])%*%(base::diag(sqrt(eigB$values[1:m])))
  return(X)
}

# # personal test : seems like it's working well enough
# x = rnorm(5, mean=3)
# y = matrix(rnorm(10*5),ncol=5)
# 
# dmat = as.matrix(dist(rbind(x,y)))
# val.alg = aux_pseudomean(dmat)
# val.true = sqrt(sum((x-as.vector(colMeans(y)))^2))

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

bwp_project_gaussians_rcpp <- function(M, Sigma, d, max_iter = 200L, step_init = 1.0, tol_grad = 1e-6, tol_obj = 1e-10, eps = 1e-12, verbose = FALSE, armijo_c = 1e-4, armijo_beta = 0.5, ls_max = 25L, ls_min_step = 1e-12) {
    .Call('_maotai_bwp_project_gaussians_rcpp', PACKAGE = 'maotai', M, Sigma, d, max_iter, step_init, tol_grad, tol_obj, eps, verbose, armijo_c, armijo_beta, ls_max, ls_min_step)
}

sphere_dist <- function(vecx, vecy) {
    .Call('_maotai_sphere_dist', PACKAGE = 'maotai', vecx, vecy)
}

cpp_WL_normalise <- function(X) {
    .Call('_maotai_cpp_WL_normalise', PACKAGE = 'maotai', X)
}

cpp_WL_weighted_mean <- function(X, weights) {
    .Call('_maotai_cpp_WL_weighted_mean', PACKAGE = 'maotai', X, weights)
}

compute_SSR <- function(D, Delta) {
    .Call('_maotai_compute_SSR', PACKAGE = 'maotai', D, Delta)
}

compute_stress <- function(D, Dhat) {
    .Call('_maotai_compute_stress', PACKAGE = 'maotai', D, Dhat)
}

main_bmds <- function(D, X0, sigg0, a, alpha, maxiter, constant, verbose, betas) {
    .Call('_maotai_main_bmds', PACKAGE = 'maotai', D, X0, sigg0, a, alpha, maxiter, constant, verbose, betas)
}

caml_frechet_mean <- function(Y_list_R, w, maxit = 200L, tol = 1e-8, step0 = 1.0, c1 = 1e-4, beta = 0.5, verbose = FALSE) {
    .Call('_maotai_caml_frechet_mean', PACKAGE = 'maotai', Y_list_R, w, maxit, tol, step0, c1, beta, verbose)
}

aux_shortestpath <- function(wmat) {
    .Call('_maotai_aux_shortestpath', PACKAGE = 'maotai', wmat)
}

cppsub_2007Wang <- function(V0, mm, d, Spu, Stu, maxiter, eps) {
    .Call('_maotai_cppsub_2007Wang', PACKAGE = 'maotai', V0, mm, d, Spu, Stu, maxiter, eps)
}

gradF <- function(func, xnow, h) {
    .Call('_maotai_gradF', PACKAGE = 'maotai', func, xnow, h)
}

dat2centers <- function(data, centers) {
    .Call('_maotai_dat2centers', PACKAGE = 'maotai', data, centers)
}

cpp_sylvester <- function(A, B, C) {
    .Call('_maotai_cpp_sylvester', PACKAGE = 'maotai', A, B, C)
}

solve_lyapunov <- function(A, B, C) {
    .Call('_maotai_solve_lyapunov', PACKAGE = 'maotai', A, B, C)
}

cpp_weiszfeld <- function(X, abstol, maxiter, xinit, weights, epsnum) {
    .Call('_maotai_cpp_weiszfeld', PACKAGE = 'maotai', X, abstol, maxiter, xinit, weights, epsnum)
}

cpp_kmeans <- function(data, k) {
    .Call('_maotai_cpp_kmeans', PACKAGE = 'maotai', data, k)
}

emds_gamma0 <- function(dmat) {
    .Call('_maotai_emds_gamma0', PACKAGE = 'maotai', dmat)
}

cpp_pairwise_L2 <- function(muA, muB, covA, covB) {
    .Call('_maotai_cpp_pairwise_L2', PACKAGE = 'maotai', muA, muB, covA, covB)
}

integrate_1d <- function(tseq, fval) {
    .Call('_maotai_integrate_1d', PACKAGE = 'maotai', tseq, fval)
}

cpp_pdist <- function(X) {
    .Call('_maotai_cpp_pdist', PACKAGE = 'maotai', X)
}

cpp_geigen <- function(A, B) {
    .Call('_maotai_cpp_geigen', PACKAGE = 'maotai', A, B)
}

cpp_triangle <- function(D) {
    .Call('_maotai_cpp_triangle', PACKAGE = 'maotai', D)
}

cpp_metricdepth <- function(D) {
    .Call('_maotai_cpp_metricdepth', PACKAGE = 'maotai', D)
}

cpp_dist_ks <- function(F) {
    .Call('_maotai_cpp_dist_ks', PACKAGE = 'maotai', F)
}

cpp_dist_lp <- function(t, F, p) {
    .Call('_maotai_cpp_dist_lp', PACKAGE = 'maotai', t, F, p)
}

cpp_dist_wasserstein <- function(q, Q, p) {
    .Call('_maotai_cpp_dist_wasserstein', PACKAGE = 'maotai', q, Q, p)
}

lp_ecdf_distance <- function(x, y, p = 2.0) {
    .Call('_maotai_lp_ecdf_distance', PACKAGE = 'maotai', x, y, p)
}

ks_ecdf_distance <- function(x, y) {
    .Call('_maotai_ks_ecdf_distance', PACKAGE = 'maotai', x, y)
}

lp_ecdf_distance_matrix <- function(samples, p = 2.0) {
    .Call('_maotai_lp_ecdf_distance_matrix', PACKAGE = 'maotai', samples, p)
}

lp_ecdf_distance_to_reference <- function(samples, reference, p = 2.0) {
    .Call('_maotai_lp_ecdf_distance_to_reference', PACKAGE = 'maotai', samples, reference, p)
}

wasserstein_p_unweighted <- function(x, y, p = 2.0) {
    .Call('_maotai_wasserstein_p_unweighted', PACKAGE = 'maotai', x, y, p)
}

wasserstein_p_weighted <- function(x, wx, y, wy, p = 2.0) {
    .Call('_maotai_wasserstein_p_weighted', PACKAGE = 'maotai', x, wx, y, wy, p)
}

wasserstein_p_distance_matrix <- function(samples, p = 2.0) {
    .Call('_maotai_wasserstein_p_distance_matrix', PACKAGE = 'maotai', samples, p)
}

cpp_mmds <- function(D, ndim, maxiter, abstol) {
    .Call('_maotai_cpp_mmds', PACKAGE = 'maotai', D, ndim, maxiter, abstol)
}

src_smacof <- function(D, W, ndim, maxiter, abstol, use_gutman) {
    .Call('_maotai_src_smacof', PACKAGE = 'maotai', D, W, ndim, maxiter, abstol, use_gutman)
}

eval_gaussian <- function(x, mu, cov) {
    .Call('_maotai_eval_gaussian', PACKAGE = 'maotai', x, mu, cov)
}

eval_gaussian_data <- function(X, mu, cov) {
    .Call('_maotai_eval_gaussian_data', PACKAGE = 'maotai', X, mu, cov)
}

eval_gmm_data <- function(X, mus, covs, weight) {
    .Call('_maotai_eval_gmm_data', PACKAGE = 'maotai', X, mus, covs, weight)
}

eval_gmm <- function(x, mus, covs, weight) {
    .Call('_maotai_eval_gmm', PACKAGE = 'maotai', x, mus, covs, weight)
}

src_construct_by_knn <- function(nn_idx, intersection) {
    .Call('_maotai_src_construct_by_knn', PACKAGE = 'maotai', nn_idx, intersection)
}

src_gaussbary_2002R <- function(array3d, weight, maxiter, abstol) {
    .Call('_maotai_src_gaussbary_2002R', PACKAGE = 'maotai', array3d, weight, maxiter, abstol)
}

src_gaussbary_2016A <- function(array3d, weight, maxiter, abstol) {
    .Call('_maotai_src_gaussbary_2016A', PACKAGE = 'maotai', array3d, weight, maxiter, abstol)
}

src_cov2corr <- function(covmat) {
    .Call('_maotai_src_cov2corr', PACKAGE = 'maotai', covmat)
}

src_leida <- function(phase) {
    .Call('_maotai_src_leida', PACKAGE = 'maotai', phase)
}

#' Distance Measures between Samples through Empirical Cumulative Distribution Functions
#' 
#' 
#' We measure distance between two empirical cumulative distribution functions of the data. 
#' Unlike \code{\link[maotai]{ecdfdist}}, this function takes raw data samples as input, and 
#' internally computes the empirical cumulative distribution functions (ECDF) for distance calculations.
#' 
#' @param veclist a length \eqn{N} list of vectors.
#' @param method name of the distance/dissimilarity measure. Case insensitive (default: \code{ks}).
#' @param p exponent for \code{Lp} or \code{Wasserstein} distance (default: \code{p=1}).
#' @param as.dist a logical; \code{TRUE} to return \code{dist} object, \code{FALSE} to return an \eqn{(N\times N)} symmetric matrix of pairwise distances (default: \code{FALSE}).
#' 
#' @return either \code{dist} object of an \eqn{(N\times N)} symmetric matrix of pairwise distances by \code{as.dist} argument.
#' 
#' @examples
#' \donttest{
#' ## toy example : 10 of random and uniform distributions
#' mylist = list()
#' for (i in 1:10){
#'   mylist[[i]] = stats::rnorm(50, sd=2)
#' }
#' for (i in 11:20){
#'   mylist[[i]] = stats::runif(50, min=-5)
#' }
#' 
#' ## compute three distances
#' d_KS = ecdfdistS(mylist, method="KS")
#' d_LP = ecdfdistS(mylist, method="Lp")
#' d_OT = ecdfdistS(mylist, method="Wasserstein")
#' 
#' ## visualize
#' opar = par(no.readonly=TRUE)
#' par(mfrow=c(1,3), pty="s")
#' image(d_KS[,nrow(d_KS):1], axes=FALSE, main="Kolmogorov-Smirnov")
#' image(d_LP[,nrow(d_LP):1], axes=FALSE, main="Lp (p=1)")
#' image(d_OT[,nrow(d_OT):1], axes=FALSE, main="Wasserstein (p=1)")
#' par(opar)
#' }
#' 
#' @export
ecdfdistS <- function(veclist, method=c("KS","Lp","Wasserstein"), p=1, as.dist=FALSE){
  ###############################################
  # Preprocessing
  if (!is.list(veclist)){
    stop("* ecdfdistS : input 'veclist' should be a list of numeric vectors.")
  }
  if (!check_veclist(veclist)){
    stop("* ecdfdistS : input 'veclist' should be a list of numeric vectors.")
  }
  methodss = c("ks","wasserstein","lp")
  mymethod = tolower(method)
  mymethod = match.arg(mymethod, methodss)
  myp      = round(p)
  if (myp < 1){
    stop("* ecdfdistS : exponent 'p' should be >= 1.")
  }
  
  ################################################
  # Computation
  output = switch(mymethod,
                  "lp" = lp_ecdf_distance_matrix(veclist, myp),
                  "ks" = lp_ecdf_distance_matrix(veclist, Inf),
                  "wasserstein" = wasserstein_p_distance_matrix(veclist, myp))
  
  ###############################################
  # Report
  if (as.dist){
    return(stats::as.dist(output))
  } else {
    return(output)
  }
}


# ## toy example : N(0,1), Unif[0,10], and N(-5,4)
# n_test  = 900
# n_third = round(n_test/3)
# 
# list_sam  = vector("list", length=n_test)
# list_ecdf = vector("list", length=n_test)
# 
# idnow = 0
# for (i in seq_len(n_third)){
#   idnow = idnow+1
#   n_sam = sample(100:500, 1)
#   list_sam[[idnow]] = stats::rnorm(n_sam, mean=0, sd=1)
#   list_ecdf[[idnow]] = stats::ecdf(list_sam[[idnow]])
# }
# for (i in seq_len(n_third)){
#   idnow = idnow+1
#   n_sam = sample(100:500, 1)
#   list_sam[[idnow]] = stats::runif(n_sam, min=0, max=10)
#   list_ecdf[[idnow]] = stats::ecdf(list_sam[[idnow]])
# }
# for (i in seq_len(n_third)){
#   idnow = idnow+1
#   n_sam = sample(100:500, 1)
#   list_sam[[idnow]] = stats::rnorm(n_sam, mean=-5, sd=2)
#   list_ecdf[[idnow]] = stats::ecdf(list_sam[[idnow]])
# }
# 
# R_ks = ecdfdist(list_ecdf, method="KS")
# R_lp = ecdfdist(list_ecdf, method="Lp")
# R_wt = ecdfdist(list_ecdf, method="Wasserstein")
# S_ks = ecdfdistS(list_sam, method="KS")
# S_lp = ecdfdistS(list_sam, method="Lp")
# S_wt = ecdfdistS(list_sam, method="Wasserstein")
# 
# par(mfrow=c(1,3), pty="s")
# image(abs(R_ks-S_ks), main=paste0("Diff:KS-",round(norm(R_ks-S_ks,"F"),3)))
# image(abs(R_lp-S_lp), main=paste0("Diff:LP-",round(norm(R_lp-S_lp,"F"),3)))
# image(abs(R_wt-S_wt), main=paste0("Diff:OT-",round(norm(R_wt-S_wt,"F"),3)))
# #   
# # 
# # 
# # 
# # # 
# # compute Kolmogorov-Smirnov distance
# microbenchmark::microbenchmark(
#   R_ks = ecdfdist(list_ecdf, method="KS"),
#   R_lp = ecdfdist(list_ecdf, method="Lp"),
#   R_wt = ecdfdist(list_ecdf, method="Wasserstein"),
#   S_ks = ecdfdistS(list_sam, method="KS"),
#   S_lp = ecdfdistS(list_sam, method="Lp"),
#   S_wt = ecdfdistS(list_sam, method="Wasserstein"),
#   times=5
# )
# 
# 
# 
# 
# 

#' Kernel Two-sample Test with Maximum Mean Discrepancy
#' 
#' Maximum Mean Discrepancy (MMD) as a measure of discrepancy between 
#' samples is employed as a test statistic for two-sample hypothesis test 
#' of equal distributions. Kernel matrix \eqn{K} is a symmetric square matrix 
#' that is positive semidefinite.
#' 
#' @param K kernel matrix or an object of \code{kernelMatrix} class from \pkg{kernlab} package.
#' @param label label vector of class indices. 
#' @param method type of estimator to be used. \code{"b"} for biased and \code{"u"} for unbiased estimator of MMD.
#' @param mc.iter the number of Monte Carlo resampling iterations.
#' 
#' @return a (list) object of \code{S3} class \code{htest} containing: \describe{
#' \item{statistic}{a test statistic.}
#' \item{p.value}{\eqn{p}-value under \eqn{H_0}.}
#' \item{alternative}{alternative hypothesis.}
#' \item{method}{name of the test.}
#' \item{data.name}{name(s) of provided kernel matrix.}
#' }
#' 
#' @examples 
#' ## small test for CRAN submission
#' dat1 <- matrix(rnorm(60, mean= 1), ncol=2) # group 1 : 30 obs of mean  1
#' dat2 <- matrix(rnorm(50, mean=-1), ncol=2) # group 2 : 25 obs of mean -1
#' 
#' dmat <- as.matrix(dist(rbind(dat1, dat2)))  # Euclidean distance matrix
#' kmat <- exp(-(dmat^2))                      # build a gaussian kernel matrix
#' lab  <- c(rep(1,30), rep(2,25))             # corresponding label
#' 
#' mmd2test(kmat, lab)                         # run the code !
#' 
#' \dontrun{
#' ## WARNING: computationally heavy. 
#' #  Let's compute empirical Type 1 error at alpha=0.05
#' niter = 496  
#' pvals1 = rep(0,niter)
#' pvals2 = rep(0,niter)
#' for (i in 1:niter){
#'   dat = matrix(rnorm(200),ncol=2)
#'   lab = c(rep(1,50), rep(2,50))
#'   lbd = 0.1
#'   kmat = exp(-lbd*(as.matrix(dist(dat))^2))
#'   pvals1[i] = mmd2test(kmat, lab, method="b")$p.value
#'   pvals2[i] = mmd2test(kmat, lab, method="u")$p.value
#'   print(paste("iteration ",i," complete..",sep=""))
#' }
#' 
#' #  Visualize the above at multiple significance levels
#' alphas = seq(from=0.001, to=0.999, length.out=100)
#' errors1 = rep(0,100)
#' errors2 = rep(0,100)
#' for (i in 1:100){
#'    errors1[i] = sum(pvals1<=alphas[i])/niter
#'    errors2[i] = sum(pvals2<=alphas[i])/niter
#' }
#' 
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,2), pty="s")
#' plot(alphas, errors1, "b", main="Biased Estimator Error", 
#'      xlab="alpha", ylab="error", cex=0.5)
#' abline(a=0,b=1, lwd=1.5, col="red")
#' plot(alphas, errors2, "b", main="Unbiased Estimator Error", 
#'      xlab="alpha", ylab="error", cex=0.5)
#' abline(a=0,b=1, lwd=1.5, col="blue")
#' par(opar)
#' }
#' 
#' @references 
#' \insertRef{gretton_kernel_2012}{maotai}
#' 
#' @export
mmd2test <- function(K, label, method=c("b","u"), mc.iter=999){
  ###############################################
  # Preprocessing
  DNAME = deparse(substitute(K))
  # 1. K : kernel matrix
  if (inherits(K, "kernelMatrix")){
    kmat = as.matrix(K)
  } else {
    kmat = as.matrix(K)
  }
  cond1 = (is.matrix(kmat))
  cond2 = (nrow(K)==ncol(kmat))
  cond3 = isSymmetric(kmat)
  if (!(cond1&&cond2&&cond3)){
    stop("* mmd2test : 'K' should be a kernel matrix.")
  }
  mineval = min(base::eigen(kmat, only.values = TRUE)$values)
  if (mineval<0){
    wm = paste("* mmd2test : 'K' may not be PD. Minimum eigenvalue is ",mineval,".",sep="")
    warning(wm)
  }
  # 2. label
  label = as.vector(as.integer(as.factor(label)))
  if ((length(label)!=nrow(kmat))||(length(unique(label))!=2)){
    stop("* mmd2test : 'label' should be a vector of proper length with 2 classes.")
  }
  ulabel = unique(label)
  # 3. method
  allmmm = c("b","u")
  method = match.arg(tolower(method), allmmm)
  
  ###############################################
  # compute statistic
  id1 = which(label==ulabel[1]); m = length(id1)
  id2 = which(label==ulabel[2]); n = length(id2)
  thestat = switch(method,
                   "b" = mmd_biased(kmat[id1,id1], kmat[id2,id2], kmat[id1,id2]),
                   "u" = mmd_unbiased(kmat[id1,id1], kmat[id2,id2], kmat[id1,id2]))
  
  ###############################################
  # Iteration
  mciter = round(mc.iter)
  itervals = rep(0,mciter)
  for (i in 1:mciter){
    permuted = sample(m+n)
    tmpid1 = permuted[1:m]
    tmpid2 = permuted[(m+1):(m+n)]
    
    itervals[i] = switch(method,
                         "b" = mmd_biased(kmat[tmpid1,tmpid1], kmat[tmpid2,tmpid2], kmat[tmpid1,tmpid2]),
                         "u" = mmd_unbiased(kmat[tmpid1,tmpid1], kmat[tmpid2,tmpid2], kmat[tmpid1,tmpid2]))
  }
  pvalue = (sum(itervals>=thestat)+1)/(mciter+1)
  
  ###############################################
  # REPORT
  hname   = "Kernel Two-sample Test with Maximum Mean Discrepancy"
  Ha    = "two distributions are not equal"
  names(thestat) = "MMD"
  res   = list(statistic=thestat, p.value=pvalue, alternative = Ha, method=hname, data.name = DNAME)
  class(res) = "htest"
  return(res)
}



# compute two squared statistics ------------------------------------------
#' @keywords internal
#' @noRd
mmd_biased <- function(XX, YY, XY){
  # parameters
  m = nrow(XX)
  n = nrow(YY)
  
  # computation
  return((sum(XX)/(m^2)) + (sum(YY)/(n^2)) - ((2/(m*n))*sum(XY)))
}
#' @keywords internal
#' @noRd
mmd_unbiased <- function(XX, YY, XY){
  # parameters
  m = nrow(XX)
  n = nrow(YY)
  
  # computation
  term1 = (sum(XX)-sum(diag(XX)))/(m*(m-1))
  term2 = (sum(YY)-sum(diag(YY)))/(n*(n-1))
  term3 = (2/(m*n))*sum(XY)
  return((term1+term2-term3))
}

#' Trace Ratio Optimation
#' 
#' This function provides several algorithms to solve the following problem
#' \deqn{\textrm{max} \frac{tr(V^\top A V)}{tr(V^\top B V)} \textrm{ such that } V^\top C V = I}
#' where \eqn{V} is a projection matrix, i.e., \eqn{V^\top V = I}. Trace ratio optimization 
#' is pertained to various linear dimension reduction methods. It should be noted that 
#' when \eqn{C = I}, the above problem is often reformulated as a generalized eigenvalue problem 
#' since it's an easier proxy with faster computation.
#' 
#' @param A a \eqn{(p\times p)} symmetric matrix in the numerator term.
#' @param B a \eqn{(p\times p)} symmetric matrix in the denomiator term.
#' @param C a \eqn{(p\times p)} symmetric constraint matrix. If not provided, it is set as identical matrix automatically.
#' @param dim an integer for target dimension. It can be considered as the number of loadings.
#' @param method the name of algorithm to be used. Default is \code{2003Guo}.
#' @param maxiter maximum number of iterations to be performed.
#' @param eps stopping criterion for iterative algorithms. 
#' 
#' @return a named list containing
#' \describe{
#' \item{V}{a \eqn{(p\times dim)} projection matrix.}
#' \item{tr.val}{an attained maximum scalar value.}
#' }
#' 
#' @examples 
#' ## simple test
#' #  problem setting
#' p = 5
#' mydim = 2
#' A = matrix(rnorm(p^2),nrow=p); A=A%*%t(A)
#' B = matrix(runif(p^2),nrow=p); B=B%*%t(B)
#' C = diag(p)
#' 
#' #  approximate solution via determinant ratio problem formulation
#' eigAB  = eigen(solve(B,A)) 
#' V      = eigAB$vectors[,1:mydim]
#' eigval = sum(diag(t(V)%*%A%*%V))/sum(diag(t(V)%*%B%*%V))
#' 
#' #  solve using 4 algorithms
#' m12 = trio(A,B,dim=mydim, method="2012Ngo")
#' m09 = trio(A,B,dim=mydim, method="2009Jia")
#' m07 = trio(A,B,dim=mydim, method="2007Wang")
#' m03 = trio(A,B,dim=mydim, method="2003Guo")
#' 
#' #  print the results
#' line1 = '* Evaluation of the cost function'
#' line2 = paste("* approx. via determinant : ",eigval,sep="")
#' line3 = paste("* trio by 2012Ngo         : ",m12$tr.val, sep="")
#' line4 = paste("* trio by 2009Jia         : ",m09$tr.val, sep="")
#' line5 = paste("* trio by 2007Wang        : ",m07$tr.val, sep="")
#' line6 = paste("* trio by 2003Guo         : ",m03$tr.val, sep="")
#' cat(line1,"\n",line2,"\n",line3,"\n",line4,"\n",line5,"\n",line6)
#' 
#' @references 
#' \insertRef{guo_generalized_2003}{maotai}
#' 
#' \insertRef{wang_trace_2007}{maotai}
#' 
#' \insertRef{yangqingjia_trace_2009}{maotai}
#' 
#' \insertRef{ngo_trace_2012}{maotai}
#' 
#' @export
trio <- function(A, B, C, dim=2, method=c("2003Guo","2007Wang","2009Jia","2012Ngo"), maxiter=1000, eps=1e-10){
  ###################################################################
  # not completed yet.
  if (missing(C)){
    C = diag(nrow(A))
    myflag = TRUE
  } else {
    myflag = FALSE
  }
  
  if (!check_symm(A)){
    stop("* trio : an input matrix 'A' should be a square, symmetric matrix.")
  }
  if (!check_symm(B)){
    stop("* trio : an input matrix 'B' should be a square, symmetric matrix.")
  }
  if (!check_symm(C)){
    stop("* trio : an input matrix 'C' should be a square, symmetric matrix.")
  }
  sizes = rep(0,3)
  sizes[1] = nrow(A)
  sizes[2] = nrow(B)
  sizes[3] = nrow(C)
  
  if (length(unique(sizes))!=1){
    stop("* trio : all input matrices should be of same size.")
  }
  
  if (!myflag){
    eigC  = eigen(C)
    Cinv2 = eigC$vectors%*%diag(1/sqrt(eigC$values))%*%t(eigC$vectors)
    A = Cinv2%*%A%*%Cinv2
    B = Cinv2%*%B%*%Cinv2  
  }
  
  # 2009 Jia's note : B should have rank >= (m-d)
  if (as.integer(Matrix::rankMatrix(B))<(nrow(B)-dim)){
    warning("* trio : null space of 'B' is excessive. trace ratio value may diverge.")
  }
  
  
  ###################################################################
  # switch case
  V = switch(method,
             "2007Wang" = trio2007Wang(A, B, dim, eps, maxiter),
             "2003Guo"  = trio2003Guo(A, B, dim, eps, maxiter),
             "2009Jia"  = trio2009Jia(A, B, dim, eps, maxiter),
             "2012Ngo"  = trio2012Ngo(A, B, dim, eps, maxiter))
  output = list()
  output$V = V
  output$tr.val = sum(diag(t(V)%*%A%*%V))/sum(diag(t(V)%*%B%*%V))
  return(output)
}



# subroutines -------------------------------------------------------------
#' 2003 Guo et al.
#' Title : A generalized Foley-Sammon transform based on generalized fisher discriminant ...
#' 
#' @keywords internal
#' @noRd
trio2003Guo <- function(A, B, dim, eps, maxiter){
  ## translate into the language
  d = dim
  Sp = A
  Sl = B
  
  ## bisection 
  #   1. initialization
  lbd1 = 0; f1 = evalGuoDiff(lbd1, Sp, Sl, d)
  lbd2 = 1; f2 = evalGuoDiff(lbd2, Sp, Sl, d)
  if (f2 >= 0){
    while (f2 > 0){
      lbd1 = lbd2; f1 = f2;
      lbd2 = lbd2*2; f2 = evalGuoDiff(lbd2, Sp, Sl, d)  
    }
  }
  
  for (i in 1:maxiter){
    lbdm = (lbd1+lbd2)/2
    fm   = evalGuoDiff(lbdm, Sp, Sl, d)
    if (fm > 0){
      lbd1 = lbdm
      f1   = fm
    } else {
      lbd2 = lbdm
      f2   = fm
    }
    if (abs(lbd1-lbd2) < eps){
      break
    }
  }
  lbdm = (lbd1+lbd2)/2
  # W = eigen(Sp-lbdm*Sl)$vectors[,1:d] ## use RSpectra for only top 'd' ones
  W = RSpectra::eigs(Sp-lbdm*Sl,d,which="LR")$vectors
  
  ## let's try to return !
  return(W)
}


#' @keywords internal
#' @noRd
evalGuoDiff <- function(lbd, A, B, dim){
  W = RSpectra::eigs(A-lbd*B,dim,which="LR")$vectors
  # W = eigen(A-lbd*B)$vectors[,1:dim] ## use RSpectra for only top 'd' ones
  return(sum(diag(t(W)%*%(A-lbd*B)%*%W)))
}

#' 2007 Wang et al.
#' Title : Trace Ratio vs. Ratio Trace for Dimensionality Reduction
#' 
#' @keywords internal
#' @noRd
trio2007Wang <- function(A, B, dim, eps, maxiter){
  ## translate into this language
  Sp = A
  St = A+B
  m = nrow(A)
  d = dim 
  
  eigSt = base::eigen(St, symmetric = TRUE)
  mm = sum(eigSt$values > 0)
  if (mm < 1){
    stop("* (A+B) has at least one nonnegative eigenvalues.")
  }
  U  = eigSt$vectors[,1:mm]
  
  ## transform into the problem of V now.
  Spu = t(U)%*%Sp%*%U
  Stu = t(U)%*%St%*%U
  Vold = qr.Q(qr(matrix(rnorm(mm*d),ncol=d)))
  
  ## main computation
  V = cppsub_2007Wang(Vold, mm, d, Spu, Stu, maxiter, eps)
  
  ## adjust back to the original problem
  W = U%*%V
  
  ## let's try to return !
  return(W)
}

#' 2009 Jia et al 
#' Title : Trace Ratio Problem Revisited (DNM)
#' 
#' @keywords internal
#' @noRd
trio2009Jia <- function(A, B, dim, eps, maxiter){
  ## translate into the language
  d = dim
  Sp = A
  Sl = B
  
  ## Decomposed Newton Method
  lbdold = 0
  for (i in 1:maxiter){
    ## 1. compute eigendecomposition
    eigS = RSpectra::eigs(Sp-lbdold*Sl,d,which="LR")
    top.val = eigS$values   # top 'd' eigenvalues
    top.vec = eigS$vectors  # top 'd' eigenvectors
    
    ## 2. lbdnew
    lbdnew = solve2009Jia(lbdold, top.val, top.vec, Sl)
    inc    = abs(lbdnew-lbdold)
    
    ## 3. updating information
    lbdold = lbdnew
    if (inc<eps){
      break
    }
  }
  
  ## now lbdold is our solution
  W = RSpectra::eigs(Sp-lbdold*Sl,d,which="LR")$vectors
  
  ## let's try to return !
  return(W)
}


#' @keywords internal
#' @noRd
solve2009Jia <- function(lbd, eval, evec, Sl){
  d = length(eval)
  primesum = 0
  for (i in 1:d){
    tgt = as.vector(evec[,i])
    primesum = primesum - sum(as.vector(Sl%*%tgt)*tgt)
  }
  sol = (lbd*primesum - sum(eval))/primesum
  return(sol)
}

#' 2012 Ngo et al 
#' Title : The Trace Ratio Optimization Problem (Newton-Lanczos Algorithm)
#' 
#' 
#' @keywords internal
#' @noRd
trio2012Ngo <- function(A, B, dim, eps, maxiter){
  ## in the the language
  n = nrow(A)
  p = dim
  
  ## prepare the initializer
  Vold = qr.Q(qr(matrix(rnorm(n*p),ncol=p)))
  rhoold = 0
  for (i in 1:maxiter){
    Vnew   = RSpectra::eigs(A-rhoold*B,p,which="LR")$vectors
    rhonew = sum(diag(t(Vnew)%*%A%*%Vnew))/sum(diag(t(Vnew)%*%B%*%Vnew))
    
    rhoinc = abs(rhonew-rhoold)
    Vold   = Vnew
    rhoold = rhonew
    
    if (rhoinc < eps){
      break
    }
  }
  
  ## let's try to return !
  return(Vold)
}


# # 
# # simple test
# p = 100
# mydim = 10
# A = matrix(rnorm(p^2),nrow=p); A=A%*%t(A)
# B = matrix(runif(p^2),nrow=p); B=B%*%t(B)
# C = diag(p)
# 
# 
# library(geigen)
# eigAB = eigen(solve(B,A)) ## geigen(B,A) # we need largest so be careful of the order
# mylist = list()
# V = eigAB$vectors[,1:mydim]; mylist$V = V
# myval = sum(diag(t(V)%*%A%*%V))/sum(diag(t(V)%*%B%*%V)); mylist$tr.val = myval
# 
# m12 = trio(A,B,dim=mydim, method="2012Ngo")
# m9 = trio(A,B,dim=mydim, method="2009Jia")
# m7 = trio(A,B,dim=mydim, method="2007Wang")
# m3 = trio(A,B,dim=mydim, method="2003Guo")
# 
#  # try rstiefel's optimization
# library(rstiefel)
# f1 = function(w){
#   trA = sum(diag(t(w)%*%A%*%w)); trB = sum(diag(t(w)%*%B%*%w));
#   return(-trA/trB)
# }
# df1 = function(w){
#   trA = sum(diag(t(w)%*%A%*%w)); trB = sum(diag(t(w)%*%B%*%w));
#   t1 = -2*(A%*%w)*trB + trA*2*(B%*%w);
#   t2 = trB^2;
#   return(t1/t2)
# }
# f2 = function(w){
#   trA = sum(diag(t(w)%*%A%*%w)); trB = sum(diag(t(w)%*%B%*%w));
#   return(trB/trA)
# }
# df2 = function(w){
#   trA = sum(diag(t(w)%*%A%*%w)); trB = sum(diag(t(w)%*%B%*%w));
#   t1 = 2*trA*(B%*%w) - 2*trB*(A%*%w);
#   t2 = trA^2;
#   return(t1/t2)
# }
# V1 <- optStiefel(f1, df1, Vinit=rustiefel(p,mydim), maxIters = 999, tol=1e-10, verbose=TRUE)
# V2 <- optStiefel(f2, df2, Vinit=rustiefel(p,mydim), maxIters = 999, tol=1e-10, verbose=TRUE)
# print(sprintf("result of (2) inverse  : %f", 1/f2(V2)))

#' Convert 'movMF' object
#' 
#' Given an output from the movMF package's movMF function, 
#' convert them into the standard mixture parameter format.
#' 
#' @param movMF_object a movMF object of \eqn{K} components in \eqn{d} dimensions.
#'
#' @return a named list containing \describe{
#' \item{means}{a \eqn{(K \times d)} matrix of means}
#' \item{concentrations}{a \eqn{K} vector of concentrations}
#' \item{weights}{a \eqn{K} vector of weights}
#' }
#' 
#' @export
movMF_convert <- function(movMF_object){
  ###############################################
  # Preprocessing
  if (!inherits(movMF_object, "movMF")){
    stop("* movMF_convert : Input object is not a 'movMF' object")
  }
  
  ###############################################
  # Change
  output = list()
  output$means <- movMF_object$theta/sqrt(rowSums(movMF_object$theta^2))
  output$concentrations <- sqrt(rowSums(movMF_object$theta^2))
  output$weights <- movMF_object$alpha
  return(output)
}

#' Find Shortest Path using Floyd-Warshall Algorithm
#'
#' This is a fast implementation of Floyd-Warshall algorithm to find the
#' shortest path in a pairwise sense using \code{RcppArmadillo}. A logical input
#' is also accepted. The given matrix should contain pairwise distance values \eqn{d_{i,j}} where 
#' \eqn{0} means there exists no path for node \eqn{i} and \eqn{j}.
#'
#' @param dist either an \eqn{(n\times n)} matrix or a \code{dist} class object.
#' 
#' @return an \eqn{(n\times n)} matrix containing pairwise shortest path length.
#' 
#' @examples
#' ## simple example : a ring graph
#' #  edges exist for pairs
#' A = array(0,c(10,10))
#' for (i in 1:9){
#'   A[i,i+1] = 1
#'   A[i+1,i] = 1
#' }
#' A[10,1] <- A[1,10] <- 1
#' 
#' # compute shortest-path and show the matrix
#' sdA <- shortestpath(A)
#' 
#' # visualize
#' opar <- par(no.readonly=TRUE)
#' par(pty="s")
#' image(sdA, main="shortest path length for a ring graph")
#' par(opar)
#'
#' @references 
#' \insertRef{floyd_algorithm_1962}{maotai}
#' 
#' \insertRef{warshall_theorem_1962}{maotai}
#' 
#' @export
shortestpath <- function(dist){
  input = dist
  # class determination
  if (inherits(dist,"matrix")){
    distnaive = dist
  } else if (inherits(dist, "dist")){
    distnaive = as.matrix(dist)
  } else {
    stop("* shortestpath : input 'dist' should be either (n*n) matrix or 'dist' class object.")
  }
   
  # consider logical input
  if (any(is.logical(distnaive))){
    distnaive = distnaive*1
  }
  # set as -Inf for 0 values
  mepsil  = .Machine$double.eps
  distnaive[which(distnaive<5*mepsil)] = -Inf
  distgeo   = aux_shortestpath(distnaive)
  return(distgeo)
}

#' Geometric Median of vMF Distributions Under a Wasserstein-Like Geometry
#' 
#' Given a collection of von Mises-Fisher (vMF) distributions, each characterized 
#' by a mean direction \eqn{\mathbf{\mu}} and a concentration parameter \eqn{\kappa},
#' this function solves the geometric median problem to compute the vMF 
#' distribution that minimizes the weighted sum of distances under an approximate Wasserstein geometry.
#' 
#' @param means An \eqn{(n \times p)} matrix where each row represents the mean 
#' direction of one of the \eqn{n} vMF distributions.
#' @param concentrations A length-\eqn{n} vector of nonnegative concentration parameters.
#' @param weights A weight vector of length \eqn{n}. If \code{NULL}, equal weights 
#' (\code{rep(1/n, n)}) are used.
#' 
#' @return A named list containing:
#' \describe{
#'   \item{mean}{A length-\eqn{p} vector representing the median direction.}
#'   \item{concentration}{A scalar representing the median concentration.}
#' }
#' 
#' @examples
#' \donttest{
#' # Set seed for reproducibility
#' set.seed(123)
#' 
#' # Number of vMF distributions
#' n <- 5   
#' 
#' # Generate mean directions concentrated around a specific angle (e.g., 45 degrees)
#' base_angle <- pi / 4  # 45 degrees in radians
#' angles <- rnorm(n, mean = base_angle, sd = pi / 20)  # Small deviation from base_angle
#' means <- cbind(cos(angles), sin(angles))  # Convert angles to unit vectors
#' 
#' # Generate concentration parameters with large magnitudes (tight distributions)
#' concentrations <- rnorm(n, mean = 50, sd = 5)  # Large values around 50
#' 
#' # Compute the median under the Wasserstein-like geometry
#' barycenter <- WLmedian(means, concentrations)
#' 
#' # Convert median mean direction to an angle
#' bary_angle <- atan2(barycenter$mean[2], barycenter$mean[1])
#' 
#' ## Visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,2), pty="s")
#' 
#' # Plot the unit circle
#' plot(cos(seq(0, 2 * pi, length.out = 200)), sin(seq(0, 2 * pi, length.out = 200)), 
#'      type = "l", col = "gray", lwd = 2, xlab = "x", ylab = "y", 
#'      main = "Median of vMF Distributions on S^1")
#' 
#' # Add input mean directions
#' points(means[,1], means[,2], col = "blue", pch = 19, cex = 1.5)
#' 
#' # Add the computed barycenter
#' points(cos(bary_angle), sin(bary_angle), col = "red", pch = 17, cex = 2)
#' 
#' # Add legend
#' legend("bottomleft", legend = c("vMF Means", "Median"), col = c("blue", "red"), 
#'        pch = c(19, 17), cex = 1)
#' 
#' # Plot the concentration parameters
#' hist(concentrations, main = "Concentration Parameters", xlab = "Concentration")
#' abline(v=barycenter$concentration, col="red", lwd=2)
#' par(opar)
#' }
#' 
#' 
#' @export
WLmedian <- function(means, concentrations, weights=NULL){
  ###############################################
  # Preprocessing
  # means
  if (!is.matrix(means)){
    data_X = cpp_WL_normalise(as.matrix(means))
  } else {
    data_X = cpp_WL_normalise(means)
  }
  
  # concentrations
  data_kappa = as.vector(concentrations)
  if (length(data_kappa)!=base::nrow(data_X)){
    stop("* WLmedian : cardinalities of the means and concentrations do not match.")
  }
  
  # weights
  if ((length(weights)==0)&&(is.null(weights))){
    data_weights = rep(1/length(data_kappa), length(data_kappa))
  } else {
    data_weights = as.vector(weights)
    data_weights = data_weights/base::sum(data_weights)
  }
  if (any(data_weights < .Machine$double.eps)||(length(data_weights)!=length(data_kappa))){
    stop("* WLmedian : invalid 'weights'. Please see the documentation.")
  }
  
  
  ###############################################
  # COMPUTATION
  # initialize
  old_mean  = as.vector(cpp_WL_weighted_mean(data_X, data_weights))
  old_kappa = base::mean(data_kappa) 
  
  # iterate
  par_nsample = base::nrow(data_X)
  par_dim     = base::ncol(data_X)
  for (it in 1:100){
    # compute the new weights
    now_weights = rep(0, par_nsample)
    for (i in 1:par_nsample){
      delta <- old_mean - as.vector(data_X[i, ])
      c <- sqrt(sum(delta^2))                # chord length
      term1 <- (2 * asin(pmin(1, c/2)))^2    # squared geodesic
      term2 = (par_dim)*((1/sqrt(old_kappa) - 1/sqrt(data_kappa[i]))^2)
      dist_old2sam   = sqrt(term1+term2)
      now_weights[i] = data_weights[i]/dist_old2sam 
    }
    now_weights = now_weights/base::sum(now_weights)
    
    # compute the new mean and kappa
    new_mean  = as.vector(cpp_WL_weighted_mean(data_X, now_weights))
    new_w_sum = base::sum(now_weights)
    new_kappa_tmp = base::sum(now_weights/sqrt(data_kappa))
    new_kappa = (new_w_sum/new_kappa_tmp)^2
    
    # update
    inc_mean  = sqrt(sum((old_mean-new_mean)^2))
    inc_kappa = abs(old_kappa-new_kappa)
    inc_max   = max(inc_mean, inc_kappa)
    
    old_mean = new_mean
    old_kappa = new_kappa
    if (inc_max < 1e-8){
      break
    }
  }
  
  ###############################################
  # Return
  output = list(mean=old_mean,
                concentration=old_kappa)
  return(output)
}

#' K-Means++ Clustering Algorithm
#' 
#' \eqn{k}-means++ algorithm is known to be a smart, careful initialization 
#' technique. It is originally intended to return a set of \eqn{k} points 
#' as initial centers though it can still be used as a rough clustering algorithm 
#' by assigning points to the nearest points.
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' @param k the number of clusters.
#' 
#' @return a length-\eqn{n} vector of class labels.
#' 
#' @examples 
#' ## use simple example of iris dataset
#' data(iris) 
#' mydata = as.matrix(iris[,1:4])
#' mycol  = as.factor(iris[,5])
#' 
#' ## find the low-dimensional embedding for visualization
#' my2d = cmds(mydata, ndim=2)$embed
#' 
#' ## apply 'kmeanspp' with different numbers of k's.
#' k2 = kmeanspp(mydata, k=2)
#' k3 = kmeanspp(mydata, k=3)
#' k4 = kmeanspp(mydata, k=4)
#' k5 = kmeanspp(mydata, k=5)
#' k6 = kmeanspp(mydata, k=6)
#' 
#' ## visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(2,3))
#' plot(my2d, col=k2, main="k=2", pch=19, cex=0.5)
#' plot(my2d, col=k3, main="k=3", pch=19, cex=0.5)
#' plot(my2d, col=k4, main="k=4", pch=19, cex=0.5)
#' plot(my2d, col=k5, main="k=5", pch=19, cex=0.5)
#' plot(my2d, col=k6, main="k=6", pch=19, cex=0.5)
#' plot(my2d, col=mycol, main="true cluster", pch=19, cex=0.5)
#' par(opar)
#' 
#' @references 
#' \insertRef{arthur_kmeans_2007}{maotai}
#' 
#' @export
kmeanspp <- function(data, k=2){
  ############################################################
  # Preprocessing 
  if (!check_datamat(data)){
    stop("* kmeanspp : an input 'data' should be a matrix without any missing/infinite values.")
  }
  xdiss = stats::as.dist(cpp_pdist(data))
  myk   = round(k)
  
  ############################################################
  # Run and Return
  output = hidden_kmeanspp(xdiss,k=myk)$cluster
  return(output)
}

# Hidden Functions for Future Use
# these functions can be loaded using 'utils::getFromNamespace'
# by the command 'getFromNamespace("function_name","maotai");
# 
# Here, all the functions require 'diss' object from 'stats' package.
#
# 01. hidden_kmedoids      : PAM algorithm 
#     hidden_kmedoids_best : PAM algorithm + use Silhouette (maximum)
# 02. hidden_bmds          : Bayesian Multidimensional Scaling
# 03. hidden_cmds          : Classical Multidimensional Scaling
# 04. hidden_kmeanspp      : k-means++ algorithm.
# 05. hidden_tsne          : t-SNE visualization.
# 06. hidden_nem           : Negative Eigenvalue Magnitude
# 07. hidden_nef           : Negative Eigenfraction
# 08. hidden_emds          : Euclified Multidimensional Scaling
# 09. hidden_hclust        : FASTCLUSTER - hclust function
# 10. hidden_dbscan        : DBSCAN      - dbscan function
# 11. hidden_silhouette    : mimics that of cluster's silhouette
# 12. hidden_mmds          : metric multidimensional scaling by SMACOF 
# 13. hidden_PHATE         : return row-stochastic matrix & time stamp
# 14. hidden_smacof        : a generalized version of SMACOF with weights
# 15. hidden_hydra         : Hyperbolic Distance Recovery and Approximation\
# 16. hidden_metricdepth   : compute the metric depth



# 00. hidden_checker ------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_checker <- function(xobj){
  if (inherits(xobj, "dist")){
    return(as.matrix(xobj))
  } else if (inherits(xobj, "matrix")){
    check1 = (nrow(xobj)==ncol(xobj))
    check2 = isSymmetric(xobj)
    check3 = all(abs(diag(xobj))<.Machine$double.eps*10)
    if (check1&&check2&&check3){
      return(as.matrix(xobj))
    } else {
      stop("* hidden : matrix is not valid.")
    }
  } else {
    stop("* hidden : input is not valid.")
  }
}

# 01. hidden_kmedoids & hidden_kmedoids_best ------------------------------
#' @keywords internal
#' @noRd
hidden_kmedoids <- function(distobj, nclust=2){
  distobj = stats::as.dist(hidden_checker(distobj))
  myk     = round(nclust)
  return(cluster::pam(distobj, k = myk))
}
#' @keywords internal
#' @noRd
hidden_kmedoids_best <- function(distobj, mink=2, maxk=10){
  # prepare
  kvec = seq(from=round(mink),to=round(maxk), by = 1)
  knum = length(kvec)
  svec = rep(0,knum)
  nobj = nrow(as.matrix(distobj))
  clut = array(0,c(nobj,knum))
  
  for (k in 1:knum){
    know = kvec[k]
    if (know < 2){
      svec[k]  = 0
      clut[,k] = rep(1,nobj)
    } else {
      pamx     = hidden_kmedoids(distobj, nclust=kvec[k])
      svec[k]  = pamx$silinfo$avg.width
      clut[,k] = pamx$clustering  
    }
  }
  # return the output
  output = list()
  output$opt.k = kvec[which.max(svec)]
  output$score = svec     # knum-vector of silhouette scores
  output$label = clut     # (n,knum) cluster labels
  return(output)
}

# 02. hidden_bmds ---------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_bmds <- function(x, ndim=2, par.a=5, par.alpha=0.5, par.step=1, mc.iter=8128, verbose=FALSE){
  ######################################################
  # Initialization
  ndim   = round(ndim)
  embedy = hidden_cmds(x, ndim)$embed
  x      = as.matrix(x)
  ndim   = round(ndim)
  
  if ((length(ndim)>1)||(ndim<1)||(ndim>=nrow(x))){
    stop("* bmds - 'ndim' should be an integer in [1,ncol(data)). ")
  }
  
  n = nrow(x)
  m = n*(n-1)/2
  
  ######################################################
  # Preliminary Computation
  # 1. apply CMDS for initialization
  y     = as.matrix(base::scale(embedy,       # (N x ndim) centered 
                                center=TRUE, scale=FALSE)) 
  Delta = as.matrix(stats::dist(y))           # (N x N) pairwise distances
  
  # 2. initialization
  eigy   = base::eigen(stats::cov(y))       
  X0     = y%*%eigy$vectors         # (N x ndim) rotated
  gamma0 = diag(X0)                 # variances ?
  sigg0  = compute_SSR(x, Delta)/m;  
  beta0  = apply(X0,2,var)/2
  
  # 3. run the main part
  runcpp <- main_bmds(x, X0, sigg0, par.a, par.alpha, mc.iter, par.step, verbose, beta0)
  Xsol   <- runcpp$solX
  Xdist  <- as.matrix(stats::dist(Xsol))
  
  output = list()
  output$embed  = Xsol
  output$stress = compute_stress(x, Xdist)
  return(output)
  # return Rcpp::List::create(Rcpp::Named("solX")=Xsol,Rcpp::Named("solSSR")=SSRsol);
}


# 03. hidden_cmds ---------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_cmds <- function(x, ndim=2){
  ##################################################3
  # Check Input and Transform
  x    = hidden_checker(x)
  ndim = round(ndim)
  k  = as.integer(ndim)
  D2 = (x^2)          # now squared matrix
  n  = nrow(D2)
  if ((length(ndim)>1)||(ndim<1)||(ndim>=nrow(x))){
    stop("* cmds : 'ndim' should be an integer in [1,ncol(data)). ")
  }
  
  ##################################################3
  # Computation
  J = diag(n) - (1/n)*outer(rep(1,n),rep(1,n))
  B = -0.5*J%*%D2%*%J
  eigB = eigen(B)
  
  LL = eigB$values[1:k]
  EE = eigB$vectors[,1:k]
  
  # Y  = as.matrix(base::scale((EE%*%diag(sqrt(LL))), center=TRUE, scale=FALSE))
  Y  = EE%*%diag(sqrt(LL))
  DY = as.matrix(stats::dist(Y))
  
  output = list()
  output$embed  = Y
  output$stress = compute_stress(x, DY)
  return(output)
}

# 04. hidden_kmeanspp -----------------------------------------------------
#' @keywords internal
#' @noRd
hidden_kmeanspp <- function(x, k=2){
  ##################################################3
  # Check Input and Transform
  x  = hidden_checker(x)
  n  = nrow(x)
  K  = round(k)
  if (K >= n){
    stop("* kmeanspp : 'k' should be smaller than the number of observations.")
  }
  if (K < 2){
    stop("* kmeanspp : 'k' should be larger than 1.")
  }
  id.now = 1:n
  
  ##################################################3
  # Computation
  #   initialize
  id.center = base::sample(id.now, 1)
  id.now    = base::setdiff(id.now, id.center)
  #   iterate
  for (i in 1:(K-1)){
    # compute distance to the nearest
    tmpdmat = x[id.now, id.center]
    if (i==1){
      d2vec = as.vector(tmpdmat)^2
      d2vec = d2vec/base::sum(d2vec)
    } else {
      d2vec = as.vector(base::apply(tmpdmat, 1, base::min))^2
      d2vec = d2vec/base::sum(d2vec)
    }
    # sample one
    id.tmp = base::sample(id.now, 1, prob=d2vec)
    # update
    id.center = c(id.center, id.tmp)
    id.now    = base::setdiff(id.now, id.tmp)
  }
  #   let's compute label
  dmat    = x[,id.center]
  cluster = base::apply(dmat, 1, base::which.min)
  
  ##################################################
  # Return
  output = list()
  output$center  = id.center
  output$cluster = cluster
  return(output)
}


# 05. hidden_tsne ---------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_tsne <- function(dx, ndim=2, ...){
  ##################################################
  # Pass to 'Rtsne'
  dx     = hidden_checker(dx)
  k      = round(ndim)
  tmpout = Rtsne::Rtsne(dx, dims=k, ..., is_distance=TRUE)
  Y   = tmpout$Y
  DY  = as.matrix(stats::dist(Y))
  
  ##################################################
  # Return
  output = list()
  output$embed  = Y
  output$stress = compute_stress(dx, DY)
  return(output)
}

# 06. hidden_nem ----------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_nem <- function(xdiss){
  ##################################################3
  # Check Input and Transform
  xx = hidden_checker(xdiss)
  D2 = (xx^2)
  n  = nrow(D2)

  ##################################################3
  # Computation
  H = diag(n) - (1/n)*base::outer(rep(1,n),rep(1,n))
  S = -0.5*(H%*%D2%*%H)
  eigS = base::eigen(S, only.values = TRUE)
  evals = eigS$values

  ##################################################3
  # Finalize
  output = abs(min(evals))/max(evals)
  return(output)
}


# 07. hidden_nef ----------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_nef <- function(xdiss){
  ##################################################3
  # Check Input and Transform
  xx = hidden_checker(xdiss)
  D2 = (xx^2)
  n  = nrow(D2)
  
  ##################################################3
  # Computation
  H = diag(n) - (1/n)*base::outer(rep(1,n),rep(1,n))
  S = -0.5*(H%*%D2%*%H)
  eigS = base::eigen(S)
  evals = eigS$values
  
  ##################################################3
  # Finalize
  output = sum(abs(evals[which(evals<0)]))/sum(abs(evals))
  return(output)
}

# 08. hidden_emds ---------------------------------------------------------
#' Euclified Multidimensional Scaling
#' 
#' strategy 1 : transitive closure of the triangle inequality (labdsv)
#' strategy 2 : Non-Euclidean or Non-metric Measures Can Be Informative; adding positive numbers to all off-diagonal entries
#' 
#' @keywords internal
#' @noRd
hidden_emds <- function(xdiss, ndim=2, method=c("closure","gram")){
  ##################################################3
  # Check Input and Transform
  x    = hidden_checker(xdiss)
  ndim = round(ndim)
  k  = as.integer(ndim)
  n  = nrow(x)
  if ((length(ndim)>1)||(ndim<1)||(ndim>=nrow(x))){
    stop("* emds : 'ndim' should be an integer in [1,nrow(x)). ")
  }
  
  method = match.arg(method) 
  mydim  = round(ndim)
  
  ##################################################3
  # Branching
  if (hidden_nef(x) < 100*.Machine$double.eps){ # if Euclidean, okay
    output = hidden_cmds(x, ndim=mydim)
  } else { # if not Euclidean
    if (method=="closure"){ # strategy 1 : transitive closure of the triangle inequality
      xnew   = as.matrix(labdsv::euclidify(stats::as.dist(x))) # well it seems to work well..
      output = hidden_cmds(xnew, ndim = mydim)
    } else {                # strategy 2 : add positive numbers to all off-diagonal entries
      gamma0 = emds_gamma0(x)
      ggrid  = seq(from=min(0.001, gamma0/1000), to=(gamma0*0.999), length.out=20) # just try 20 cases
      vgrid  = rep(0,20)
      for (i in 1:20){
        xtmp = x + ggrid[i]
        diag(xtmp) = rep(0,nrow(xtmp))
        vgrid[i]   = hidden_nef(xtmp)
      }
      idopts = which.min(vgrid)
      if (length(idopts)>1){ # if multiple, use the first one.
        idopts = idopts[1]
      }
      optgamma   = ggrid[idopts]
      xnew       = x + optgamma
      diag(xnew) = rep(0,nrow(xnew))
      output     = hidden_cmds(xnew, ndim = mydim)
    } 
  }
  
  ##################################################3
  # Report 
  return(output)
}

# 09. hidden_hclust -------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_hclust <- function(xdiss, mymethod, mymembers){
  return(fastcluster::hclust(xdiss, method=mymethod,
                             members=mymembers))
}

# 10. hidden_dbscan -------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_dbscan <- function(Xdiss, myeps, myminPts=5, ...){
  output = dbscan::dbscan(Xdiss, eps = myeps, minPts=myminPts, ...)
  return(output)
}

# 11. hidden_silhouette --------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_silhouette <- function(xdiss, label){
  x    = as.integer(as.factor(label))
  hsil = cluster::silhouette(x, xdiss)
  
  output = list()
  output$local  = as.vector(hsil[,3])
  output$global = base::mean(as.vector(hsil[,3]))
  return(output)
}

# 12. hidden_mmds          : metric multidimensional scaling by SMACOF  --------
#     note that from version 0.2.2, I"m using a version from the modern MDS book.
#' @keywords internal
#' @noRd
hidden_mmds <- function(x, ndim=2, maxiter=200, abstol=1e-5){
  # Check Input and Transform
  x    = hidden_checker(x)
  n    = base::nrow(x)
  ndim = round(ndim)
  myiter = max(50, round(maxiter))
  mytol  = max(100*.Machine$double.eps, as.double(abstol))
  
  WW = array(1,c(n,n))
  return(as.matrix(src_smacof(x, WW, ndim, myiter, mytol, TRUE)$embed))
  # # Run with Rcpp
  # return(cpp_mmds(x, ndim, myiter, mytol))
}



# 13. hidden_PHATE --------------------------------------------------------
#' @keywords internal
#' @noRd
hidden_PHATE <- function(x, nbdk=5, alpha=2){
  # Check Input and Transform
  x = hidden_checker(x) # now it's a matrix
  n = base::nrow(x)
  nbdk  = max(1, round(nbdk))
  alpha = max(sqrt(.Machine$double.eps), as.double(alpha))
  
  # k-th nearest distance
  nndist = rep(0,n)
  for (i in 1:n){
    tgt = as.vector(x[i,])
    nndist[i] = tgt[order(tgt)[nbdk+1]]
  }
  
  # Build Kernel Matrix
  matK = array(1,c(n,n))
  for (i in 1:(n-1)){
    for (j in (i+1):n){
      term1 = exp(-((x[i,j]/nndist[i])^(alpha)))
      term2 = exp(-((x[i,j]/nndist[j])^(alpha)))
      matK[i,j] <- matK[j,i] <- 0.5*(term1+term2)
    }
  }
  vecD = base::rowSums(matK)
  matP = base::diag(1/vecD)%*%matK
  matA = base::diag(1/base::sqrt(vecD))%*%matK%*%base::diag(1/base::sqrt(vecD))
  
  # Eigenvalues and Von-Neumann Entropy
  eigA  = eigen(matA)$values
  eigA  = eigA[(eigA>0)]
  vec.t = 1:1000
  vec.H = rep(0,1000)
  for (i in 1:1000){
    eig.t  = (eigA^i) + (1e-7) # modified for zero-padding
    eig.t  = eig.t/base::sum(eig.t)
    term.t = -base::sum(eig.t*base::log(eig.t))
    if (is.na(term.t)){
      vec.t = vec.t[1:(i-1)]
      vec.H = vec.H[1:(i-1)]
      break
    } else {
      vec.H[i] = term.t
    }
  }
  
  # Optimal Stopping Criterion
  Pout  = matP
  opt.t = round(hidden_knee_clamped(vec.t, vec.H))
  for (i in 1:(opt.t-1)){
    Pout = Pout%*%matP
  }
  
  # return the output
  output = list()
  output$P = Pout
  output$t = opt.t
  return(output)
}

# X   = as.matrix(iris[,1:4])
# lab = as.factor(iris[,5])
# 
# D    = stats::dist(X)
# cmd2 = cmdscale(D, k=2)
# mmdA = hidden_mmds(D, ndim=2, abstol=1e-2)
# mmdB = hidden_mmds(D, ndim=2, abstol=1e-10)
# 
# par(mfrow=c(1,3), pty="s")
# plot(cmd2, col=lab, main = "cmds")
# plot(mmdA, col=lab, main="mmds-2")
# plot(mmdB, col=lab, main="mmds-8")


# # example -----------------------------------------------------------------
# library(labdsv)
# data(bryceveg) # returns a vegetation data.frame
# dis.bc <- as.matrix(dsvdis(bryceveg,'bray/curtis')) # calculate a Bray/Curtis
# 
# emds = getFromNamespace("hidden_emds","maotai")
# out.cmds <- cmds(dis.bc, ndim=2)$embed
# out.emds1 <- emds(dis.bc, ndim=2, method="closure")$embed
# out.emds2 <- emds(dis.bc, ndim=2, method="gram")$embed
# 
# par(mfrow=c(3,1))
# plot(out.cmds, main="cmds")
# plot(out.emds1, main="emds::closure")
# plot(out.emds2, main="emds::gram")



# 14. hidden_smacof        : a generalized version of SMACOF with weights ======
#     returns both {embed} and {stress}
#' @keywords internal
#' @noRd
hidden_smacof <- function(D, W=NULL, ndim=2, maxiter=100, abstol=(1e-7)){
  myiter  = round(maxiter)
  mytol   = as.double(abstol)
  myndim  = round(ndim)
  
  DD = hidden_checker(D) # now it's a matrix
  nn = base::nrow(DD)
  if (is.null(W)&&(length(W)==0)){
    use.gutman = TRUE
    WW = array(1,c(nn,nn))
  } else {
    use.gutmat = FALSE
    WW = as.matrix(W)
  }
  
  output = src_smacof(DD, WW, myndim, myiter, mytol, use.gutman)
  return(output)
}

# fun_cmds   <- utils::getFromNamespace("hidden_cmds", "maotai")
# fun_mmds   <- utils::getFromNamespace("hidden_mmds", "maotai")
# fun_smacof <- utils::getFromNamespace("hidden_smacof", "maotai")
# 
# D   = stats::dist(as.matrix(iris[,1:4]))
# lab = factor(iris[,5])
# 
# Yc = fun_cmds(D)$embed
# Ym = fun_mmds(D, maxiter=500)
# Ys = fun_smacof(D, maxiter=500)$embed
# 
# par(mfrow=c(1,3), pty="s")
# plot(Yc, col=lab, pch=19, main="CMDS")
# plot(Ym, col=lab, pch=19, main="MMDS")
# plot(Ys, col=lab, pch=19, main="smacof")




# 15. hidden_hydra --------------------------------------------------------
#     note that 'iso.adjust=TRUE' is a direct application of the algorithm
#     as stated in the paper. For 2-dimensional embedding, 'FALSE' is default 
#     in the 'hydra' package's implementation.
#' @keywords internal
#' @noRd
hidden_hydra <- function(distobj, ndim=2, kappa=1, iso.adjust=TRUE){
  # preprocess
  D     = as.matrix(distobj)
  n     = base::nrow(D)
  dim   = round(ndim)
  kappa = base::max(base::sqrt(.Machine$double.eps), as.double(kappa))
  A     = base::cosh(base::sqrt(kappa)*D)
  
  # eigen-decomposition : 100 dimensions
  if (n < 100){ 
    # regular 'base::eigen'
    eigA = base::eigen(A, TRUE)
    
    # top vector
    x0 = sqrt(eigA$values[1])*as.vector(eigA$vectors[,1])
    if (x0[1] < 0){
      x0 = -x0
    }
    
    # others
    bot_vals = eigA$values[(n-dim+1):n]
    bot_vecs = eigA$vectors[,(n-dim+1):n]
  } else {
    # or use 'RSpectra::eigs_sym'
    eig_top = RSpectra::eigs_sym(A, 1,    which="LA")
    eig_bot = RSpectra::eigs_sym(A, ndim, which="SA")
    
    # top vector
    x0 = sqrt(eig_top$values)*as.vector(eig_top$vectors)
    if (x0[1] < 0){
      x0 = -x0
    }
    # others
    bot_vecs = eig_bot$vectors
    bot_vals = eig_bot$values
  }
  
  # component : radial
  x_min  = base::min(x0)
  radial = sqrt((x0-x_min)/(x0+x_min))
  
  # component : directional
  if (iso.adjust){
    X_last      = bot_vecs%*%diag(sqrt(pmax(-bot_vals,0)))
    sqnorms     = base::apply(X_last, 1, function(x) 1/sqrt(sum(x^2)))
    directional = base::diag(sqnorms)%*%X_last
    
  } else {
    sqnorms     = base::apply(bot_vecs, 1, function(x) 1/sqrt(sum(x^2)))
    directional = base::diag(sqnorms)%*%bot_vecs
  }

  # return the output
  return(list(radial=radial, directional=directional))
}

# library(hydra)
# data(karate)
# D = as.dist(karate$distance)
# 
# X = as.matrix(iris[,1:4])
# D = stats::dist(X)
# Y = factor(iris[,5])
# 
# run_hydra = hydra(as.matrix(D), equi.adj =-1, alpha = 1, curvature = 5)
# my_hydra  = hidden_hydra(D, kappa=5, iso.adjust = FALSE)
# 
# X1 = cbind(run_hydra$r*cos(run_hydra$theta), run_hydra$r*sin(run_hydra$theta))
# X2 = diag(my_hydra$radial)%*%my_hydra$directional
# 
# par(mfrow=c(1,3), pty="s")
# plot(run_hydra)
# plot(X1, pch=19, xlim=c(-1,1), ylim=c(-1,1), col=Y, main="HYDRA package")
# plot(X2, pch=19, xlim=c(-1,1), ylim=c(-1,1), col=Y, main="my implementation")



# 16. hidden_metricdepth --------------------------------------------------
#' @keywords internal
#' @noRd
hidden_metricdepth <- function(distobj){
  # intermediate assignment
  D = as.matrix(distobj)
  # compute
  return(as.vector(cpp_metricdepth(D)))
}

#' Weiszfeld Algorithm for Computing L1-median
#' 
#' Geometric median, also known as L1-median, is a solution to the following problem
#' \deqn{\textrm{argmin} \sum_{i=1}^n \| x_i - y \|_2 }
#' for a given data \eqn{x_1,x_2,\ldots,x_n \in R^p}. 
#' 
#' @param X an \eqn{(n\times p)} matrix for \eqn{p}-dimensional signal. If vector is given, it is assumed that \eqn{p=1}.
#' @param weights \code{NULL} for equal weight \code{rep(1/n,n)}; otherwise, it has to be a vector of length \eqn{n}.
#' @param maxiter maximum number of iterations.
#' @param abstol stopping criterion 
#' 
#' @examples 
#' ## generate sin(x) data with noise for 100 replicates
#' set.seed(496)
#' t = seq(from=0,to=10,length.out=20)
#' X = array(0,c(100,20))
#' for (i in 1:100){
#'    X[i,] = sin(t) + stats::rnorm(20, sd=0.5)
#' }
#' 
#' ## compute L1-median and L2-mean
#' vecL2 = base::colMeans(X)
#' vecL1 = weiszfeld(X)
#' 
#' ## visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,3), pty="s")
#' matplot(t(X[1:5,]), type="l", main="5 generated data", ylim=c(-2,2))
#' plot(t, vecL2, type="l", col="blue", main="L2-mean",   ylim=c(-2,2))
#' plot(t, vecL1, type="l", col="red",  main="L1-median", ylim=c(-2,2))
#' par(opar)
#' 
#' @export
weiszfeld <- function(X, weights=NULL, maxiter=496, abstol=1e-6){
  ###############################################
  # Preprocessing
  if (is.vector(X)){
    X = matrix(X, ncol = 1)
  }
  n = nrow(X)
  d = ncol(X)
  
  if ((length(weights)==0)&&(is.null(weights))){
    weights = rep(1/n, n)
  }
  if ((!is.vector(weights))||(length(weights)!=n)){
    stop(paste0("* weiszfeld : 'weights' should be a vector of length ",n))
  }
  
  ###############################################
  # Prepare
  myiter = round(maxiter)
  mytol  = as.double(abstol)
  xinit  = as.vector(base::colMeans(X))
  epsnum = (100*.Machine$double.eps)
  
  ###############################################
  # Compute and Return
  return(as.vector(cpp_weiszfeld(X, mytol, myiter, xinit, weights, epsnum)))
}



# ## example from Gmedian:weiszfeld
# ## Robustness of the geometric median of n=3 points in dimension d=2.
# library(Gmedian)
# 
# ## Computation speed
# ## Simulated data - Brownian paths
# n <- 1e4
# d <- 50
# x <- matrix(rnorm(n*d,sd=1/sqrt(d)), n, d)
# x <- t(apply(x,1,cumsum))
# 
# library(microbenchmark)
# microbenchmark::microbenchmark(
#   "resG1" = as.vector(Weiszfeld(x)$median),
#   "resG2" = as.vector(Gmedian(x)),
#   "resWW" = as.vector(maotai::weiszfeld(x))  
# )

test_that("weiszfeld", {
  n = 20
  p = 10
  t = seq(from=0,to=10,length.out=p)
  
  X = array(0,c(n,p))
  for (i in 1:n){
     X[i,] = sin(t) + stats::rnorm(p, sd=0.5)
  }

  vecL1 = as.vector(weiszfeld(X))
  expect_equal(length(vecL1), p)
})

#' Convert Covariance into Partial Correlation Matrix
#' 
#' Given a covariance matrix, return a partial correlation matrix that has unit diagonals. 
#' We strictly impose (and check) whether the given input is a symmetric matrix 
#' of full-rank.
#' 
#' @param mat a \eqn{(p\times p)} covariance matrix.
#' 
#' @return a \eqn{(p\times p)} partial correlation matrix.
#' 
#' @examples 
#' \donttest{
#' # generate an empirical covariance scaled
#' prep_mat = stats::cov(matrix(rnorm(100*10),ncol=10))
#' prep_vec = diag(as.vector(stats::runif(10, max=5)))
#' prep_cov = prep_vec%*%prep_mat%*%prep_vec
#' 
#' # compute correlation and partial correlation matrices
#' prep_cor = cov2corr(prep_cov)
#' prep_par = cov2pcorr(prep_cov)
#' 
#' # visualize
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,3), pty="s")
#' image(prep_cov, axes=FALSE, main="covariance")
#' image(prep_cor, axes=FALSE, main="correlation")
#' image(prep_par, axes=FALSE, main="partial correlation")
#' par(opar)
#' }
#' 
#' @export
cov2pcorr <- function(mat){
  # checker
  if (!check_covariance(mat)){
    stop("* cov2corr : an input 'mat' is not a valid covariance matrix.")
  }
  return(src_cov2corr(mat))
}

library(maotai)

test_that("cov2corr works", {
  myp = 5
  myn = 50
  
  prep_mat = stats::cov(matrix(rnorm(myn*myp), ncol=myp))
  vec_corr = diag(maotai::cov2corr(prep_mat))
  
  expect_equal(vec_corr, rep(1, myp))
})

test_that("cov2pcorr works", {
  myp = 5
  myn = 50
  
  prep_mat = stats::cov(matrix(rnorm(myn*myp), ncol=myp))
  vec_pcor = diag(maotai::cov2pcorr(prep_mat))
  
  expect_equal(vec_pcor, rep(1, myp))
})

#' Tools for Matrix Algebra, Optimization and Inference
#' 
#' A matrix is an universal and sometimes primary object/unit in applied mathematics and statistics. We provide a number of algorithms for selected problems in optimization and statistical inference.
#' 
#' @keywords internal
#' @name package-maotai
#' @import Rdpack
#' @noRd
#' @importFrom dbscan dbscan
#' @importFrom fastcluster hclust
#' @importFrom RANN nn2
#' @importFrom cluster pam silhouette agnes
#' @importFrom stats as.dist knots ecdf rnorm runif quantile dist rgamma rgeom var cov lm
#' @importFrom Rtsne Rtsne
#' @importFrom pracma cross detrend
#' @importFrom utils packageVersion
#' @importFrom RSpectra eigs
#' @importFrom Matrix rankMatrix
#' @importFrom Rcpp evalCpp
#' @importFrom gsignal hilbert butter filtfilt
#' @useDynLib maotai
"_PACKAGE"
# pack <- "maotai"
# path <- find.package(pack)
# system(paste(shQuote(file.path(R.home("bin"), "R")),
#              "CMD", "Rd2pdf", shQuote(path)))

#' Metric Depth
#' 
#' Compute the metric depth proposed by \insertCite{geenens_2023_StatisticalDepthAbstract;textual}{maotai}, which is 
#' one generalization of statistical depth function onto the arbitrary metric space. Our implementation assumes that 
#' given the multivariate data it computes the (empirical) depth for all observations using under the Euclidean regime.
#' 
#' @param data an \eqn{(n\times p)} matrix whose rows are observations.
#' @return a length-\eqn{n} vector of empirical metric depth values.
#' 
#' @examples 
#' \dontrun{
#' ## use simple example of iris dataset
#' data(iris) 
#' X <- as.matrix(iris[,1:4])
#' y <- as.factor(iris[,5])
#' 
#' ## compute the metric depth
#' mdX <- metricdepth(X)
#' 
#' ## visualize
#' #  2-d embedding for plotting by MDS
#' X2d <- maotai::cmds(X, ndim=2)$embed
#' 
#' # get a color code for the metric depth
#' pal = colorRampPalette(c("yellow","red"))
#' 
#' # draw
#' opar <- par(no.readonly=TRUE)
#' par(mfrow=c(1,2), pty="s")
#' plot(X2d, pch=19, main="by class", xlab="", ylab="", col=y)
#' plot(X2d, pch=19, main="by depth", xlab="", ylab="", col=pal(150)[order(mdX)])
#' legend("bottomright", col=pal(2), pch=19, legend=round(range(mdX), 2))
#' par(opar)
#' }
#' 
#' @references 
#' \insertAllCited{}
#' 
#' @export
metricdepth <- function(data){
  ## PREPROCESSING 
  if (!check_datamat(data)){
    stop("* metricdepth : an input 'data' should be a matrix without any missing/infinite values.")
  }
  xdiss = stats::as.dist(cpp_pdist(data))
  
  ## RUN AND RETURN
  output = hidden_metricdepth(xdiss)
  return(output)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rotationS2.R
\name{rotationS2}
\alias{rotationS2}
\title{Compute a Rotation on the 2-dimensional Sphere}
\usage{
rotationS2(x, y)
}
\arguments{
\item{x}{a length-\eqn{3} vector. If \eqn{\|x\|\neq 1}, normalization is applied.}

\item{y}{a length-\eqn{3} vector. If \eqn{\|y\|\neq 1}, normalization is applied.}
}
\value{
a \eqn{(3\times 3)} rotation matrix.
}
\description{
A vector of unit norm is an element on the hypersphere. When two unit-norm 
vectors \eqn{x} and \eqn{y} in 3-dimensional space are given, this function 
computes a rotation matrix \eqn{Q} on the 2-dimensional sphere such that
\deqn{y=Qx}.
}
\examples{
\donttest{
## generate two data points
#  one randomly and another on the north pole
x = stats::rnorm(3)
x = x/sqrt(sum(x^2))
y = c(0,0,1)

## compute the rotation
Q = rotationS2(x,y)

## compare 
Qx = as.vector(Q\%*\%x)

## print
printmat = rbind(Qx, y)
rownames(printmat) = c("rotated:", "target:")
print(printmat)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shortestpath.R
\name{shortestpath}
\alias{shortestpath}
\title{Find Shortest Path using Floyd-Warshall Algorithm}
\usage{
shortestpath(dist)
}
\arguments{
\item{dist}{either an \eqn{(n\times n)} matrix or a \code{dist} class object.}
}
\value{
an \eqn{(n\times n)} matrix containing pairwise shortest path length.
}
\description{
This is a fast implementation of Floyd-Warshall algorithm to find the
shortest path in a pairwise sense using \code{RcppArmadillo}. A logical input
is also accepted. The given matrix should contain pairwise distance values \eqn{d_{i,j}} where 
\eqn{0} means there exists no path for node \eqn{i} and \eqn{j}.
}
\examples{
## simple example : a ring graph
#  edges exist for pairs
A = array(0,c(10,10))
for (i in 1:9){
  A[i,i+1] = 1
  A[i+1,i] = 1
}
A[10,1] <- A[1,10] <- 1

# compute shortest-path and show the matrix
sdA <- shortestpath(A)

# visualize
opar <- par(no.readonly=TRUE)
par(pty="s")
image(sdA, main="shortest path length for a ring graph")
par(opar)

}
\references{
\insertRef{floyd_algorithm_1962}{maotai}

\insertRef{warshall_theorem_1962}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/movMF_convert.R
\name{movMF_convert}
\alias{movMF_convert}
\title{Convert 'movMF' object}
\usage{
movMF_convert(movMF_object)
}
\arguments{
\item{movMF_object}{a movMF object of \eqn{K} components in \eqn{d} dimensions.}
}
\value{
a named list containing \describe{
\item{means}{a \eqn{(K \times d)} matrix of means}
\item{concentrations}{a \eqn{K} vector of concentrations}
\item{weights}{a \eqn{K} vector of weights}
}
}
\description{
Given an output from the movMF package's movMF function, 
convert them into the standard mixture parameter format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WLmedian.R
\name{WLmedian}
\alias{WLmedian}
\title{Geometric Median of vMF Distributions Under a Wasserstein-Like Geometry}
\usage{
WLmedian(means, concentrations, weights = NULL)
}
\arguments{
\item{means}{An \eqn{(n \times p)} matrix where each row represents the mean 
direction of one of the \eqn{n} vMF distributions.}

\item{concentrations}{A length-\eqn{n} vector of nonnegative concentration parameters.}

\item{weights}{A weight vector of length \eqn{n}. If \code{NULL}, equal weights 
(\code{rep(1/n, n)}) are used.}
}
\value{
A named list containing:
\describe{
  \item{mean}{A length-\eqn{p} vector representing the median direction.}
  \item{concentration}{A scalar representing the median concentration.}
}
}
\description{
Given a collection of von Mises-Fisher (vMF) distributions, each characterized 
by a mean direction \eqn{\mathbf{\mu}} and a concentration parameter \eqn{\kappa},
this function solves the geometric median problem to compute the vMF 
distribution that minimizes the weighted sum of distances under an approximate Wasserstein geometry.
}
\examples{
\donttest{
# Set seed for reproducibility
set.seed(123)

# Number of vMF distributions
n <- 5   

# Generate mean directions concentrated around a specific angle (e.g., 45 degrees)
base_angle <- pi / 4  # 45 degrees in radians
angles <- rnorm(n, mean = base_angle, sd = pi / 20)  # Small deviation from base_angle
means <- cbind(cos(angles), sin(angles))  # Convert angles to unit vectors

# Generate concentration parameters with large magnitudes (tight distributions)
concentrations <- rnorm(n, mean = 50, sd = 5)  # Large values around 50

# Compute the median under the Wasserstein-like geometry
barycenter <- WLmedian(means, concentrations)

# Convert median mean direction to an angle
bary_angle <- atan2(barycenter$mean[2], barycenter$mean[1])

## Visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2), pty="s")

# Plot the unit circle
plot(cos(seq(0, 2 * pi, length.out = 200)), sin(seq(0, 2 * pi, length.out = 200)), 
     type = "l", col = "gray", lwd = 2, xlab = "x", ylab = "y", 
     main = "Median of vMF Distributions on S^1")

# Add input mean directions
points(means[,1], means[,2], col = "blue", pch = 19, cex = 1.5)

# Add the computed barycenter
points(cos(bary_angle), sin(bary_angle), col = "red", pch = 17, cex = 2)

# Add legend
legend("bottomleft", legend = c("vMF Means", "Median"), col = c("blue", "red"), 
       pch = c(19, 17), cex = 1)

# Plot the concentration parameters
hist(concentrations, main = "Concentration Parameters", xlab = "Concentration")
abline(v=barycenter$concentration, col="red", lwd=2)
par(opar)
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecdfdist.R
\name{ecdfdist}
\alias{ecdfdist}
\title{Distance Measures between Multiple Empirical Cumulative Distribution Functions}
\usage{
ecdfdist(
  elist,
  method = c("KS", "Lp", "Wasserstein"),
  p = 2,
  as.dist = FALSE,
  useR = TRUE
)
}
\arguments{
\item{elist}{a length \eqn{N} list of \code{ecdf} objects.}

\item{method}{name of the distance/dissimilarity measure. Case insensitive (default: \code{ks}).}

\item{p}{exponent for \code{Lp} or \code{Wasserstein} distance.}

\item{as.dist}{a logical; \code{TRUE} to return \code{dist} object, \code{FALSE} to return an \eqn{(N\times N)} symmetric matrix of pairwise distances (default: \code{FALSE}).}

\item{useR}{a logical; \code{TRUE} to use R implementation, \code{FALSE} to use C++ implementation (default: \code{TRUE}).}
}
\value{
either \code{dist} object of an \eqn{(N\times N)} symmetric matrix of pairwise distances by \code{as.dist} argument.
}
\description{
We measure distance between two empirical cumulative distribution functions (ECDF). For 
simplicity, we only take an input of \code{\link[stats]{ecdf}} objects from \pkg{stats} package.
}
\examples{
\donttest{
## toy example : 10 of random and uniform distributions
mylist = list()
for (i in 1:10){
  mylist[[i]] = stats::ecdf(stats::rnorm(50, sd=2))
}
for (i in 11:20){
  mylist[[i]] = stats::ecdf(stats::runif(50, min=-5))
}

## compute Kolmogorov-Smirnov distance
dm = ecdfdist(mylist, method="KS")

## visualize
mks  =" KS distances of 2 Types"
opar = par(no.readonly=TRUE)
par(pty="s")
image(dm[,nrow(dm):1], axes=FALSE, main=mks)
par(opar)
}

}
\seealso{
\code{\link[stats]{ecdf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/movMF_reduce_partitional.R
\name{movMF_reduce_partitional}
\alias{movMF_reduce_partitional}
\title{von Mises-Fisher mixture model reduction - partitional method}
\usage{
movMF_reduce_partitional(
  means,
  concentrations,
  weights,
  target.num = 2,
  method = c("hclust", "kmedoids")
)
}
\arguments{
\item{means}{a \eqn{(K \times p)} matrix of means of the von Mises-Fisher components.}

\item{concentrations}{a \eqn{K} vector of concentrations of the von Mises-Fisher components.}

\item{weights}{a \eqn{K} vector of weights of the von Mises-Fisher components.}

\item{target.num}{a desired number of components after reduction. Default is 2.}

\item{method}{a clustering method to be used. Default is "hclust".}
}
\value{
a named list of the reduced mixture model containing \describe{
\item{means}{a \eqn{(\code{target.num} \times p)} matrix of means of the von Mises-Fisher components.}
\item{concentrations}{a \eqn{\code{target.num}} vector of concentrations of the von Mises-Fisher components.}
\item{weights}{a \eqn{\code{target.num}} vector of weights of the von Mises-Fisher components.}
}
}
\description{
When given parameters of the von Mises-Fisher mixture model, this function 
aims at mixture model reduction using a partitional method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boot.stationary.R
\name{boot.stationary}
\alias{boot.stationary}
\title{Generate Index for Stationary Bootstrapping}
\usage{
boot.stationary(N, p = 0.25)
}
\arguments{
\item{N}{the number of observations.}

\item{p}{parameter for geometric distribution with the size of each block.}
}
\value{
a vector of length \code{N} for moving block bootstrap sampling.
}
\description{
Assuming data being dependent with cardinality \code{N}, \code{boot.stationary} returns 
a vector of index that is used for stationary bootstrapping. To describe, starting points 
are drawn from uniform distribution over \code{1:N} and the size of each block is 
determined from geometric distribution with parameter \eqn{p}.
}
\examples{
\donttest{
## example : bootstrap confidence interval of mean and variances
vec.x = seq(from=0,to=10,length.out=100)
vec.y = sin(1.21*vec.x) + 2*cos(3.14*vec.x) + rnorm(100,sd=1.5)
data.mu  = mean(vec.y)
data.var = var(vec.y)

## apply stationary bootstrapping
nreps   = 50
vec.mu  = rep(0,nreps)
vec.var = rep(0,nreps)
for (i in 1:nreps){
   sample.id = boot.stationary(100)
   sample.y  = vec.y[sample.id]
   vec.mu[i]  = mean(sample.y)
   vec.var[i] = var(sample.y)
   print(paste("iteration ",i,"/",nreps," complete.", sep=""))
}

## visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,3), pty="s")
plot(vec.x, vec.y, type="l", main="1d signal")  # 1d signal
hist(vec.mu, main="mean CI", xlab="mu")         # mean
abline(v=data.mu, col="red", lwd=4)
hist(vec.var, main="variance CI", xlab="sigma") # variance
abline(v=data.var, col="blue", lwd=4)
par(opar)
}

}
\references{
\insertRef{politis_stationary_1994}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cayleymenger.R
\name{cayleymenger}
\alias{cayleymenger}
\title{Cayley-Menger Determinant}
\usage{
cayleymenger(data)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix of row-stacked observations.}
}
\value{
a list containing\describe{
\item{det}{determinant value.}
\item{vol}{volume attained from the determinant.}
}
}
\description{
Cayley-Menger determinant is a formula of a \eqn{n}-dimensional simplex 
with respect to the squares of all pairwise distances of its vertices.
}
\examples{
## USE 'IRIS' DATASET
data(iris)
X = as.matrix(iris[,1:4])

## COMPUTE CAYLEY-MENGER DETERMINANT
#  since k=4 < n=149, it should be zero.
cayleymenger(X)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdeterminant.R
\name{pdeterminant}
\alias{pdeterminant}
\title{Calculate the Pseudo-Determinant of a Matrix}
\usage{
pdeterminant(A)
}
\arguments{
\item{A}{a square matrix whose pseudo-determinant be computed.}
}
\value{
a scalar value for computed pseudo-determinant.
}
\description{
When a given square matrix \eqn{A} is rank deficient, determinant is zero. 
Still, we can compute the pseudo-determinant by multiplying all non-zero 
eigenvalues. Since thresholding to determine near-zero eigenvalues is subjective, 
we implemented the function as of original limit problem. When matrix is 
non-singular, it coincides with traditional determinant.
}
\examples{
## show the convergence of pseudo-determinant
#  settings
n = 10
A = cov(matrix(rnorm(5*n),ncol=n))   # (n x n) matrix
k = as.double(Matrix::rankMatrix(A)) # rank of A

# iterative computation
ntry = 11
del.vec = exp(-(1:ntry))
det.vec = rep(0,ntry)
for (i in 1:ntry){
  del = del.vec[i]
  det.vec[i] = det(A+del*diag(n))/(del^(n-k))
}

# visualize the results
opar <- par(no.readonly=TRUE)
plot(1:ntry, det.vec, main=paste("true rank is ",k," out of ",n,sep=""),"b", xlab="iterations")
abline(h=pdeterminant(A),col="red",lwd=1.2)
par(opar)

}
\references{
\insertRef{holbrook_differentiating_2018}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sylvester.R
\name{sylvester}
\alias{sylvester}
\title{Solve Sylvester Equation}
\usage{
sylvester(A, B, C)
}
\arguments{
\item{A}{a \eqn{(p\times p)} matrix as above.}

\item{B}{a \eqn{(p\times p)} matrix as above.}

\item{C}{a \eqn{(p\times p)} matrix as above.}
}
\value{
a solution matrix \eqn{X} of size \eqn{(p\times p)}.
}
\description{
The Sylvester equation is of form
\deqn{AX + XB = C}
where \eqn{X} is the unknown and others are given. Though it's possible to have non-square \eqn{A} and \eqn{B} matrices, 
we currently support square matrices only. This is a wrapper of \code{armadillo}'s \code{sylvester} function.
}
\examples{
## simulated example
#  generate square matrices
A = matrix(rnorm(25),nrow=5)
X = matrix(rnorm(25),nrow=5)
B = matrix(rnorm(25),nrow=5)
C = A\%*\%X + X\%*\%B

#  solve using 'sylvester' function
solX = sylvester(A,B,C)
pm1 = "* Experiment with Sylvester Solver"
pm2 = paste("* Absolute Error  : ",norm(solX-X,"f"),sep="")
pm3 = paste("* Relative Error  : ",norm(solX-X,"f")/norm(X,"f"),sep="")
cat(paste(pm1,"\n",pm2,"\n",pm3,sep=""))


}
\references{
\insertRef{sanderson_armadillo_2016}{maotai}

\insertRef{eddelbuettel_rcpparmadillo_2014}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metricdepth.R
\name{metricdepth}
\alias{metricdepth}
\title{Metric Depth}
\usage{
metricdepth(data)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}
}
\value{
a length-\eqn{n} vector of empirical metric depth values.
}
\description{
Compute the metric depth proposed by \insertCite{geenens_2023_StatisticalDepthAbstract;textual}{maotai}, which is 
one generalization of statistical depth function onto the arbitrary metric space. Our implementation assumes that 
given the multivariate data it computes the (empirical) depth for all observations using under the Euclidean regime.
}
\examples{
\dontrun{
## use simple example of iris dataset
data(iris) 
X <- as.matrix(iris[,1:4])
y <- as.factor(iris[,5])

## compute the metric depth
mdX <- metricdepth(X)

## visualize
#  2-d embedding for plotting by MDS
X2d <- maotai::cmds(X, ndim=2)$embed

# get a color code for the metric depth
pal = colorRampPalette(c("yellow","red"))

# draw
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2), pty="s")
plot(X2d, pch=19, main="by class", xlab="", ylab="", col=y)
plot(X2d, pch=19, main="by depth", xlab="", ylab="", col=pal(150)[order(mdX)])
legend("bottomright", col=pal(2), pch=19, legend=round(range(mdX), 2))
par(opar)
}

}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkdist.R
\name{checkdist}
\alias{checkdist}
\title{Check for Distance Matrix}
\usage{
checkdist(d)
}
\arguments{
\item{d}{\code{"dist"} object or \eqn{(N\times N)} matrix of pairwise distances.}
}
\value{
a logical; \code{TRUE} if it satisfies metric property, \code{FALSE} otherwise.
}
\description{
This function checks whether the distance matrix \eqn{D:=d_{ij} = d(x_i, x_j)} satisfies 
three axioms to make itself a semimetric, which are (1) \eqn{d_{ii} = 0}, (2) \eqn{d_{ij} > 0} for \eqn{i\neq j}, and 
(3) \eqn{d_{ij} = d_{ji}}.
}
\examples{
## Let's use L2 distance matrix of iris dataset
data(iris)
dx = as.matrix(stats::dist(iris[,1:4]))

# perturb d(i,j) 
dy = dx  
dy[1,2] <- dy[2,1] <- 10

# run the algorithm
checkdist(dx)
checkdist(dy)

}
\seealso{
\code{\link{checkmetric}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kmeanspp.R
\name{kmeanspp}
\alias{kmeanspp}
\title{K-Means++ Clustering Algorithm}
\usage{
kmeanspp(data, k = 2)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}

\item{k}{the number of clusters.}
}
\value{
a length-\eqn{n} vector of class labels.
}
\description{
\eqn{k}-means++ algorithm is known to be a smart, careful initialization 
technique. It is originally intended to return a set of \eqn{k} points 
as initial centers though it can still be used as a rough clustering algorithm 
by assigning points to the nearest points.
}
\examples{
## use simple example of iris dataset
data(iris) 
mydata = as.matrix(iris[,1:4])
mycol  = as.factor(iris[,5])

## find the low-dimensional embedding for visualization
my2d = cmds(mydata, ndim=2)$embed

## apply 'kmeanspp' with different numbers of k's.
k2 = kmeanspp(mydata, k=2)
k3 = kmeanspp(mydata, k=3)
k4 = kmeanspp(mydata, k=4)
k5 = kmeanspp(mydata, k=5)
k6 = kmeanspp(mydata, k=6)

## visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(2,3))
plot(my2d, col=k2, main="k=2", pch=19, cex=0.5)
plot(my2d, col=k3, main="k=3", pch=19, cex=0.5)
plot(my2d, col=k4, main="k=4", pch=19, cex=0.5)
plot(my2d, col=k5, main="k=5", pch=19, cex=0.5)
plot(my2d, col=k6, main="k=6", pch=19, cex=0.5)
plot(my2d, col=mycol, main="true cluster", pch=19, cex=0.5)
par(opar)

}
\references{
\insertRef{arthur_kmeans_2007}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bmds.R
\name{bmds}
\alias{bmds}
\title{Bayesian Multidimensional Scaling}
\usage{
bmds(
  data,
  ndim = 2,
  par.a = 5,
  par.alpha = 0.5,
  par.step = 1,
  mc.iter = 8128,
  verbose = TRUE
)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}

\item{ndim}{an integer-valued target dimension.}

\item{par.a}{hyperparameter for conjugate prior on variance term, i.e., \eqn{\sigma^2 \sim IG(a,b)}. Note that \eqn{b} is chosen appropriately as in paper.}

\item{par.alpha}{hyperparameter for conjugate prior on diagonal term, i.e., \eqn{\lambda_j \sim IG(\alpha, \beta_j)}. Note that \eqn{\beta_j} is chosen appropriately as in paper.}

\item{par.step}{stepsize for random-walk, which is standard deviation of Gaussian proposal.}

\item{mc.iter}{the number of MCMC iterations.}

\item{verbose}{a logical; \code{TRUE} to show iterations, \code{FALSE} otherwise.}
}
\value{
a named list containing
\describe{
\item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
\item{stress}{discrepancy between embedded and origianl data as a measure of error.}
}
}
\description{
A Bayesian formulation of classical Multidimensional Scaling is presented.
Even though this method is based on MCMC sampling, we only return maximum a posterior (MAP) estimate
that maximizes the posterior distribution. Due to its nature without any special tuning,
increasing \code{mc.iter} requires much computation.
}
\examples{
\donttest{
## use simple example of iris dataset
data(iris) 
idata = as.matrix(iris[,1:4])

## run Bayesian MDS
#  let's run 10 iterations only.
iris.cmds = cmds(idata, ndim=2)
iris.bmds = bmds(idata, ndim=2, mc.iter=5, par.step=(2.38^2)) 

## extract coordinates and class information
cx = iris.cmds$embed # embedded coordinates of CMDS
bx = iris.bmds$embed #                         BMDS
icol = iris[,5]      # class information

## visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(2,1))
mc = paste0("CMDS with STRESS=",round(iris.cmds$stress,4))
mb = paste0("BMDS with STRESS=",round(iris.bmds$stress,4))
plot(cx, col=icol,pch=19,main=mc)
plot(bx, col=icol,pch=19,main=mb)
par(opar)
}

}
\references{
\insertRef{oh_bayesian_2001a}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nem.R
\name{nem}
\alias{nem}
\title{Negative Eigenvalue Magnitude}
\usage{
nem(data)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}
}
\value{
a nonnegative NEM value.
}
\description{
Negative Eigenvalue Magnitude (NEM) is a measure of distortion for the data 
whether they are lying in Euclidean manner or not. When the value is exactly 0, it means 
the data is Euclidean. On the other hand, when NEM is far away from 0, it means not Euclidean. 
The concept of NEM is closely related to the definiteness of a Gram matrix.
}
\examples{
## use simple example of iris dataset 
data(iris) 
mydat = as.matrix(iris[,1:4])

## calculate NEM
nem(mydat)

}
\references{
\insertRef{pekalska_noneuclidean_2006}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lyapunov.R
\name{lyapunov}
\alias{lyapunov}
\title{Solve Lyapunov Equation}
\usage{
lyapunov(A, Q)
}
\arguments{
\item{A}{a \eqn{(p\times p)} matrix as above.}

\item{Q}{a \eqn{(p\times p)} matrix as above.}
}
\value{
a solution matrix \eqn{X} of size \eqn{(p\times p)}.
}
\description{
The Lyapunov equation is of form
\deqn{AX + XA^\top = Q}
where \eqn{A} and \eqn{Q} are square matrices of same size. Above form is also known as \emph{continuous} form. 
This is a wrapper of \code{armadillo}'s \code{sylvester} function.
}
\examples{
## simulated example
#  generate square matrices
A = matrix(rnorm(25),nrow=5)
X = matrix(rnorm(25),nrow=5)
Q = A\%*\%X + X\%*\%t(A)

#  solve using 'lyapunov' function
solX = lyapunov(A,Q)
\dontrun{
pm1 = "* Experiment with Lyapunov Solver"
pm2 = paste("* Absolute Error  : ",norm(solX-X,"f"),sep="")
pm3 = paste("* Relative Error  : ",norm(solX-X,"f")/norm(X,"f"),sep="")
cat(paste(pm1,"\n",pm2,"\n",pm3,sep=""))
}

}
\references{
\insertRef{sanderson_armadillo_2016}{maotai}

\insertRef{eddelbuettel_rcpparmadillo_2014}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/epmeans.R
\name{epmeans}
\alias{epmeans}
\title{EP-means Algorithm for Clustering Empirical Distributions}
\usage{
epmeans(elist, k = 2)
}
\arguments{
\item{elist}{a length \eqn{N} list of either vector or \code{ecdf} objects.}

\item{k}{the number of clusters.}
}
\value{
a named list containing \describe{
\item{cluster}{an integer vector indicating the cluster to which each \code{ecdf} is allocated.}
\item{centers}{a length \eqn{k} list of centroid \code{ecdf} objects.}
}
}
\description{
EP-means is a variant of k-means algorithm adapted to cluster 
multiple empirical cumulative distribution functions under metric structure 
induced by Earth Mover's Distance.
}
\examples{
\donttest{
## two sets of 1d samples, 10 each and add some noise
#    set 1 : mixture of two gaussians
#    set 2 : single gamma distribution

# generate data
elist = list()
for (i in 1:10){
   elist[[i]] = stats::ecdf(c(rnorm(100, mean=-2), rnorm(50, mean=2)))
}
for (j in 11:20){
   elist[[j]] = stats::ecdf(rgamma(100,1) + rnorm(100, sd=sqrt(0.5)))
}

# run EP-means with k clusters 
# change the value below to see different settings
myk   = 2
epout = epmeans(elist, k=myk)

# visualize
opar = par(no.readonly=TRUE)
par(mfrow=c(1,myk))
for (k in 1:myk){
  idk = which(epout$cluster==k)
  for (i in 1:length(idk)){
    if (i<2){
      pm = paste("class ",k," (size=",length(idk),")",sep="")
      plot(elist[[idk[i]]], verticals=TRUE, lwd=0.25, do.points=FALSE, main=pm)
    } else {
      plot(elist[[idk[i]]], add=TRUE, verticals=TRUE, lwd=0.25, do.points=FALSE)
    }
    plot(epout$centers[[k]], add=TRUE, verticals=TRUE, lwd=2, col="red", do.points=FALSE)
  }
}
par(opar)
}

}
\references{
\insertRef{henderson_epmeans_2015}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nef.R
\name{nef}
\alias{nef}
\title{Negative Eigenfraction}
\usage{
nef(data)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}
}
\value{
a nonnegative NEF value.
}
\description{
Negative Eigenfraction (NEF) is a measure of distortion for the data 
whether they are lying in Euclidean manner or not. When the value is exactly 0, it means 
the data is Euclidean. On the other hand, when NEF is far away from 0, it means not Euclidean. 
The concept of NEF is closely related to the definiteness of a Gram matrix.
}
\examples{
## use simple example of iris dataset 
data(iris) 
mydat = as.matrix(iris[,1:4])

## calculate NEF
nef(mydat)

}
\references{
\insertRef{pekalska_noneuclidean_2006}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecdfdistS.R
\name{ecdfdistS}
\alias{ecdfdistS}
\title{Distance Measures between Samples through Empirical Cumulative Distribution Functions}
\usage{
ecdfdistS(
  veclist,
  method = c("KS", "Lp", "Wasserstein"),
  p = 1,
  as.dist = FALSE
)
}
\arguments{
\item{veclist}{a length \eqn{N} list of vectors.}

\item{method}{name of the distance/dissimilarity measure. Case insensitive (default: \code{ks}).}

\item{p}{exponent for \code{Lp} or \code{Wasserstein} distance (default: \code{p=1}).}

\item{as.dist}{a logical; \code{TRUE} to return \code{dist} object, \code{FALSE} to return an \eqn{(N\times N)} symmetric matrix of pairwise distances (default: \code{FALSE}).}
}
\value{
either \code{dist} object of an \eqn{(N\times N)} symmetric matrix of pairwise distances by \code{as.dist} argument.
}
\description{
We measure distance between two empirical cumulative distribution functions of the data. 
Unlike \code{\link[maotai]{ecdfdist}}, this function takes raw data samples as input, and 
internally computes the empirical cumulative distribution functions (ECDF) for distance calculations.
}
\examples{
\donttest{
## toy example : 10 of random and uniform distributions
mylist = list()
for (i in 1:10){
  mylist[[i]] = stats::rnorm(50, sd=2)
}
for (i in 11:20){
  mylist[[i]] = stats::runif(50, min=-5)
}

## compute three distances
d_KS = ecdfdistS(mylist, method="KS")
d_LP = ecdfdistS(mylist, method="Lp")
d_OT = ecdfdistS(mylist, method="Wasserstein")

## visualize
opar = par(no.readonly=TRUE)
par(mfrow=c(1,3), pty="s")
image(d_KS[,nrow(d_KS):1], axes=FALSE, main="Kolmogorov-Smirnov")
image(d_LP[,nrow(d_LP):1], axes=FALSE, main="Lp (p=1)")
image(d_OT[,nrow(d_OT):1], axes=FALSE, main="Wasserstein (p=1)")
par(opar)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trio.R
\name{trio}
\alias{trio}
\title{Trace Ratio Optimation}
\usage{
trio(
  A,
  B,
  C,
  dim = 2,
  method = c("2003Guo", "2007Wang", "2009Jia", "2012Ngo"),
  maxiter = 1000,
  eps = 1e-10
)
}
\arguments{
\item{A}{a \eqn{(p\times p)} symmetric matrix in the numerator term.}

\item{B}{a \eqn{(p\times p)} symmetric matrix in the denomiator term.}

\item{C}{a \eqn{(p\times p)} symmetric constraint matrix. If not provided, it is set as identical matrix automatically.}

\item{dim}{an integer for target dimension. It can be considered as the number of loadings.}

\item{method}{the name of algorithm to be used. Default is \code{2003Guo}.}

\item{maxiter}{maximum number of iterations to be performed.}

\item{eps}{stopping criterion for iterative algorithms.}
}
\value{
a named list containing
\describe{
\item{V}{a \eqn{(p\times dim)} projection matrix.}
\item{tr.val}{an attained maximum scalar value.}
}
}
\description{
This function provides several algorithms to solve the following problem
\deqn{\textrm{max} \frac{tr(V^\top A V)}{tr(V^\top B V)} \textrm{ such that } V^\top C V = I}
where \eqn{V} is a projection matrix, i.e., \eqn{V^\top V = I}. Trace ratio optimization 
is pertained to various linear dimension reduction methods. It should be noted that 
when \eqn{C = I}, the above problem is often reformulated as a generalized eigenvalue problem 
since it's an easier proxy with faster computation.
}
\examples{
## simple test
#  problem setting
p = 5
mydim = 2
A = matrix(rnorm(p^2),nrow=p); A=A\%*\%t(A)
B = matrix(runif(p^2),nrow=p); B=B\%*\%t(B)
C = diag(p)

#  approximate solution via determinant ratio problem formulation
eigAB  = eigen(solve(B,A)) 
V      = eigAB$vectors[,1:mydim]
eigval = sum(diag(t(V)\%*\%A\%*\%V))/sum(diag(t(V)\%*\%B\%*\%V))

#  solve using 4 algorithms
m12 = trio(A,B,dim=mydim, method="2012Ngo")
m09 = trio(A,B,dim=mydim, method="2009Jia")
m07 = trio(A,B,dim=mydim, method="2007Wang")
m03 = trio(A,B,dim=mydim, method="2003Guo")

#  print the results
line1 = '* Evaluation of the cost function'
line2 = paste("* approx. via determinant : ",eigval,sep="")
line3 = paste("* trio by 2012Ngo         : ",m12$tr.val, sep="")
line4 = paste("* trio by 2009Jia         : ",m09$tr.val, sep="")
line5 = paste("* trio by 2007Wang        : ",m07$tr.val, sep="")
line6 = paste("* trio by 2003Guo         : ",m03$tr.val, sep="")
cat(line1,"\n",line2,"\n",line3,"\n",line4,"\n",line5,"\n",line6)

}
\references{
\insertRef{guo_generalized_2003}{maotai}

\insertRef{wang_trace_2007}{maotai}

\insertRef{yangqingjia_trace_2009}{maotai}

\insertRef{ngo_trace_2012}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecdfdist2.R
\name{ecdfdist2}
\alias{ecdfdist2}
\title{Pairwise Measures for Two Sets of Empirical CDFs}
\usage{
ecdfdist2(elist1, elist2, method = c("KS", "Lp", "Wasserstein"), p = 2)
}
\arguments{
\item{elist1}{a length \eqn{M} list of \code{ecdf} objects.}

\item{elist2}{a length \eqn{N} list of \code{ecdf} objects.}

\item{method}{name of the distance/dissimilarity measure. Case insensitive.}

\item{p}{exponent for \code{Lp} or \code{Wasserstein} distance.}
}
\value{
an \eqn{(M\times N)} matrix of pairwise distances.
}
\description{
We measure distance between two sets of empirical cumulative distribution functions (ECDF). For 
simplicity, we only take an input of \code{\link[stats]{ecdf}} objects from \pkg{stats} package.
}
\examples{
\donttest{
## toy example
#  first list : 10 of random and uniform distributions
mylist1 = list()
for (i in 1:10){ mylist1[[i]] = stats::ecdf(stats::rnorm(50, sd=2))}
for (i in 11:20){mylist1[[i]] = stats::ecdf(stats::runif(50, min=-5))}

#  second list : 15 uniform and random distributions
mylist2 = list()
for (i in 1:15){ mylist2[[i]] = stats::ecdf(stats::runif(50, min=-5))}
for (i in 16:30){mylist2[[i]] = stats::ecdf(stats::rnorm(50, sd=2))}

## compute Kolmogorov-Smirnov distance
dm2ks = ecdfdist2(mylist1, mylist2, method="KS")
dm2lp = ecdfdist2(mylist1, mylist2, method="lp")
dm2wa = ecdfdist2(mylist1, mylist2, method="wasserstein")
nrs   = nrow(dm2ks)

## visualize
opar = par(no.readonly=TRUE)
par(mfrow=c(1,3), pty="s")
image(dm2ks[,nrs:1], axes=FALSE, main="Kolmogorov-Smirnov")
image(dm2lp[,nrs:1], axes=FALSE, main="L2")
image(dm2wa[,nrs:1], axes=FALSE, main="Wasserstein")
par(opar)
}

}
\seealso{
\code{\link[stats]{ecdf}} \code{\link{ecdfdist}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cov2corr.R
\name{cov2corr}
\alias{cov2corr}
\title{Convert Covariance into Correlation Matrix}
\usage{
cov2corr(mat)
}
\arguments{
\item{mat}{a \eqn{(p\times p)} covariance matrix.}
}
\value{
a \eqn{(p\times p)} correlation matrix.
}
\description{
Given a covariance matrix, return a correlation matrix that has unit diagonals. 
We strictly impose (and check) whether the given input is a symmetric matrix 
of full-rank.
}
\examples{
\donttest{
# generate an empirical covariance scaled
prep_mat = stats::cov(matrix(rnorm(100*10),ncol=10))
prep_vec = diag(as.vector(stats::runif(10, max=5)))
prep_cov = prep_vec\%*\%prep_mat\%*\%prep_vec

# compute correlation matrix
prep_cor = cov2corr(prep_cov)

# visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2), pty="s")
image(prep_cov, axes=FALSE, main="covariance")
image(prep_cor, axes=FALSE, main="correlation")
par(opar)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WLpdist.R
\name{WLpdist}
\alias{WLpdist}
\title{Pairwise Wasserstein-like Distance between two vMF distributions}
\usage{
WLpdist(means, concentrations)
}
\arguments{
\item{means}{An \eqn{(n \times p)} matrix where each row represents the mean 
direction of one of the \eqn{n} vMF distributions.}

\item{concentrations}{A length-\eqn{n} vector of nonnegative concentration parameters.}
}
\value{
An \eqn{(n \times n)} matrix of pairwise distances.
}
\description{
Given a collection of von Misees-Fisher (vMF) distributions, compute the pairwise 
distance using the Wasserstein-like distance from an approximate Wasserstein geometry.
}
\examples{
\donttest{
# Set seed for reproducibility
set.seed(123)

# Generate two classes of mean directions around north and south poles
means1 = array(0,c(50,2)); means1[,2] = rnorm(50, mean=1, sd=0.25)
means2 = array(0,c(50,2)); means2[,2] = rnorm(50, mean=-1, sd=0.25)
means1 = means1/sqrt(rowSums(means1^2))
means2 = means2/sqrt(rowSums(means2^2))

# Concatenate the mean directions
data_means = rbind(means1, means2)

# Generate concentration parameters
data_concentrations = rnorm(100, mean=20, sd=1)

# Compute the pairwise distance matrix
pdmat = WLpdist(data_means, data_concentrations)

# Visualise the pairwise distance matrix
opar <- par(no.readonly=TRUE)
image(pdmat, main="Pairwise Wasserstein-like Distance")
par(opar)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bwproj.R
\name{bwproj}
\alias{bwproj}
\title{Bures-Wasserstein Projection}
\usage{
bwproj(means, covs, target_dim = 2, max_iter = 100, verbose = TRUE)
}
\arguments{
\item{means}{an \eqn{(n,p)} matrix of Gaussian means, where \eqn{n} is the number of Gaussians and \eqn{p} is the original dimension.}

\item{covs}{a \eqn{(p,p,n)} array of Gaussian covariance matrices.}

\item{target_dim}{an integer specifying the target lower dimension \eqn{d} (default is 2).}

\item{max_iter}{an integer specifying the maximum number of iterations for the optimization (default is 100).}

\item{verbose}{a logical flag indicating whether to print progress messages (default is TRUE).}
}
\value{
a named list containing \describe{
\item{U}{the \eqn{(p,d)} projection matrix mapping original space to the lower-dimensional space.}
\item{proj_means}{the \eqn{(n,d)} matrix of projected Gaussian means.}
\item{proj_covs}{the \eqn{(d,d,n)} array of projected Gaussian covariance matrices.}
\item{iter_obj}{a vector of objective function values at each iteration.}
\item{iter_gnorm}{a vector of gradient norms at each iteration.}
}
}
\description{
Projects a set of high-dimensional Gaussian distributions specified by their means and covariance matrices 
onto a lower-dimensional subspace using the distance-preserving projection method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dpmeans.R
\name{dpmeans}
\alias{dpmeans}
\title{DP-means Algorithm for Clustering Euclidean Data}
\usage{
dpmeans(
  data,
  lambda = 1,
  maxiter = 1234,
  abstol = 1e-06,
  permute.order = FALSE
)
}
\arguments{
\item{data}{an \eqn{(n\times p)} data matrix for each row being an observation.}

\item{lambda}{a threshold to define a new cluster.}

\item{maxiter}{maximum number of iterations.}

\item{abstol}{stopping criterion}

\item{permute.order}{a logical; \code{TRUE} if random order for permutation is used, \code{FALSE} otherwise.}
}
\value{
a named list containing
\describe{
\item{cluster}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
\item{centers}{a list containing information for out-of-sample prediction.}
}
}
\description{
DP-means is a nonparametric clustering method motivated by DP mixture model in that 
the number of clusters is determined by a parameter \eqn{\lambda}. The larger 
the \eqn{\lambda} value is, the smaller the number of clusters is attained. 
In addition to the original paper, we added an option to randomly permute 
an order of updating for each observation's membership as a common 
heuristic in the literature of cluster analysis.
}
\examples{
## define data matrix of two clusters
x1  = matrix(rnorm(50*3,mean= 2), ncol=3)
x2  = matrix(rnorm(50*3,mean=-2), ncol=3)
X   = rbind(x1,x2)
lab = c(rep(1,50),rep(2,50))

## run dpmeans with several lambda values
solA <- dpmeans(X, lambda= 5)$cluster
solB <- dpmeans(X, lambda=10)$cluster
solC <- dpmeans(X, lambda=20)$cluster

## visualize the results
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,4), pty="s")
plot(X,col=lab,  pch=19, cex=.8, main="True", xlab="x", ylab="y")
plot(X,col=solA, pch=19, cex=.8, main="dpmeans lbd=5", xlab="x", ylab="y")
plot(X,col=solB, pch=19, cex=.8, main="dpmeans lbd=10", xlab="x", ylab="y")
plot(X,col=solC, pch=19, cex=.8, main="dpmeans lbd=20", xlab="x", ylab="y")
par(opar)

\donttest{
## let's find variations by permuting orders of update
## used setting : lambda=20, we will 8 runs
sol8 <- list()
for (i in 1:8){
  sol8[[i]] = dpmeans(X, lambda=20, permute.order=TRUE)$cluster
}

## let's visualize
vpar <- par(no.readonly=TRUE)
par(mfrow=c(2,4), pty="s")
for (i in 1:8){
  pm = paste("permute no.",i,sep="")
  plot(X,col=sol8[[i]], pch=19, cex=.8, main=pm, xlab="x", ylab="y")
}
par(vpar)
}

}
\references{
\insertRef{kulis_revisiting_2012}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/movMF_info.R
\name{movMF_info}
\alias{movMF_info}
\title{Extract meaningful information from the von Mises-Fisher mixture model}
\usage{
movMF_info(data, means, concentrations, weights)
}
\arguments{
\item{data}{an \eqn{(n\times d)} data matrix.}

\item{means}{an \eqn{(k\times d)} matrix of means.}

\item{concentrations}{a vector of length \eqn{k} of concentration parameters.}

\item{weights}{a vector of length \eqn{k} of mixing weights.}
}
\value{
a named list containing \describe{
\item{densities}{a vector of length \eqn{n} of the densities of the data points.}
\item{clustering}{a vector of length \eqn{n} of the hard clustering results.}
\item{loglkd}{the log-likelihood of the data.}
\item{AIC}{the Akaike information criterion.}
\item{BIC}{the Bayesian information criterion.}
\item{HQIC}{the Hannan-Quinn information criterion.}
}
}
\description{
Given a mixture of von Mises-Fisher distributions, this function computes 
several related quantities of the data on the unit hypersphere with respect 
to the specified model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cmds.R
\name{cmds}
\alias{cmds}
\title{Classical Multidimensional Scaling}
\usage{
cmds(data, ndim = 2)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}

\item{ndim}{an integer-valued target dimension.}
}
\value{
a named list containing
\describe{
\item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
\item{stress}{discrepancy between embedded and origianl data as a measure of error.}
}
}
\description{
Classical multidimensional scaling aims at finding low-dimensional structure 
by preserving pairwise distances of data.
}
\examples{
## use simple example of iris dataset
data(iris) 
idata = as.matrix(iris[,1:4])
icol  = as.factor(iris[,5])   # class information

## run Classical MDS
iris.cmds = cmds(idata, ndim=2)

## visualize
opar <- par(no.readonly=TRUE)
plot(iris.cmds$embed, col=icol, 
     main=paste0("STRESS=",round(iris.cmds$stress,4)))
par(opar)

}
\references{
\insertRef{torgerson_multidimensional_1952}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mmd2test.R
\name{mmd2test}
\alias{mmd2test}
\title{Kernel Two-sample Test with Maximum Mean Discrepancy}
\usage{
mmd2test(K, label, method = c("b", "u"), mc.iter = 999)
}
\arguments{
\item{K}{kernel matrix or an object of \code{kernelMatrix} class from \pkg{kernlab} package.}

\item{label}{label vector of class indices.}

\item{method}{type of estimator to be used. \code{"b"} for biased and \code{"u"} for unbiased estimator of MMD.}

\item{mc.iter}{the number of Monte Carlo resampling iterations.}
}
\value{
a (list) object of \code{S3} class \code{htest} containing: \describe{
\item{statistic}{a test statistic.}
\item{p.value}{\eqn{p}-value under \eqn{H_0}.}
\item{alternative}{alternative hypothesis.}
\item{method}{name of the test.}
\item{data.name}{name(s) of provided kernel matrix.}
}
}
\description{
Maximum Mean Discrepancy (MMD) as a measure of discrepancy between 
samples is employed as a test statistic for two-sample hypothesis test 
of equal distributions. Kernel matrix \eqn{K} is a symmetric square matrix 
that is positive semidefinite.
}
\examples{
## small test for CRAN submission
dat1 <- matrix(rnorm(60, mean= 1), ncol=2) # group 1 : 30 obs of mean  1
dat2 <- matrix(rnorm(50, mean=-1), ncol=2) # group 2 : 25 obs of mean -1

dmat <- as.matrix(dist(rbind(dat1, dat2)))  # Euclidean distance matrix
kmat <- exp(-(dmat^2))                      # build a gaussian kernel matrix
lab  <- c(rep(1,30), rep(2,25))             # corresponding label

mmd2test(kmat, lab)                         # run the code !

\dontrun{
## WARNING: computationally heavy. 
#  Let's compute empirical Type 1 error at alpha=0.05
niter = 496  
pvals1 = rep(0,niter)
pvals2 = rep(0,niter)
for (i in 1:niter){
  dat = matrix(rnorm(200),ncol=2)
  lab = c(rep(1,50), rep(2,50))
  lbd = 0.1
  kmat = exp(-lbd*(as.matrix(dist(dat))^2))
  pvals1[i] = mmd2test(kmat, lab, method="b")$p.value
  pvals2[i] = mmd2test(kmat, lab, method="u")$p.value
  print(paste("iteration ",i," complete..",sep=""))
}

#  Visualize the above at multiple significance levels
alphas = seq(from=0.001, to=0.999, length.out=100)
errors1 = rep(0,100)
errors2 = rep(0,100)
for (i in 1:100){
   errors1[i] = sum(pvals1<=alphas[i])/niter
   errors2[i] = sum(pvals2<=alphas[i])/niter
}

opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2), pty="s")
plot(alphas, errors1, "b", main="Biased Estimator Error", 
     xlab="alpha", ylab="error", cex=0.5)
abline(a=0,b=1, lwd=1.5, col="red")
plot(alphas, errors2, "b", main="Unbiased Estimator Error", 
     xlab="alpha", ylab="error", cex=0.5)
abline(a=0,b=1, lwd=1.5, col="blue")
par(opar)
}

}
\references{
\insertRef{gretton_kernel_2012}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matderiv.R
\name{matderiv}
\alias{matderiv}
\title{Numerical Approximation to Gradient of a Function with Matrix Argument}
\usage{
matderiv(fn, x, h = 0.001)
}
\arguments{
\item{fn}{a function that takes a matrix of size \eqn{(n\times p)} and returns a scalar value.}

\item{x}{an \eqn{(n\times p)} matrix where the gradient is to be computed.}

\item{h}{step size for centered difference scheme.}
}
\value{
an approximate numerical gradient matrix of size \eqn{(n\times p)}.
}
\description{
For a given function \eqn{f:\mathbf{R}^{n\times p} \rightarrow \mathbf{R}}, 
we use finite difference scheme that approximates a gradient at a given point \eqn{x}. 
In Riemannian optimization, this can be used as a proxy for 
ambient gradient. Use with care since it may accumulate numerical error.
}
\examples{
## function f(X) = <a,Xb> for two vectors 'a' and 'b'
#  derivative w.r.t X is ab'
#  take an example of (5x5) symmetric positive definite matrix

#  problem settings
a   <- rnorm(5)
b   <- rnorm(5)
ftn <- function(X){
  return(sum(as.vector(X\%*\%b)*a))
}       # function to be taken derivative
myX <- matrix(rnorm(25),nrow=5)  # point where derivative is evaluated
myX <- myX\%*\%t(myX)

# main computation
sol.true <- base::outer(a,b)
sol.num1 <- matderiv(ftn, myX, h=1e-1) # step size : 1e-1
sol.num2 <- matderiv(ftn, myX, h=1e-5) #             1e-3
sol.num3 <- matderiv(ftn, myX, h=1e-9) #             1e-5

## visualize/print the results
expar = par(no.readonly=TRUE)
par(mfrow=c(2,2),pty="s")
image(sol.true, main="true solution")
image(sol.num1, main="h=1e-1")
image(sol.num2, main="h=1e-5")
image(sol.num3, main="h=1e-9")
par(expar)

\donttest{
ntrue = norm(sol.true,"f")
cat('* Relative Errors in Frobenius Norm ')
cat(paste("*  h=1e-1   : ",norm(sol.true-sol.num1,"f")/ntrue,sep=""))
cat(paste("*  h=1e-5   : ",norm(sol.true-sol.num2,"f")/ntrue,sep=""))
cat(paste("*  h=1e-9   : ",norm(sol.true-sol.num3,"f")/ntrue,sep=""))
}

}
\references{
\insertRef{kincaid_numerical_2009}{maotai}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsne.R
\name{tsne}
\alias{tsne}
\title{t-SNE Embedding}
\usage{
tsne(data, ndim = 2, ...)
}
\arguments{
\item{data}{an \eqn{(n\times p)} matrix whose rows are observations.}

\item{ndim}{an integer-valued target dimension.}

\item{...}{extra parameters to be used in \code{\link[Rtsne]{Rtsne}} function.}
}
\value{
a named list containing
\describe{
\item{embed}{an \eqn{(n\times ndim)} matrix whose rows are embedded observations.}
\item{stress}{discrepancy between embedded and origianl data as a measure of error.}
}
}
\description{
This function is a simple wrapper of \code{\link[Rtsne]{Rtsne}} function for 
t-Stochastic Neighbor Embedding for finding low-dimensional structure of 
the data embedded in the high-dimensional space.
}
\examples{
\donttest{
## use simple example of iris dataset 
data(iris) 
mydat = as.matrix(iris[,1:4])
mylab = as.factor(iris[,5])

## run t-SNE and MDS for comparison
iris.cmds = cmds(mydat, ndim=2)
iris.tsne = tsne(mydat, ndim=2)

## extract coordinates and class information
cx = iris.cmds$embed # embedded coordinates of CMDS
tx = iris.tsne$embed #                         t-SNE

## visualize
#  main title
mc = paste("CMDS with STRESS=",round(iris.cmds$stress,4),sep="")
mt = paste("tSNE with STRESS=",round(iris.tsne$stress,4),sep="")

#  draw a figure
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,2))
plot(cx, col=mylab, pch=19, main=mc)
plot(tx, col=mylab, pch=19, main=mt)
par(opar)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weiszfeld.R
\name{weiszfeld}
\alias{weiszfeld}
\title{Weiszfeld Algorithm for Computing L1-median}
\usage{
weiszfeld(X, weights = NULL, maxiter = 496, abstol = 1e-06)
}
\arguments{
\item{X}{an \eqn{(n\times p)} matrix for \eqn{p}-dimensional signal. If vector is given, it is assumed that \eqn{p=1}.}

\item{weights}{\code{NULL} for equal weight \code{rep(1/n,n)}; otherwise, it has to be a vector of length \eqn{n}.}

\item{maxiter}{maximum number of iterations.}

\item{abstol}{stopping criterion}
}
\description{
Geometric median, also known as L1-median, is a solution to the following problem
\deqn{\textrm{argmin} \sum_{i=1}^n \| x_i - y \|_2 }
for a given data \eqn{x_1,x_2,\ldots,x_n \in R^p}.
}
\examples{
## generate sin(x) data with noise for 100 replicates
set.seed(496)
t = seq(from=0,to=10,length.out=20)
X = array(0,c(100,20))
for (i in 1:100){
   X[i,] = sin(t) + stats::rnorm(20, sd=0.5)
}

## compute L1-median and L2-mean
vecL2 = base::colMeans(X)
vecL1 = weiszfeld(X)

## visualize
opar <- par(no.readonly=TRUE)
par(mfrow=c(1,3), pty="s")
matplot(t(X[1:5,]), type="l", main="5 generated data", ylim=c(-2,2))
plot(t, vecL2, type="l", col="blue", main="L2-mean",   ylim=c(-2,2))
plot(t, vecL1, type="l", col="red",  main="L1-median", ylim=c(-2,2))
par(opar)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LEiDA.R
\name{LEiDA}
\alias{LEiDA}
\title{Leading Eigenvector Dynamics Analysis}
\usage{
LEiDA(X, TR, bp = c(0.01, 0.1), b_ord = 2)
}
\arguments{
\item{X}{A \eqn{(T,N)} matrix of multivariate time series data, where \eqn{T} is the number of time points and \eqn{N} is the number of ROIs.}

\item{TR}{Repetition time (in secounds)}

\item{bp}{Bandpass filter, a vector of length 2 with the lower and upper bounds of the bandpass filter in Hz. Default is \code{c(0.01, 0.10)}.}

\item{b_ord}{Butterworth order, a positive integer. Default is 2.}
}
\value{
a list containing \describe{
\item{V}{A \eqn{(T,N)} matrix of the leading eigenvector time series.}
\item{FCD_cos}{A \eqn{(T,T)} matrix of the functional connectivity dynamics (FCD) using cosine similarity.}
\item{FCD_cor}{A \eqn{(T,T)} matrix of the functional connectivity dynamics (FCD) using Pearson correlation.}
}
}
\description{
Compute the leading eigenvector dynamics analysis (LEiDA) of a 
multivariate time series as appearing in computational neuroscience.
}
